/*
 * Copyright Fraunhofer ISE, energy & meteo Systems GmbH, and other contributors 2011
 *
 * This file is part of openIEC61850.
 * For more information visit http://www.openmuc.org 
 *
 * openIEC61850 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * openIEC61850 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with openIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openiec61850.common.model;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * 
 * @author Stefan Feuerhahn
 * 
 */
public class DataObject extends ModelNode {

	private String cdcName;

	/**
	 * DataObject can have the ObjectType DATA_OBJECT or FC_DATA_OBJECT
	 * 
	 * @param objectReference
	 * @param children
	 *            can have ObjectType of DATA_OBJECT, FC_DATA_OBJECT,
	 *            ARRAY_OF_DO, ARRAY_OF_FCDO, ARRAY_OF_DA,
	 *            CONSTRUCTED_DATA_ATTRIBUTE, or BASIC_DATA_ATTRIBUTE
	 * @param cdcName
	 * @param fc
	 */
	public DataObject(ObjectReference objectReference, List<ModelNode> children, String cdcName, FunctionalConstraint fc) {

		this.children = new LinkedHashMap<String, ModelNode>();
		this.objectReference = objectReference;
		for (ModelNode child : children) {
			this.children.put(child.getReference().getName(), child);
		}
		this.cdcName = cdcName;
		this.fc = fc;
	}

	public DataObject(ObjectReference objectReference, List<ModelNode> children, String cdcName) {
		this(objectReference, children, cdcName, null);
	}

	@Override
	public DataObject copy() {
		List<ModelNode> childCopies = new ArrayList<ModelNode>(children.size());
		for (ModelNode childNode : children.values()) {
			childCopies.add(childNode.copy());
		}
		return new DataObject(objectReference, childCopies, cdcName, fc);
	}

	public List<DataObject> getFCDataObjects() {

		List<DataObject> fcdos = new LinkedList<DataObject>();

		if (fc != null) {
			fcdos.add(this);
			return fcdos;
		}

		Map<FunctionalConstraint, List<ModelNode>> subFCDataMap = new LinkedHashMap<FunctionalConstraint, List<ModelNode>>();

		for (FunctionalConstraint fc : FunctionalConstraint.values()) {
			subFCDataMap.put(fc, new LinkedList<ModelNode>());
		}

		for (ModelNode childNode : children.values()) {
			if (childNode.getFunctionalConstraint() == null) {
				if (childNode instanceof DataObject) {
					for (ModelNode fcChildNode : ((DataObject) childNode).getFCDataObjects()) {
						subFCDataMap.get(fcChildNode.getFunctionalConstraint()).add(fcChildNode);
					}
				}
				else {
					for (ModelNode fcChildNode : ((Array) childNode).getArraysOfFCData()) {
						subFCDataMap.get(fcChildNode.getFunctionalConstraint()).add(fcChildNode);
					}
				}
			}
			else {
				subFCDataMap.get(childNode.getFunctionalConstraint()).add(childNode);
			}

		}

		for (FunctionalConstraint fc : FunctionalConstraint.values()) {
			if (subFCDataMap.get(fc).size() > 0) {
				fcdos.add(new DataObject(objectReference, subFCDataMap.get(fc), cdcName, fc));
			}
		}

		return fcdos;

	}

	public String getCDCName() {
		return cdcName;
	}

}