/*
 * Copyright Fraunhofer ISE, energy & meteo Systems GmbH, and other contributors 2011
 *
 * This file is part of openIEC61850.
 * For more information visit http://www.openmuc.org 
 *
 * openIEC61850 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * openIEC61850 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with openIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openiec61850.common.model.datypes.commonacsi;

import java.util.Date;

import org.openiec61850.common.model.datypes.DAType;
import org.openiec61850.common.model.BasicDataAttribute;
import org.openiec61850.common.model.FunctionalConstraint;
import org.openiec61850.common.model.ObjectReference;
import org.openiec61850.common.model.datypes.DAType;

/**
 * See 61850-7-2 chapter TypeDefinitions/BasicTypes
 * 
 * @author bertram
 */
public class Timestamp extends BasicDataAttribute implements Comparable<Timestamp> {
	/**
	 * The SecondSinceEpoch shall be the interval in seconds continuously
	 * counted from the epoch 1970-01-01 00:00:00 UTC
	 */
	public long secondsSinceEpoch;

	/**
	 * The attribute FractionOfSecond shall be the fraction of the current
	 * second when the value of the TimeStamp has been determined. The fraction
	 * of second shall be calculated as
	 * <code>(SUM from I = 0 to 23 of bi*2**–(I+1) s).</code>
	 * 
	 * NOTE 1 The resolution is the smallest unit by which the time stamp is
	 * updated. The 24 bits of the integer provides 1 out of 16777216 counts as
	 * the smallest unit; calculated by 1/2**24 which equals approximately 60
	 * ns.
	 * 
	 * NOTE 2 The resolution of a time stamp may be 1/2**1 (= 0,5 s) if only the
	 * first bit is used; or may be 1/2**2 (= 0,25 s) if the first two bits are
	 * used; or may be approximately 60 ns if all 24 bits are used. The
	 * resolution provided by an IED is outside the scope of this standard.
	 */
	public int fractionOfSecond;
	/**
	 * The TimeQuality shall provide information about the time source of the
	 * sending IED
	 */
	public TimeQuality timeQuality = new TimeQuality();

	// public Timestamp() {
	// this(new Date());
	// }
	//
	// public Timestamp(Date date) {
	// setValue(date);
	// }

	public Timestamp(ObjectReference objectReference, FunctionalConstraint fc, String sAddr, Date date, boolean dchg,
			boolean dupd) {
		this.objectReference = objectReference;
		this.fc = fc;
		this.sAddr = sAddr;
		this.basicType = DAType.TIMESTAMP;
		setValue(date);
		this.dchg = dchg;
		this.dupd = dupd;
	}

	public void setValue(Date newValue) {
		if (newValue == null) {
			timeQuality = TimeQuality.INVALID;
		}
		else {
			secondsSinceEpoch = newValue.getTime() / 1000L;
			fractionOfSecond = (int) ((newValue.getTime() % 1000L) / 1000.0 * (2 ^ 24));
			timeQuality = TimeQuality.DEFAULT;
		}
	}

	public Date getValue() {
		long time = secondsSinceEpoch * 1000L + (fractionOfSecond / (2 ^ 24) * 1000);
		return new Date(time);
	}

	public int compareTo(Timestamp o) {
		if (secondsSinceEpoch < o.secondsSinceEpoch) {
			return -1;
		}
		if (secondsSinceEpoch > o.secondsSinceEpoch) {
			return 1;
		}
		if (fractionOfSecond < o.fractionOfSecond) {
			return -1;
		}
		if (fractionOfSecond > o.fractionOfSecond) {
			return 1;
		}
		return 0;
	}

	@Override
	public boolean equals(Object obj) {
		return obj instanceof Timestamp && compareTo((Timestamp) obj) == 0;
	}

	@Override
	public int hashCode() {
		return (int) (secondsSinceEpoch * 1237L + fractionOfSecond);
	}

	// @Override
	// public String toString() {
	// return DateFormat.getDateTimeInstance().format(getValue());
	// }

	public void setDefault() {
		// set to current date
		setValue(new Date());
	}

	@Override
	public Timestamp copy() {
		return new Timestamp(this.getReference(), this.getFunctionalConstraint(), this.sAddr, this.getValue(),
				this.dchg, this.dupd);
	}
}
