/*
 * Copyright Fraunhofer ISE, energy & meteo Systems GmbH, and other contributors 2011
 *
 * This file is part of openIEC61850.
 * For more information visit http://www.openmuc.org 
 *
 * openIEC61850 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * openIEC61850 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with openIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openiec61850.server;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.openiec61850.common.ACSIClass;
import org.openiec61850.common.ServiceError;
import org.openiec61850.common.model.Array;
import org.openiec61850.common.model.BasicDataAttribute;
import org.openiec61850.common.model.ConstructedDataAttribute;
import org.openiec61850.common.model.DataObject;
import org.openiec61850.common.model.DataSet;
import org.openiec61850.common.model.FunctionalConstraint;
import org.openiec61850.common.model.LogicalNode;
import org.openiec61850.common.model.ModelNode;
import org.openiec61850.common.model.ObjectReference;
import org.openiec61850.common.model.Server;
import org.openiec61850.common.model.datypes.basic.BOOLEAN;
import org.openiec61850.common.model.datypes.basic.INT32U;
import org.openiec61850.common.model.datypes.basic.INT64;
import org.openiec61850.common.model.datypes.basic.INT8U;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_129;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_255;
import org.openiec61850.common.model.datypes.commonacsi.Timestamp;
import org.openiec61850.common.model.report.BufferedReportControlBlock;
import org.openiec61850.common.model.report.OptFields;
import org.openiec61850.common.model.report.Report;
import org.openiec61850.common.model.report.ReportControlBlock;
import org.openiec61850.common.model.report.ReportEntryData;
import org.openiec61850.common.model.report.ReportEntryData.ReasonCode;
import org.openiec61850.common.model.report.TriggerConditions;
import org.openiec61850.common.model.report.UnbufferedReportContrlBlock;
import org.openiec61850.server.data.ConfigurationException;
import org.openiec61850.server.data.DataSource;
import org.openiec61850.server.scsm.SCSMConnectionHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Stefan Feuerhahn
 * @author Michael Zillgith
 * @author Tobias Weidelt
 * 
 */
public class AccessPoint implements ServerACSI {

	private final Server serverModel;
	private final String name;
	private final ServiceSupport serviceSupport;
	private DataSource dataSource;
	private Thread dataSourceThread;
	private final Map<Long, Map<String, DataSet>> nonPersistentDS = new LinkedHashMap<Long, Map<String, DataSet>>();

	// TODO is SqNum for both buffered and unbuffered reports?
	private int reportSqNum = 0;
	private int reportSubSqNum = 0;

	// TODO synchronize these lists?
	private final Map<String, List<DataSet>> triggerDchg = new LinkedHashMap<String, List<DataSet>>();
	private final Map<String, List<DataSet>> triggerDupd = new LinkedHashMap<String, List<DataSet>>();
	private final Map<String, List<DataSet>> triggerQchg = new LinkedHashMap<String, List<DataSet>>();

	// DataSet references to report control blocks
	private final Map<DataSet, List<ReportControlBlock>> dsRCB = new LinkedHashMap<DataSet, List<ReportControlBlock>>();

	// lock for data set related functionality
	Lock dsLock = new ReentrantLock();

	private static Logger logger = LoggerFactory.getLogger(AccessPoint.class);

	/**
	 * Menge aktiver {@link Association}s. Wenn man die füllt, muss man auch
	 * dafür sorgen, dass {@link Association}s nach Ende ihrer Verbindung da
	 * wieder raus entfernt werden, damit sie Garbage-Collected werden können.
	 */
	private final List<Association> associations;

	public Association associate(SCSMConnectionHandler clientHandler, String auth) throws ServiceError {

		// public Association associate(SCSMConnectionHandler clientHandler,
		// String auth, byte[] clientIPAddress) throws ServiceError {

		// TODO check authentication string

		Association association = new Association(clientHandler);
		associations.add(association);
		return association;
	}

	public AccessPoint(String name, Server server, ServiceSupport serviceSupport) {
		this.name = name;
		this.serverModel = server;
		this.serviceSupport = serviceSupport;

		// generate trigger table for reports
		populateTriggers();

		this.associations = new ArrayList<Association>();

	}

	public Server getServer() {
		return serverModel;
	}

	public String getName() {
		return name;
	}

	public void abort(Association assoc, String reason) throws ServiceError {
		// TODO Auto-generated method stub
		nonPersistentDS.clear();
	}

	public void release(Association assoc) throws ServiceError {
		// FIXME Auto-generated method stub
		nonPersistentDS.clear();

	}

	public void close() {
		// TODO Determine what to do here
		// clear non-persistent data sets
		nonPersistentDS.clear();
	}

	public List<String> getServerDirectory(Association assoc, ServerDirectoryType serverDirectoryType)
			throws ServiceError {
		if (serverDirectoryType == ServerDirectoryType.LOGICAL_DEVICES) {
			List<String> logicalDeviceNames = new ArrayList<String>(serverModel.getChildren().size());
			for (ModelNode logicalDevice : serverModel.getChildren()) {
				logicalDeviceNames.add(logicalDevice.getNodeName());
			}
			return logicalDeviceNames;
		}
		else {
			throw new ServiceError(ServiceError.CLASS_NOT_SUPPORTED);
		}
	}

	public List<String> getLogicalDeviceDirectory(Association assoc, String logicalDeviceName) throws ServiceError {
		ModelNode logicalDevice = serverModel.getChild(logicalDeviceName);
		if (logicalDevice == null) {
			throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE);
		}
		List<String> logicalNodeNames = new ArrayList<String>(logicalDevice.getChildren().size());
		for (ModelNode logicalNode : logicalDevice) {
			logicalNodeNames.add(logicalNode.getReference().toString());
		}
		return logicalNodeNames;
	}

	public List<String> getLogicalNodeDirectory(Association assoc, String logicalNodeReference, ACSIClass acsiClass)
			throws ServiceError {
		ModelNode modelNode = serverModel.findSubNode(new ObjectReference(logicalNodeReference), null);
		if (modelNode == null || !(modelNode instanceof LogicalNode)) {
			throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE);
		}
		LogicalNode logicalNode = (LogicalNode) modelNode;

		List<String> instanceNames = new LinkedList<String>();

		switch (acsiClass) {
		case DATA_OBJECT:
			for (ModelNode dataObject : logicalNode.getChildren()) {
				instanceNames.add(dataObject.getReference().getName());
			}
			break;
		case DATA_SET:
			// adding configured and persistent data sets
			try {
				dsLock.lock();
				for (DataSet dataSet : logicalNode.getDataSets()) {
					instanceNames.add(dataSet.getNodeName());
				}
			} finally {
				dsLock.unlock();
			}

			break;

		case URCB:
			for (ReportControlBlock cb : logicalNode.getReportControlBlocks()) {
				if (cb instanceof UnbufferedReportContrlBlock) {
					instanceNames.add(cb.getReference().getName());
				}
			}

		case BRCB:
			for (ReportControlBlock cb : logicalNode.getReportControlBlocks()) {
				if (cb instanceof BufferedReportControlBlock) {
					instanceNames.add(cb.getReference().getName());
				}
			}

		default:
			// TODO Implement other ACSI classes
		}
		return instanceNames;
	}

	public ModelNode getDataDefinition(Association association, ObjectReference objectReference, FunctionalConstraint fc)
			throws ServiceError {
		return serverModel.findSubNode(objectReference, fc);
	}

	public ModelNode getDataValues(Association association, String ref, FunctionalConstraint fc) throws ServiceError {

		ModelNode node = serverModel.findSubNode(new ObjectReference(ref), fc);
		if (node == null) {
			throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE, "object not found: " + ref);
		}

		if (!(node instanceof DataObject) && !(node instanceof Array) && !(node instanceof BasicDataAttribute)
				&& !(node instanceof ConstructedDataAttribute)) {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "invalid object reference " + ref);
		}

		ModelNode nodeCopy = node.copy();

		List<BasicDataAttribute> basicDataAttributes = nodeCopy.getBasicDataAttributes();
		dataSource.readValues(basicDataAttributes);
		return nodeCopy;
	}

	public ModelNode getRCB(Association association, ObjectReference objRef) throws ServiceError {

		ModelNode lnNode = serverModel.findSubNode(new ObjectReference(objRef.getParentRef()), null);

		if (!(lnNode instanceof LogicalNode)) {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "invalid object reference " + objRef);
		}

		List<ReportControlBlock> rcbs = ((LogicalNode) lnNode).getReportControlBlocks();

		for (ReportControlBlock rcb : rcbs) {
			if (rcb.getReference().toString().equals(objRef.toString())) {
				return rcb;
			}
		}

		return null;
	}

	// public ModelNode getRCB(Association association, String objRef,
	// FunctionalConstraint fc) throws ServiceError {
	//
	// ModelNode node = serverModel.findSubNode(new ObjectReference(objRef),
	// fc);
	// if (node == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "object not found: " + objRef);
	// }
	//
	// if (!(node instanceof ReportControlBlock)) {
	// throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
	// "invalid object reference " + objRef);
	// }
	//
	// return node.copy();
	//
	// }

	public void setDataValues(Association association, ModelNode container) throws ServiceError {
		// container is already a copy
		List<BasicDataAttribute> basicDataAttributes = container.getBasicDataAttributes();
		// try {
		dataSource.writeValues(basicDataAttributes);
		// } catch (ConfigurationException e) {
		// throw new ServiceError(ServiceError.CLASS_NOT_SUPPORTED, e);
		// }
	}

	public List<ServiceError> setDataSetValues(Association association, List<ModelNode> members) {
		// members are a copy already
		List<ServiceError> results = new ArrayList<ServiceError>();

		for (ModelNode mem : members) {
			try {
				dataSource.writeValues(mem.getBasicDataAttributes());
				results.add(new ServiceError(ServiceError.NO_ERROR));
			} catch (ServiceError e) {
				results.add(e);
			}
		}

		return results;
	}

	public void createDataSet(Association association, String dataSetName, List<String> dsMemberList,
			List<FunctionalConstraint> fc) throws ServiceError {
		// LDName/LNName.FC.DO.DA etc...
		// construct the data set

		List<ModelNode> members = new ArrayList<ModelNode>();
		Iterator<String> itMem = dsMemberList.iterator();
		Iterator<FunctionalConstraint> itFC = fc.iterator();

		// construct list of members
		while (itMem.hasNext()) {

			ModelNode mem = serverModel.findSubNode(new ObjectReference(itMem.next()), itFC.next());

			if (mem == null) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT, "Cannot find member of data set");
			}
			members.add(mem);

		}

		// add data set
		ObjectReference dsRef = new ObjectReference(dataSetName);
		int at = dataSetName.indexOf("@");

		if (at < 0) { // persistent
			ModelNode lnNode = serverModel.findSubNode(new ObjectReference(dsRef.getParentRef()), null);

			if (lnNode == null || !(lnNode instanceof LogicalNode)) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "Data set reference incorrect");
			}

			try {
				dsLock.lock();
				((LogicalNode) lnNode).addDataSet(new DataSet(dsRef, members, true));
			} finally {
				dsLock.unlock();
			}
		}

		else {// non-persistent
				// no concurrent issues since each thread should only have
				// access to
				// its own list of non-persistent data sets
			if (nonPersistentDS.get(Thread.currentThread().getId()) == null) {
				nonPersistentDS.put(Thread.currentThread().getId(), new LinkedHashMap<String, DataSet>());
			}
			nonPersistentDS.get(Thread.currentThread().getId()).put(dsRef.getName(), new DataSet(dsRef, members, true));
		}
	}

	public ModelNode getDataSetDirectory(Association association, String dataSetName) throws ServiceError {

		ModelNode ds = null;

		int at = dataSetName.indexOf('@');

		if (at < 0) {// persistent
			int dot = dataSetName.indexOf('.');
			ModelNode lnNode = serverModel.findSubNode(new ObjectReference(dataSetName.substring(0, dot)), null);
			if (lnNode == null) {
				throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
						"getDataSetDirectory: cannot find logical node");
			}

			try {
				dsLock.lock();
				ds = ((LogicalNode) lnNode).getDataSet(dataSetName.substring(dot + 1));
				if (ds == null) {
					throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE, "Data set does not exist");
				}
				ds = ds.copy();
			} finally {
				dsLock.unlock();
			}

		}
		else {
			ds = nonPersistentDS.get(Thread.currentThread().getId()).get(
					dataSetName.substring(dataSetName.lastIndexOf(".")));
			if (ds == null) {
				throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE, "Data set does not exist");
			}
			ds = ds.copy();
		}

		return ds;
	}

	public int[] deleteDataSet(Association association, String dataSetName) throws ServiceError {
		int at = dataSetName.indexOf('@');
		int[] numberMatchedDeleted = new int[2];

		if (at < 0) {// either persistent or configured
			int dot = dataSetName.indexOf('.');

			LogicalNode lnNode = (LogicalNode) serverModel.findSubNode(
					new ObjectReference(dataSetName.substring(0, dot)), null);
			if (lnNode == null) {
				throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE, "deleteDataSet: cannot find logical node");
			}
			DataSet ds = lnNode.getDataSet(dataSetName.substring(dot + 1));

			if (ds == null) {
				throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE, "deleteDataSet: cannot find data set");
			}
			numberMatchedDeleted[0] = 1; // there is a match
			// otherwise of type persistent
			// TODO need to check if reports/logs are pointing to this data set

			if (ds.isDeletable()) {// need to modify check

				try {
					dsLock.lock();
					lnNode.deleteDataSet(dataSetName.substring(dot + 1));
					numberMatchedDeleted[1] = 1;
					return numberMatchedDeleted;
				} finally {
					dsLock.unlock();
				}
			}

		}

		else {// non-persistent
				// TODO need to check if any reports/logs are referring to
			nonPersistentDS.get(Thread.currentThread().getId()).remove(dataSetName);
			numberMatchedDeleted[1] = 1;
			return numberMatchedDeleted;

		}
		// was not deleted
		numberMatchedDeleted[1] = 0;
		return numberMatchedDeleted;
	}

	public List<ModelNode> getDataSetValues(Association association, Collection<ModelNode> members) throws ServiceError {
		// members is not a copy at this point
		List<ModelNode> membersCopy = new ArrayList<ModelNode>();
		List<BasicDataAttribute> readValues = new ArrayList<BasicDataAttribute>();
		ModelNode nodeCopy;

		for (ModelNode mem : members) {
			nodeCopy = mem.copy();
			membersCopy.add(nodeCopy);
			readValues.addAll(nodeCopy.getBasicDataAttributes());
		}

		dataSource.readValues(readValues);

		return membersCopy;
	}

	public void initDataSource(String dataSourceClassName) throws ConfigurationException {
		Object instance = null;
		try {
			instance = Class.forName(dataSourceClassName).newInstance();
		} catch (Exception e) {
			throw new ConfigurationException("Exception instantiating DataSource: " + dataSourceClassName + e);
		}
		if (!(instance instanceof DataSource)) {
			throw new ConfigurationException(instance.getClass().getName() + " does not implement the "
					+ DataSource.class.getName() + " interface");
		}

		dataSource = (DataSource) instance;

		/*
		 * provide a link to the AccessPoint and pass in the shared memory (i.e.
		 * rcbs, list of associations, etc)
		 */
		dataSource.initialize(this, triggerDchg.keySet(), triggerQchg.keySet(), triggerDupd.keySet());
		dataSourceThread = new Thread(dataSource);
		dataSourceThread.start();

	}

	public void selectEditSG(Association association, String ref, int sGN) throws ServiceError {
		// TODO Auto-generated method stub

	}

	public void setURCBValues(Association association, UnbufferedReportContrlBlock urcb, ModelNode value)
			throws ServiceError {

		ModelNode resv = urcb.getChild("Resv");
		if (((BOOLEAN) resv).getValue() == true) {// only the client that
			// reserved it can make changes
			// need to check if it is the client that reserved it

			// TODO handle how?
			// if (urcb.getOwner() != association.getIPAddress()) {
			// throw new ServiceError(ServiceError.ACCESS_VIOLATION,
			// "setURCBValues: this client cannot set values");
			// }

		}

		ModelNode child = urcb.getChild(value.getNodeName());
		if (child == null) {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
					"setURCBValues: report attribute is not valid");
		}

		if (value.getNodeName().equals("RptEna")) {

			if (((BOOLEAN) value).getValue() == true) {

				if (((BOOLEAN) child).getValue() == false) {
					// resetting the value of sqNum to 0 each time it has been
					// enabled from idle
					INT8U sqNum = (INT8U) urcb.getChild("SqNum");
					sqNum.setValue((byte) 0);
				}

				// if(((BOOLEAN)urcb.getChild("Resv")).getValue() == false){
				((BOOLEAN) urcb.getChild("Resv")).setValue(true);
				// }

				// if(((OCTET_STRING_64)urcb.getChild("Owner")).getValue() ==
				// null){
				// ((OCTET_STRING_64)urcb.getChild("Owner")).setValue(association.getIPAddress());
				// }

				// TODO
				// urcb.setOwner(association.getIPAddress());

			}
			else {
				// reset the reservation and owner of the rcb
				// if(((BOOLEAN)urcb.getChild("Resv")).getValue() == true){
				((BOOLEAN) urcb.getChild("Resv")).setValue(false);
				// }

				// if(urcb.getChild("Owner") != null ||
				// ((OCTET_STRING_64)urcb.getChild("Owner")).getValue() !=
				// null){
				// ((OCTET_STRING_64)urcb.getChild("Owner")).setValue(new
				// byte[0]);
				// }
				urcb.setOwner(null);
			}

			((BOOLEAN) child).setValue(((BOOLEAN) value).getValue());

		}

		else if (value.getNodeName().equals("Resv")) {
			((BOOLEAN) child).setValue(((BOOLEAN) value).getValue());

			if (((BOOLEAN) value).getValue() == true) {
				// urcb.addChild(new OCTET_STRING_64(new ObjectReference(value
				// .getReference().toString()
				// + ".Owner"), null, "", association
				// .getIPAddress(), false, false));
				// ((OCTET_STRING_64)urcb.getChild("Owner")).setValue(association.getIPAddress());

				// TODO
				// urcb.setOwner(association.getIPAddress());
			}
		}

		else if (value.getNodeName().equals("ConfRev")) {
			// ((INT32U)child).setValue(((INT32U)value).getValue());
			// not writable but IEDScout sends a write request
			// TODO return an error?
			return;
		}

		else if (value.getNodeName().equals("SqNum")) {
			// not writable but IEDScout sends a write request
			// TODO return an error?
			return;
		}

		else if (value.getNodeName().equals("IntgPd")) {
			if (((TriggerConditions) urcb.getChild("TrgOps")).isIntegrity()) {
				((INT64) child).setValue(((INT64) value).getValue());

				// TODO implement setup of integrity reports
			}

			else {
				throw new ServiceError(ServiceError.ACCESS_VIOLATION);
			}
		}

		// Can be set with RptEna = false or true
		else if (value.getNodeName().equals("GI")) {

			if (((TriggerConditions) urcb.getChild("TrgOps")).isGeneralInterrogation()) {
				((BOOLEAN) child).setValue(((BOOLEAN) value).getValue());
				// TODO start process of sending GI report: after setting to
				// true, the GI process shall be started.
				// After initializing this process, the value of GI shall be set
				// back to false (see iec61850-7-2, page 103
			}
			else {
				throw new ServiceError(ServiceError.ACCESS_VIOLATION);
			}

		}

		else {
			// these can only be set when rptEna = false
			if (((BOOLEAN) urcb.getChild("RptEna")).getValue() == false) {

				if (value.getNodeName().equals("RptID")) {
					((VISIBLE_STRING_129) child).setValue(((VISIBLE_STRING_129) value).getValue());
				}
				// else if (value instanceof DataSet) {
				else if (value.getNodeName().equals("DatSet")) {
					// update trigger information

					// if(!(urcb.getChild(urcb.getDataSetRef()).getReference().toString().equals(value.getReference().toString()))){
					if (((VISIBLE_STRING_255) urcb.getChild("DatSet")).getValue().equals(
							((VISIBLE_STRING_255) value).getValue())) {
						INT64 confRev = (INT64) urcb.getChild("ConfRev");
						confRev.setValue(confRev.getValue() + 1);

						DataSet ds = (DataSet) getDataSetDirectory(association, ((VISIBLE_STRING_255) value).getValue());

						updateTriggers((DataSet) urcb.getDataSet(), ds, urcb);

						urcb.setDataSet(ds);
					}
				}
				else if (value.getNodeName().equals("OptFlds")) {
					((OptFields) urcb.getChild("OptFlds")).setValue(((OptFields) value).getValue());

				}

				else if (value.getNodeName().equals("BufTm")) {
					((INT32U) child).setValue(((INT32U) value).getValue());
				}

				else if (value.getNodeName().equals("TrgOps")) {
					((TriggerConditions) urcb.getChild("TrgOps")).setValue(((TriggerConditions) value).getValue());
				}
				else {
					throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
				}
			}

			else {// trying to set value when rptEna is true not allowed
				throw new ServiceError(ServiceError.ACCESS_VIOLATION);
			}
		}

		// TODO check if integrity is set in trigger options and the integrity
		// period is set to start the sending of integrity periods

		// TODO handle GI requests if GI has been written to true. Reset the
		// value back to false
	}

	public void notificationIndication(ModelNode node, ReasonCode rc) {
		// TODO figure out if it is a buffered/unbuffered/log - call appropriate
		// functions
		List<DataSet> datSets = null;
		List<ReportControlBlock> rcbs = null;

		if (rc == ReasonCode.DCHG) {
			datSets = triggerDchg.get(node.getReference().toString());
		}
		else if (rc == ReasonCode.QCHG) {
			datSets = triggerQchg.get(node.getReference().toString());
		}
		else if (rc == ReasonCode.DUPD) {
			datSets = triggerDupd.get(node.getReference().toString());
		}
		// TODO handle integrity and GI cases - handled in a separate
		// method/thread?

		// TODO need to update values of DO or DA so that it can be included
		// into report

		for (DataSet ds : datSets) {

			// need to see what node is a member of the data set (i.e. DO, DA)
			// The report is sent for the member
			ModelNode reportNode = null;
			for (int i = node.getReference().size() - 1; i > 0; i--) {
				reportNode = ds.getChild(node.getReference().get(i));
				if (reportNode != null
						&& (reportNode instanceof DataObject || reportNode instanceof BasicDataAttribute || reportNode instanceof ConstructedDataAttribute)) {
					break;
				}
			}

			// update the members of the node to be reported
			if (reportNode.getReference().toString().equals(node.getReference().toString())) {
				reportNode = node; // node should have updated value
			}

			else if (reportNode instanceof DataObject) {
				// need to update the basic data attributes of the data object
				// which will be reported
				reportNode = reportNode.copy();
				try {
					dataSource.readValues(reportNode.getBasicDataAttributes());
				} catch (ServiceError e) {
					logger.warn("problem during reading member of data set for report");
				}
			}

			rcbs = dsRCB.get(ds);
			for (ReportControlBlock rcb : rcbs) {
				if (((BOOLEAN) rcb.getChild("RptEna")).getValue() == true) {
					if (rcb instanceof UnbufferedReportContrlBlock) {
						generateURCBReport(reportNode, rc, (UnbufferedReportContrlBlock) rcb);
					}
					else if (rcb instanceof BufferedReportControlBlock) {
						// TODO add to buffered report queue
					}
				}
			}

			// TODO do the same thing for logs

		}

	}

	private void generateURCBReport(ModelNode node, ReasonCode rc, UnbufferedReportContrlBlock urcb) {
		System.out.println("UnbufferedReport is going to be generated");

		// byte[] owner = ((OCTET_STRING_64) urcb.getChild("Owner")).getValue();
		byte[] owner = urcb.getOwner();
		SCSMConnectionHandler connectionHandler = null;

		for (Association association : associations) {
			// TODO
			// if (association.getIPAddress() == owner) {
			connectionHandler = association.getConnectionHandler();
			// }
		}

		if (connectionHandler != null) {

			OptFields optFlds = (OptFields) urcb.getChild("OptFlds");

			Report report = new Report();

			report.setRptId(((VISIBLE_STRING_129) urcb.getChild("RptID")).getValue());
			report.setOptFlds(optFlds);

			if (optFlds.isSeqNum()) { // SeqNum determined here
				report.setSqNum(reportSqNum);
				reportSqNum++;
				// TODO subSqNum and MoreSegmentsFollow
			}

			if (optFlds.isDataSet()) {
				// report.setDataSet(urcb.getChild("DatSet").getReference()
				// .toString());
				report.setDataSet(urcb.getDataSet().getReference().toString());
			}

			if (optFlds.isConfigRef()) {
				report.setConvRev(((INT64) urcb.getChild("ConfRev")).getValue());
			}

			if (optFlds.isTimeStamp()) {
				report.setTimeOfEntry(new Timestamp(null, null, "", new Date(), false, false));
			}

			ReportEntryData rptEntryData = new ReportEntryData();

			if (optFlds.isDataRef()) {
				rptEntryData.setDataRef(node.getReference().toString());
			}

			// rptEntryData.setValue(((BasicDataAttribute) node).getValue());
			rptEntryData.setValue(node);

			if (optFlds.isReasonCode()) {
				rptEntryData.setReasonCode(rc);
			}

			report.addEntryData(rptEntryData);

			byte[] inclusionBitString = new byte[1];
			inclusionBitString[0] = 0x01;
			report.setInclusionBitString(inclusionBitString);

			try {
				connectionHandler.sendReport(report);
			} catch (IOException e) {
				logger.warn("generateURCBReport got an IOException");
			}

			// increment the sequence number after report has been sent

			INT8U sqNum = (INT8U) urcb.getChild("SqNum");

			if (sqNum.getValue().intValue() < 256) {
				sqNum.setValue(new Byte((byte) (sqNum.getValue() + 1)));
			}

			else {
				sqNum.setValue(new Byte((byte) 0));
			}
		}

	}

	/**
	 * TODO update logs as well
	 */

	private void populateTriggers() {
		// initial update
		for (ModelNode ld : serverModel.getChildren()) {
			for (ModelNode ln : ld.getChildren()) {
				for (ReportControlBlock rcb : ((LogicalNode) ln).getReportControlBlocks()) {
					DataSet ds = (DataSet) rcb.getDataSet();

					TriggerConditions trgOps = (TriggerConditions) rcb.getChild("TrgOps");

					if (ds != null) {
						if (dsRCB.get(ds) == null) {
							dsRCB.put(ds, new ArrayList<ReportControlBlock>());
						}
						dsRCB.get(ds).add(rcb);

						for (BasicDataAttribute mem : ds.getBasicDataAttributes()) {

							if (mem.getDchg() && trgOps.isDataChange()) {
								if (triggerDchg.get(mem.getReference().toString()) == null) {
									triggerDchg.put(mem.getReference().toString(), new ArrayList<DataSet>());
								}

								triggerDchg.get(mem.getReference().toString()).add(ds);
							}

							else if (mem.getQchg() && trgOps.isQualityChange()) {
								if (triggerQchg.get(mem.getReference().toString()) == null) {
									triggerQchg.put(mem.getReference().toString(), new ArrayList<DataSet>());
								}

								triggerQchg.get(mem.getReference().toString()).add(ds);
							}

							else if (mem.getDupd() && trgOps.isDataUpdate()) {
								if (triggerDupd.get(mem.getReference().toString()) == null) {
									triggerDupd.put(mem.getReference().toString(), new ArrayList<DataSet>());
								}

								triggerDupd.get(mem.getReference().toString()).add(ds);
							}

						}
					}

				}
			}
		}

	}

	/**
	 * This method should be called when reports/logs point to a new data set so
	 * that the members to be monitored are updated TODO implement the update
	 * for logs
	 * 
	 */

	private void updateTriggers(DataSet oldDS, DataSet newDS, ReportControlBlock rcb) {

		// remove old data set reference
		for (ReportControlBlock reportCtrl : dsRCB.get(oldDS)) {
			if (reportCtrl.getReference().toString().equals(rcb.getReference().toString())) {
				dsRCB.get(oldDS).remove(reportCtrl);
				break;
			}
		}

		// add new data set reference
		if (dsRCB.get(newDS) == null) {
			dsRCB.put(newDS, new ArrayList<ReportControlBlock>());
		}

		dsRCB.get(newDS).add(rcb);

		// remove members of old data set reference from trigger
		for (BasicDataAttribute mem : oldDS.getBasicDataAttributes()) {
			if (mem.getDchg()) {
				for (DataSet datSet : triggerDchg.get(mem.getReference().toString())) {
					if (datSet.getReference().toString().equals(oldDS.getReference().toString())) {
						triggerDchg.remove(datSet);
						break;
					}
				}
			}

			else if (mem.getQchg()) {
				for (DataSet datSet : triggerQchg.get(mem.getReference().toString())) {
					if (datSet.getReference().toString().equals(oldDS.getReference().toString())) {
						triggerQchg.remove(datSet);
						break;
					}
				}
			}

			else if (mem.getDchg()) {
				for (DataSet datSet : triggerDupd.get(mem.getReference().toString())) {
					if (datSet.getReference().toString().equals(oldDS.getReference().toString())) {
						triggerDupd.remove(datSet);
						break;
					}
				}
			}
		}

		// add members of new data set reference

		for (BasicDataAttribute mem : newDS.getBasicDataAttributes()) {
			TriggerConditions trgOps = (TriggerConditions) rcb.getChild("TrgOps");

			String memRef = mem.getReference().toString();

			if (mem.getDchg() && trgOps.isDataChange()) {
				List<DataSet> datSets = triggerDchg.get(memRef);

				if (datSets == null) {
					triggerDchg.put(memRef, new ArrayList<DataSet>());
					triggerDchg.get(memRef).add(newDS);
					return;

				}

				for (DataSet datSet : datSets) {
					if (datSet.getReference().toString().equals(newDS.getReference().toString())) {
						return; // already added
					}
				}

				// otherwise not added yet
				triggerDchg.get(memRef).add(newDS);
			}

			else if (mem.getQchg() && trgOps.isQualityChange()) {
				List<DataSet> datSets = triggerQchg.get(memRef);

				if (datSets == null) {
					triggerQchg.put(memRef, new ArrayList<DataSet>());
					triggerQchg.get(memRef).add(newDS);
					return;

				}

				for (DataSet datSet : datSets) {
					if (datSet.getReference().toString().equals(newDS.getReference().toString())) {
						return; // already added
					}
				}

				// otherwise not added yet
				triggerQchg.get(memRef).add(newDS);
			}

			else if (mem.getDupd() && trgOps.isDataUpdate()) {
				List<DataSet> datSets = triggerDupd.get(memRef);

				if (datSets == null) {
					triggerDupd.put(memRef, new ArrayList<DataSet>());
					triggerDupd.get(memRef).add(newDS);
					return;

				}

				for (DataSet datSet : datSets) {
					if (datSet.getReference().toString().equals(newDS.getReference().toString())) {
						return; // already added
					}
				}

				// otherwise not added yet
				triggerDupd.get(memRef).add(newDS);
			}
		}
	}
}
