package org.openiec61850.server.data;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.openiec61850.common.ServiceError;
import org.openiec61850.common.model.BasicDataAttribute;
import org.openiec61850.common.model.FunctionalConstraint;
import org.openiec61850.common.model.ModelNode;
import org.openiec61850.server.AccessPoint;
import org.openiec61850.common.model.report.ReportEntryData.ReasonCode;;

/**
 * The IEC server AccessPoint reads and writes data from and to so called
 * DataSources. These data sources are external to the openIEC61850 library
 * stack and are accessed via this interface. For every AccessPoint there must
 * be an implementation of this interface. These implementations are not part of
 * the openIEC61850 stack but of the concrete applications that make use of the
 * stack.
 * 
 * @author C. Do
 */
public class SampleDataSource implements DataSource, Runnable {

	private AccessPoint ac;
	private Map<ReasonCode, List<ModelNode>> notifications = Collections
			.synchronizedMap(new LinkedHashMap<ReasonCode, List<ModelNode>>()); // replace
																				// the
																				// string
																				// with
																				// notifications
	private Set<String> triggerDchg = null;
	private Set<String> triggerQchg = null;
	private Set<String> triggerDupd = null;

	public SampleDataSource() {
		// this.ac = ac;
	}

	public void initialize(AccessPoint ac, Set<String> triggerDchg, Set<String> triggerQchg, Set<String> triggerDupd) {
		this.ac = ac;
		this.triggerDchg = triggerDchg;
		this.triggerQchg = triggerQchg;
		this.triggerDupd = triggerDupd;

		notifications.put(ReasonCode.DCHG, new ArrayList<ModelNode>());
		notifications.put(ReasonCode.QCHG, new ArrayList<ModelNode>());
		notifications.put(ReasonCode.DUPD, new ArrayList<ModelNode>());

		// determined in AccessPoint?
		// notifications.put(ReasonCode.INTEGRITY, new ArrayList<ModelNode>());
		// notifications.put(ReasonCode.GI, new ArrayList<ModelNode>());
	}

	/**
	 * Inserts the proper values in the given BasicDataAttributes using
	 * setValue(). This function is called by the openIEC61850 server library
	 * whenever a GetDataValues or GetDataSetValues service request comes in.
	 * 
	 */
	public void readValues(List<BasicDataAttribute> basicDataAttributes)
			throws ServiceError {

	}

	/**
	 * Takes the values from the given list of BasicDataAttributes.
	 * 
	 */
	public void writeValues(List<BasicDataAttribute> basicDataAttributes)
			throws ServiceError {
		
		for(BasicDataAttribute bda : basicDataAttributes){
			if(bda.getFunctionalConstraint() == FunctionalConstraint.MX){
				addNotification(bda, ReasonCode.DCHG);
			}
		}
	}

	/**
	 * Keeps track of events for reporting and logging. Notifies the server via
	 * the access point when necessary
	 */
	public void run() {
		while(true){
		synchronized (notifications) {
			if (!notifications.isEmpty()) {
				// TODO for all notifications:
				// ac.notificationIndication(notification);
				if (!notifications.get(ReasonCode.DCHG).isEmpty()) {
					for (ModelNode node : notifications.get(ReasonCode.DCHG)) {
						ac.notificationIndication(node, ReasonCode.DCHG);						
					}
					notifications.get(ReasonCode.DCHG).clear();
				}

				if (!notifications.get(ReasonCode.QCHG).isEmpty()) {
					for (ModelNode node : notifications.get(ReasonCode.QCHG)) {
						ac.notificationIndication(node, ReasonCode.QCHG);
					}
					notifications.get(ReasonCode.QCHG).clear();
				}

				if (!notifications.get(ReasonCode.DUPD).isEmpty()) {
					for (ModelNode node : notifications.get(ReasonCode.DUPD)) {
						ac.notificationIndication(node, ReasonCode.DUPD);
					}
					notifications.get(ReasonCode.DUPD).clear();
				}

			}
		}
		}

	}

	private void addNotification(ModelNode node, ReasonCode rc) {

		if (rc == ReasonCode.DCHG || rc == ReasonCode.DUPD
				|| rc == ReasonCode.QCHG) {

			synchronized (notifications) {
				// TODO check if it is a valid notification depending on data
				// set members
				if (((BasicDataAttribute) node).getDchg()) {
					for (String mem : triggerDchg) {
						if (mem.equals(node.getReference().toString())) {
							notifications.get(rc).add(node);
						}
					}
				}
				
				else if (((BasicDataAttribute) node).getQchg()) {
					for (String mem : triggerQchg) {
						if (mem.equals(node.getReference().toString())) {
							notifications.get(rc).add(node);
						}
					}
				}
				
				else if (((BasicDataAttribute) node).getDupd()) {
					for (String mem : triggerDupd) {
						if (mem.equals(node.getReference().toString())) {
							notifications.get(rc).add(node);
						}
					}
				}
			}
		}

	}



}