/*
 * Copyright Fraunhofer ISE, energy & meteo Systems GmbH, and other contributors 2011
 *
 * This file is part of openIEC61850.
 * For more information visit http://www.openmuc.org 
 *
 * openIEC61850 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * openIEC61850 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with openIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openiec61850.server.scl;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.openiec61850.common.model.Array;
import org.openiec61850.common.model.ConstructedDataAttribute;
import org.openiec61850.common.model.DataObject;
import org.openiec61850.common.model.DataSet;
import org.openiec61850.common.model.FunctionalConstraint;
import org.openiec61850.common.model.LogicalDevice;
import org.openiec61850.common.model.LogicalNode;
import org.openiec61850.common.model.ModelNode;
import org.openiec61850.common.model.ObjectReference;
import org.openiec61850.common.model.Server;
import org.openiec61850.common.model.datypes.DAType;
import org.openiec61850.common.model.datypes.basic.BOOLEAN;
import org.openiec61850.common.model.datypes.basic.CodedEnum;
import org.openiec61850.common.model.datypes.basic.Enumerated;
import org.openiec61850.common.model.datypes.basic.FLOAT32;
import org.openiec61850.common.model.datypes.basic.INT16;
import org.openiec61850.common.model.datypes.basic.INT16U;
import org.openiec61850.common.model.datypes.basic.INT32;
import org.openiec61850.common.model.datypes.basic.INT32U;
import org.openiec61850.common.model.datypes.basic.INT64;
import org.openiec61850.common.model.datypes.basic.INT8;
import org.openiec61850.common.model.datypes.basic.INT8U;
import org.openiec61850.common.model.datypes.basic.OCTET_STRING_6;
import org.openiec61850.common.model.datypes.basic.OCTET_STRING_64;
import org.openiec61850.common.model.datypes.basic.UNICODE_STRING_255;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_129;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_255;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_32;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_64;
import org.openiec61850.common.model.datypes.commonacsi.Quality;
import org.openiec61850.common.model.datypes.commonacsi.Timestamp;
import org.openiec61850.common.model.report.BufferedReportControlBlock;
import org.openiec61850.common.model.report.OptFields;
import org.openiec61850.common.model.report.ReportControlBlock;
import org.openiec61850.common.model.report.TriggerConditions;
import org.openiec61850.common.model.report.UnbufferedReportContrlBlock;
import org.openiec61850.server.AccessPoint;
import org.openiec61850.server.IED;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * 
 * @author mzillgit
 * @author sfeuerhahn
 * 
 */

public class ICDParser {

	private TypeDefinitions typeDefinitions;

	private String iedName;
	private IED ied;

	public ICDParser() {
	}

	public IED getIED() {
		return ied;
	}

	public void parse(String icdFile) throws SCLParserException {

		ied = null;

		Vector<AccessPoint> accessPoints = null;

		typeDefinitions = new TypeDefinitions();
		String uri = "file:" + new File(icdFile).getAbsolutePath();

		int lnTypeCount = 0;
		int doTypeCount = 0;
		int daTypeCount = 0;
		int enumTypeCount = 0;

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder;
		try {
			builder = factory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			throw new SCLParserException(e);
		}

		Document doc;
		try {
			doc = builder.parse(uri);
		} catch (SAXException e) {
			throw new SCLParserException(e);
		} catch (IOException e) {
			throw new SCLParserException(e);
		}

		// read in DataTypeTemplates
		NodeList dttSections = doc.getElementsByTagName("DataTypeTemplates");

		if (dttSections.getLength() != 1) {
			throw new SCLParserException("Only one DataTypeSection allowed");
		}

		Node dtt = dttSections.item(0);

		NodeList dataTypes = dtt.getChildNodes();

		for (int i = 0; i < dataTypes.getLength(); i++) {
			Node element = dataTypes.item(i);

			String nodeName = element.getNodeName();

			if (nodeName.equals("LNodeType")) {
				typeDefinitions.putLNodeType(new LNodeType(element));

				lnTypeCount++;
			}
			else if (nodeName.equals("DOType")) {
				typeDefinitions.putDOType(new DOType(element));
				doTypeCount++;
			}
			else if (nodeName.equals("DAType")) {
				typeDefinitions.putDAType(new org.openiec61850.server.scl.DAType(element));
				daTypeCount++;
			}
			else if (nodeName.equals("EnumType")) {
				typeDefinitions.putEnumType(new EnumType(element));
				enumTypeCount++;
			}
		}

		// read in the IED section
		NodeList iedList = doc.getElementsByTagName("IED");
		if (iedList == null || iedList.getLength() == 0) {
			throw new SCLParserException("No IED section found!");
		}
		this.iedName = parseIEDName(iedList.item(0));

		// Create access points
		NodeList aps = doc.getElementsByTagName("AccessPoint");
		if (accessPoints == null) {
			accessPoints = new Vector<AccessPoint>();
		}
		for (int i = 0; i < aps.getLength(); i++) {
			accessPoints.add(createAccessPoint(aps.item(i)));
		}

		ied = new IED(accessPoints, null);

	}

	private AccessPoint createAccessPoint(Node iedServer) throws SCLParserException {
		AccessPoint ap = null;

		NodeList elements = iedServer.getChildNodes();

		/* Create data models */
		for (int i = 0; i < elements.getLength(); i++) {
			Node element = elements.item(i);

			if (element.getNodeName().equals("Server")) {
				if (ap != null) {
					throw new SCLParserException("AccessPoint has more then one server!");
				}
				Server server = createDataModel(element);

				Node namedItem = iedServer.getAttributes().getNamedItem("name");
				if (namedItem == null) {
					throw new SCLParserException("AccessPoint has no name attribute!");
				}
				String name = namedItem.getNodeValue();
				// TODO add ServicesSupported
				ap = new AccessPoint(name, server, null);
			}
		}

		if (ap == null) {
			throw new SCLParserException("AccessPoint has no server!");
		}

		return ap;
	} /* createAccessPoint() */

	private Server createDataModel(Node serverXMLNode) throws SCLParserException {

		NodeList elements = serverXMLNode.getChildNodes();
		List<LogicalDevice> logicalDevices = new ArrayList<LogicalDevice>(elements.getLength());

		for (int i = 0; i < elements.getLength(); i++) {
			Node element = elements.item(i);

			if (element.getNodeName().equals("LDevice")) {
				logicalDevices.add(createNewLDevice(element));
			}
			else if (element.getNodeName().equals("Authentication")) {
				// TODO Authentication
				// System.out.println("TODO: Authentication");
			}
		}

		return new Server(logicalDevices);
	} /* createDataModel() */

	private LogicalDevice createNewLDevice(Node xmlNode) throws SCLParserException {

		/* Parse attributes */

		String inst = null;
		String ldName = null;

		NamedNodeMap attributes = xmlNode.getAttributes();
		Node node;

		if (attributes != null) {
			String nodeName;

			for (int i = 0; i < attributes.getLength(); i++) {
				node = attributes.item(i);
				nodeName = node.getNodeName();

				if (nodeName.equals("inst")) {
					inst = node.getNodeValue();
					// System.out.println("inst = " + inst);
				}
				else if (nodeName.equals("ldName")) {
					ldName = node.getNodeValue();
					// System.out.println("ldName = " + ldName);
				}
			}
		} // else System.out.println("createNewLDevice(): attributes = NULL!");

		if (inst == null) {
			throw new SCLParserException("Required attribute \"inst\" not found!");
		}

		/* Parse Logical Node objects */
		NodeList elements = xmlNode.getChildNodes();
		Node element;

		List<LogicalNode> logicalNodes = new ArrayList<LogicalNode>();

		String ref;
		if ((ldName != null) && (ldName.length() != 0)) {
			ref = ldName;
		}
		else {
			ref = this.iedName + inst;
		}

		for (int i = 0; i < elements.getLength(); i++) {
			element = elements.item(i);

			if (element.getNodeName().equals("LN")) {
				// System.out.println("LN:");
				logicalNodes.add(createNewLogicalNode(element, ref));
			}
			else if (element.getNodeName().equals("LN0")) {
				// System.out.println("LN0:");
				logicalNodes.add(createNewLogicalNode(element, ref));
			}
			// else throw new SCLParserException("Invalid tag name: " +
			// element.getNodeName());
		}

		LogicalDevice ldevice = new LogicalDevice(new ObjectReference(ref), logicalNodes);

		return ldevice;
	} /* createNewLDevice() */

	private LogicalNode createNewLogicalNode(Node lnXMLNode, String parentRef) throws SCLParserException {

		/* Parse attributes */

		String inst = null;
		String lnClass = null;
		String lnType = null;

		// TODO desc useful?
		String desc = null;
		String prefix = "";

		NamedNodeMap attributes = lnXMLNode.getAttributes();
		Node node;
		if (attributes != null) {
			String nodeName;

			for (int i = 0; i < attributes.getLength(); i++) {
				node = attributes.item(i);
				nodeName = node.getNodeName();

				if (nodeName.equals("inst")) {
					inst = node.getNodeValue();
				}
				else if (nodeName.equals("lnType")) {
					lnType = node.getNodeValue();
				}
				else if (nodeName.equals("lnClass")) {
					lnClass = node.getNodeValue();
				}
				else if (nodeName.equals("desc")) {
					desc = node.getNodeValue();
				}
				else if (nodeName.equals("prefix")) {
					prefix = node.getNodeValue();
				}
			}
		}

		if (inst == null) {
			throw new SCLParserException("Required attribute \"inst\" not found!");
		}
		if (lnType == null) {
			throw new SCLParserException("Required attribute \"lnType\" not found!");
		}
		if (lnClass == null) {
			throw new SCLParserException("Required attribute \"lnClass\" not found!");
		}

		String lnName = prefix + lnClass + inst;
		String ref = parentRef + '/' + lnName;

		LNodeType lnTypeDef = typeDefinitions.getLNodeType(lnType);

		List<DataObject> dataObjects = new ArrayList<DataObject>();

		if (lnTypeDef == null) {
			throw new SCLParserException("Type " + lnType + " not defined!");
		}
		for (DO dobject : lnTypeDef.getDataObjects()) {

			// look for DOI node with the name of the DO
			Node doiNodeFound = null;
			for (int i = 0; i < lnXMLNode.getChildNodes().getLength(); i++) {
				Node childNode = lnXMLNode.getChildNodes().item(i);
				if ("DOI".equals(childNode.getNodeName())) {

					NamedNodeMap doiAttributes = childNode.getAttributes();
					Node nameAttribute = doiAttributes.getNamedItem("name");
					if (nameAttribute != null && nameAttribute.getNodeValue() == dobject.getName()) {
						doiNodeFound = childNode;
					}
				}
			}

			DataObject newDataObject = createDataObject(dobject.getName(), ref, dobject.getType(), doiNodeFound);
			dataObjects.add(newDataObject);

			// fcDataObjects.add(dataObject.getFcDataObjects();
		}

		LogicalNode lnNode = new LogicalNode(new ObjectReference(ref), dataObjects, null, null);
		// this creates the fcDataObjects list

		/* Add additional elements and parse data instance related information */
		for (int i = 0; i < lnXMLNode.getChildNodes().getLength(); i++) {
			Node childNode = lnXMLNode.getChildNodes().item(i);
			if ("DataSet".equals(childNode.getNodeName())) {

				lnNode.addDataSet(createDataSet(lnNode, childNode));

			}
			else if ("ReportControl".equals(childNode.getNodeName())) {
				lnNode.addReportControlBlock(createReportControl(lnNode, childNode));
			}
			else if ("LogControl".equals(childNode.getNodeName())) {
				// TODO
				// lnode.addLogControlBlock(createLogControlBlock(lnode,
				// childNode));
			}

		}

		// return new LogicalNode(new ObjectReference(ref), dataObjects, null,
		// null);
		return lnNode;

	}

	// private void parseSDI(Node xmlNode, ModelNode parentNode) throws
	// SCLParserException {
	// String name = null;
	//
	// NamedNodeMap attributes = xmlNode.getAttributes();
	// Node node;
	// if (attributes != null) {
	// String nodeName;
	//
	// for (int i = 0; i < attributes.getLength(); i++) {
	// node = attributes.item(i);
	// nodeName = node.getNodeName();
	//
	// if (nodeName.equals("name")) {
	// name = node.getNodeValue();
	// // } else if (nodeName.equals("ix")) {
	// // String ix = node.getNodeValue();
	// // TODO support arrays
	// }
	// }
	// }
	//
	// /* Check for required attributes */
	// if (name == null) {
	// throw new SCLParserException("SDI requires \"name\" attribute!");
	// }
	//
	// ModelNode currentNode = parentNode.getChild(name);
	// if ((currentNode == null)) {
	// throw new SCLParserException("SAI is not allowed for \"" +
	// parentNode.getReference() + "." + name + "\"!");
	// }
	//
	// // String sdiRef = parRef + "." + name;
	//
	// /* parse child SDIs and DAIs */
	// for (int k = 0; k < xmlNode.getChildNodes().getLength(); k++) {
	// Node attrNode = xmlNode.getChildNodes().item(k);
	//
	// if (attrNode.getNodeName().equals("SDI")) {
	// parseSDI(attrNode, currentNode);
	// }
	// else if (attrNode.getNodeName().equals("DAI")) {
	// parseDAI(attrNode, currentNode);
	// }
	// }
	//
	// } /* parseSDI() */

	// private void parseDAI(Node xmlNode, ModelNode parentNode) throws
	// SCLParserException {
	// String name = null;
	// String sAddr = null;
	// String ix = null;
	//
	// NamedNodeMap attributes = xmlNode.getAttributes();
	// Node node;
	// if (attributes != null) {
	// String nodeName;
	//
	// for (int i = 0; i < attributes.getLength(); i++) {
	// node = attributes.item(i);
	// nodeName = node.getNodeName();
	//
	// if (nodeName.equals("name")) {
	// name = node.getNodeValue();
	// }
	// else if (nodeName.equals("sAddr")) {
	// sAddr = node.getNodeValue();
	// // } else if (nodeName.equals("valKind")) {
	// // valKind = node.getNodeValue();
	// // } else if (nodeName.equals("desc")) {
	// // desc = node.getNodeValue();
	// }
	// else if (nodeName.equals("ix")) {
	// ix = node.getNodeValue();
	// }
	// }
	// }
	//
	// /* Check for required attributes */
	// if (name == null) {
	// throw new SCLParserException("DAI requires \"name\" attribute!");
	// }
	//
	// // String daiRef = lnode.getReference() + "." + doName + "." + name;
	//
	// ModelNode currentNode = parentNode.getChild(name);
	// if ((currentNode == null) || (currentNode.hasChildren())) {
	// throw new SCLParserException("DAI is not allowed for \"" +
	// parentNode.getReference() + "\"!");
	// }
	//
	// ConstructedDataAttribute da;
	// if (currentNode.getObjectType() == ObjectType.ARRAY_OF_DA) {
	// Array array = (Array) currentNode;
	// int index = Integer.parseInt(ix);
	// if (ix == null || array.get(index) == null) {
	// throw new SCLParserException("DAI array index \"ix\" out of bound: " +
	// parentNode.getReference() + '['
	// + index + ']');
	// }
	// da = (ConstructedDataAttribute) array.get(index);
	// }
	// else {
	// da = (ConstructedDataAttribute) currentNode;
	// }
	//
	// if (sAddr != null) {
	// da.getValue().setsAddr(sAddr);
	// // System.out.println("Short address: " + sAddr + " for " + daiRef);
	// }
	// } /* parseDAI() */
	//
	private DataSet createDataSet(LogicalNode lnNode, Node xmlNode) throws SCLParserException {

		Node nameAttribute = xmlNode.getAttributes().getNamedItem("name");
		if (nameAttribute == null) {
			throw new SCLParserException("DataSet must have a name");
		}

		String name = nameAttribute.getNodeValue();

		List<ModelNode> members = new ArrayList<ModelNode>();

		for (int i = 0; i < xmlNode.getChildNodes().getLength(); i++) {
			Node childNode = xmlNode.getChildNodes().item(i);
			if ("FCDA".equals(childNode.getNodeName())) {

				Node doNameAttribute = childNode.getAttributes().getNamedItem("doName");
				Node fcAttribute = childNode.getAttributes().getNamedItem("fc");
				if (fcAttribute == null) {
					throw new SCLParserException("FCDA must have an fc");
				}
				FunctionalConstraint fc = FunctionalConstraint.fromString(fcAttribute.getNodeValue());

				if (doNameAttribute != null) {
					String doName = doNameAttribute.getNodeValue();

					if (!doName.isEmpty()) {
						ModelNode dataObject = lnNode.getChild(doName, fc);
						if (dataObject == null) {
							throw new SCLParserException("LogicalNode " + lnNode.getNodeName()
									+ " does not have a DataObject " + doName + " for DataSet " + name);
						}
						members.add(dataObject);
					}

					else {
						// add only the functionally constrained data objects if
						// doName is empty?
						List<DataObject> fcDataObjects = lnNode.getChildren(fc);
						if (fcDataObjects == null) {
							throw new SCLParserException("LogicalNode " + lnNode.getNodeName()
									+ " does not have children with functional constraint " + fc + " for DataSet "
									+ name);
						}

						for (DataObject dataObj : fcDataObjects) {
							members.add(dataObj);
						}

					}
				}
				// check if daName exists
				else {
					Node daNameAttribute = childNode.getAttributes().getNamedItem("daName");
					String daName = daNameAttribute.getNodeValue();

					List<DataObject> fcDataObjects = lnNode.getChildren(fc);

					if (fcDataObjects == null) {
						throw new SCLParserException("LogicalNode " + lnNode.getNodeName()
								+ " does not have children with functional constraint " + fc + " for DataSet " + name);
					}

					if (daNameAttribute == null || daName.isEmpty()) {

						for (DataObject dataObj : fcDataObjects) {
							members.addAll(dataObj.getBasicDataAttributes());
						}
					}

					else {
						// add all the data attributes with the same name and
						// functional constraint
						for (DataObject dataObj : fcDataObjects) {
							members.add(dataObj.getChild(daName));
						}

					}

				}

			}

		}

		DataSet dataSet = new DataSet(new ObjectReference(lnNode.getReference().toString() + '.' + name), members,
				false);
		return dataSet;
	}

	private ReportControlBlock createReportControl(LogicalNode lnode, Node xmlNode) throws SCLParserException {
		// ReportControlBlock rcb;
//		FunctionalConstraint fc;
		
		boolean bufferedReport;

		if (getAttribute(xmlNode, "buffered", null) != null) {
//			fc = FunctionalConstraint.BR;
			bufferedReport = true;
		}
		else {
//			fc = FunctionalConstraint.RP;
			bufferedReport = false;
		}
		TriggerConditions trigOps = null;
		OptFields optFields = null;
		boolean rptEna = false;

		ObjectReference reportObjRef = new ObjectReference(lnode.getReference().toString() + "."
				+ getAttribute(xmlNode, "name", "ReportControl must have a name"));
		List<ModelNode> children = new ArrayList<ModelNode>();

		for (int i = 0; i < xmlNode.getChildNodes().getLength(); i++) {
			Node childNode = xmlNode.getChildNodes().item(i);
			if ("TrgOps".equals(childNode.getNodeName())) {
				// rcb.setTrgOps(createTriggerConditions(childNode));
				trigOps = createTriggerConditions(childNode, lnode);
			}
			else if ("OptFields".equals(childNode.getNodeName())) {
				// rcb.setOptFlds(createOptFields(childNode));
				optFields = createOptFields(childNode, lnode);
			}
			else if ("RptEnabled".equals(childNode.getNodeName())) {
				// rcb.setRptEna(true);
				rptEna = false;
				// TODO there may be ClientLN fields here that describe for
				// which
				// clients
				// buffered reports shall buffer data. This is not implemented
				// yet
			}
		}
		// rcb.setName(getAttribute(xmlNode, "name",
		// "ReportControl must have a name"));

		children.add(new VISIBLE_STRING_129(new ObjectReference(reportObjRef
				.toString()
				+ ".RptID"), null, "", getAttribute(xmlNode, "rptID",
				"ReportControl must have a rptID"), false, false));
		children.add(new BOOLEAN(new ObjectReference(reportObjRef.toString()
				+ ".RptEna"), null, "", rptEna, false, false));
		
		if(!bufferedReport){		
		children.add(new BOOLEAN(new ObjectReference(reportObjRef.toString()
				+ ".Resv"), null, "", false, false, false));
		}

		String nodeValue = getAttribute(xmlNode, "datSet", "ReportControl must have attribute datSet");
		DataSet dataSet = lnode.getDataSet(nodeValue);
		if (dataSet == null) {
			throw new SCLParserException("DataSet " + nodeValue + " does not exist for ReportControl");
		}
		
		children.add(new VISIBLE_STRING_255(new ObjectReference(reportObjRef.toString() + ".DatSet"), null, "", dataSet.getReference().toString(), false, false));
		children.add(new INT64(new ObjectReference(reportObjRef.toString() + ".ConfRev"), null, "", getLongAttribute(
				xmlNode, "confRev", "ReportControl must have a convRef"), false, false));
		children.add(optFields);

		//in iec 61850-7-2 says it is of type INT32U for BufTm
		children.add(new INT32U(new ObjectReference(reportObjRef.toString()	+ ".BufTm"), null, "", new Integer((int)getLongAttribute(xmlNode, "bufTime", null)), false, false));
		children.add(new INT8U(new ObjectReference(reportObjRef.toString()	+ ".SqNum"), null, "", new Byte((byte)0), false, false));
		children.add(trigOps);
		children.add(new INT64(new ObjectReference(reportObjRef.toString() + ".IntgPd"), null, "", getLongAttribute(
				xmlNode, "intgPd", null), false, false));
		children.add(new BOOLEAN(new ObjectReference(reportObjRef.toString() + ".GI"), null, "", false, false, false));

		// rcb.setDataSet(dataSet);
		// rcb.setIntgPd(getLongAttribute(xmlNode, "intgPd", null));
		// rcb.setRptID(getAttribute(xmlNode, "rptID",
		// "ReportControl must have a rptID"));
		// rcb.setConvRef(getLongAttribute(xmlNode, "confRev",
		// "ReportControl must have a convRef"));
		// rcb.setBufTm(getLongAttribute(xmlNode, "bufTime", null));

		if (bufferedReport) {

			children.add(new BOOLEAN(new ObjectReference(reportObjRef.toString()
				+ ".PurgeBuf"), null, "", false, false, false));	
	
			children.add(new OCTET_STRING_64(new ObjectReference(reportObjRef.toString()
				+ ".EntryID"), null, "", new byte[8] , false, false));
			
			children.add(new OCTET_STRING_6(new ObjectReference(reportObjRef.toString()
				+ ".TimeOfEntry"), null, "", new byte[6], false, false));			
			
			//these are optional
//			children.add(new INT16(new ObjectReference(reportObjRef.toString()+".ResvTms"), fc, "", new Short((short) 0), false, false));
//			
//			children.add(new OCTET_STRING_64(new ObjectReference(reportObjRef.toString()
//					+ ".Owner"), null, "", new byte[0], false, false));
			BufferedReportControlBlock brcb = new BufferedReportControlBlock(reportObjRef, children);
			brcb.setDataSet(dataSet);			
			return brcb; 
		} else {
			//this is optional
//			children.add(new OCTET_STRING_64(new ObjectReference(reportObjRef.toString()
//					+ ".Owner"), null, "", new byte[0] , false, false));
			UnbufferedReportContrlBlock urcb = new UnbufferedReportContrlBlock(reportObjRef, children);
			urcb.setDataSet(dataSet);
			return urcb;
		}

	}

	private OptFields createOptFields(Node xmlNode, LogicalNode lnode) throws SCLParserException {
		OptFields of = new OptFields(new ObjectReference(lnode.getReference().toString() + ".OptFlds"));
		of.setBufOvfl(getBooleanAttribute(xmlNode, "bufOvfl"));
		of.setConfigRef(getBooleanAttribute(xmlNode, "configRef"));
		of.setDataRef(getBooleanAttribute(xmlNode, "dataRef"));
		of.setDataSet(getBooleanAttribute(xmlNode, "dataSet"));
		of.setEntryId(getBooleanAttribute(xmlNode, "entryID"));
		of.setReasonCode(getBooleanAttribute(xmlNode, "reasonCode"));
		/* segmentation is depreciated and not used */
		// of.setSegmentation(getBooleanAttribute(xmlNode, "segmentation"));
		of.setSeqNum(getBooleanAttribute(xmlNode, "seqNum"));
		of.setTimeStamp(getBooleanAttribute(xmlNode, "timeStamp"));
		return of;
	}

	private TriggerConditions createTriggerConditions(Node xmlNode, LogicalNode lnode) throws SCLParserException {
		TriggerConditions tc = new TriggerConditions(new ObjectReference(lnode.getReference().toString() + ".TrgOps"));
		tc.setDataChange(getBooleanAttribute(xmlNode, "dchg"));
		tc.setDataUpdate(getBooleanAttribute(xmlNode, "dupd"));
		tc.setQualityChange(getBooleanAttribute(xmlNode, "qchg"));
		tc.setIntegrity(getBooleanAttribute(xmlNode, "period"));

		// should be true if the gi field is not present in the SCL
		if (getAttribute(xmlNode, "gi", null) == null) {
			tc.setGeneralInterrogation(true);
		}
		else {
			tc.setGeneralInterrogation(getBooleanAttribute(xmlNode, "gi"));
		}
		return tc;
	}

	// private LogControlBlock createLogControlBlock(LogicalNode lnode,
	// Node xmlNode) throws SCLParserException {
	// // LogControlBlock lcb = new LogControlBlock();
	// // String name = getAttribute(xmlNode, "name",
	// // "LogControl must have a name");
	// // lcb.setName(name);
	// // lcb.setLogRef(new ObjectReference(lnode.getReference() + "." +
	// // name));
	// // String nodeValue = getAttribute(xmlNode, "datSet",
	// // "LogControl must have attribute datSet");
	// // DataSet dataSet = lnode.getDataSet(nodeValue);
	// // if (dataSet == null) {
	// // throw new SCLParserException("DataSet " + nodeValue
	// // + " does not exist for LogControl");
	// // }
	// // lcb.setDataSet(dataSet);
	// // lcb.setIntgPd(getLongAttribute(xmlNode, "intgPd", null));
	// // lcb.setLogEna(getBooleanAttribute(xmlNode, "logEna"));
	// // for (int i = 0; i < xmlNode.getChildNodes().getLength(); i++) {
	// // Node childNode = xmlNode.getChildNodes().item(i);
	// // if ("TrgOps".equals(childNode.getNodeName())) {
	// // lcb.setTrgOps(createTriggerConditions(childNode));
	// // }
	// // }
	// return null;
	// }

	private boolean getBooleanAttribute(Node xmlNode, String name) throws SCLParserException {
		String value = getAttribute(xmlNode, name, null);
		return value != null && "true".equalsIgnoreCase(value);
	}

	private long getLongAttribute(Node xmlNode, String name, String errorMessage) throws SCLParserException {
		String value = getAttribute(xmlNode, name, errorMessage);
		if (value != null) {
			try {
				return Long.parseLong(value);
			} catch (NumberFormatException e) {
				throw new SCLParserException("Wrong number format for node " + xmlNode.getNodeName() + ", attribute "
						+ name + ": " + e.getMessage());
			}
		}
		return 0L;
	}

	private String getAttribute(Node xmlNode, String name, String errorMessage) throws SCLParserException {
		Node dataSetAttribute = xmlNode.getAttributes().getNamedItem(name);
		if (dataSetAttribute == null) {
			if (errorMessage == null) {
				return null;
			}
			else {
				throw new SCLParserException(errorMessage);
			}
		}
		return dataSetAttribute.getNodeValue();
	}

	private DataObject createDataObject(String name, String parentRef, String doTypeID, Node doiNode)
			throws SCLParserException {

		DOType doType = typeDefinitions.getDOType(doTypeID);

		if (doType == null) {
			throw new SCLParserException("DO type " + doType + " not defined!");
		}

		String ref = parentRef + '.' + name;

		List<ModelNode> childNodes = new ArrayList<ModelNode>();

		for (DA dattr : doType.getdAttributes()) {

			// look for DAI node with the name of the DA
			Node iNodeFound = findINode(doiNode, dattr.getName());

			if (dattr.getCount() >= 1) {
				childNodes.add(createArrayOfDataAttributes(ref + '.' + dattr.getName(), dattr, iNodeFound));
			}
			else {
				childNodes.add(createDataAttribute(ref + '.' + dattr.getName(), dattr.getFc(), dattr, iNodeFound));
			}

		}

		for (SDO sdo : doType.getSdos()) {
			// TODO array of data object
			Node iNodeFound = findINode(doiNode, sdo.getName());

			childNodes.add(createDataObject(sdo.getName(), ref, sdo.getType(), iNodeFound));
		}

		if (doiNode != null) {
			NamedNodeMap doiAttributes = doiNode.getAttributes();
			for (int j = 0; j < doiAttributes.getLength(); j++) {
				String attributeName = doiAttributes.item(j).getNodeName();

				if (attributeName.equals("ix")) {
					// TODO parse "ix"
				}
				// TODO parse "accessControl"
			}
		}

		// ModelNode currentNode = lnode.getChild(doName);
		// if (currentNode == null) {
		// throw new SCLParserException("SDO not allowed for " +
		// lnode.getReference() + "." + doName);
		// }
		//
		// /* parse child SDIs and DAIs */
		// for (int k = 0; k < doiNode.getChildNodes().getLength(); k++) {
		// Node attrNode = doiNode.getChildNodes().item(k);
		//
		// if (attrNode.getNodeName().equals("SDI")) {
		// parseSDI(attrNode, currentNode);
		// }
		// else if (attrNode.getNodeName().equals("DAI")) {
		// parseDAI(attrNode, currentNode);
		// }
		// }

		return new DataObject(new ObjectReference(ref), childNodes, doType.getCdc());
	}

	private Node findINode(Node iNode, String dattrName) {

		if (iNode == null) {
			return null;
		}

		Node iNodeFound = null;
		for (int i = 0; i < iNode.getChildNodes().getLength(); i++) {
			Node childNode = iNode.getChildNodes().item(i);
			Node nameAttribute = childNode.getAttributes().getNamedItem("name");
			if (nameAttribute != null && nameAttribute.getNodeValue() == dattrName) {
				iNodeFound = childNode;
			}
		}
		return iNodeFound;
	}

	private Array createArrayOfDataAttributes(String ref, DA sclDA, Node iXMLNode) throws SCLParserException {

		FunctionalConstraint fc = sclDA.getFc();
		int size = sclDA.getCount();

		List<ModelNode> arrayItems = new ArrayList<ModelNode>();
		for (int i = 0; i < size; i++) {
			// ArrayItem item = createDataAttribute(parentRef, fc, sclDA);
			// ArrayItem item = new DataAttribute(parentRef,
			// sclDA.getName(), fc);
			// newArray.add(item);
			// TODO:
			// if (prototypeItem.getBasicType() == BasicTypeEnum.FLOAT32) {
			// prototypeItem.getValue().setValue(i / 100f); // FIXME just a
			// // test
			// }

			arrayItems.add(createDataAttribute(ref + '[' + i + ']', null, sclDA, iXMLNode));
		}

		return new Array(new ObjectReference(ref), fc, arrayItems);
	}

	// returns ConstructedDataAttribute or BasicDataAttribute
	private ModelNode createDataAttribute(String ref, FunctionalConstraint fc, AbstractDataAttribute dattr,
			Node iXMLNode) throws SCLParserException {

		// String ref = parentRef + "." + dattr.getName();

		/* Check if attribute is a basic type */
		if (Util.isBasicType(dattr.getbType())) {
			// System.out.println(" | BASIC TYPE: " + dattr.getbType());

			String sAddr = null;
			if (iXMLNode != null) {

				NamedNodeMap iAttributes = iXMLNode.getAttributes();

				// FIXME already done in AbstractDataAttribute?
				for (int j = 0; j < iAttributes.getLength(); j++) {
					String attributeName = iAttributes.item(j).getNodeName();

					if (attributeName.equals("sAddr")) {
						sAddr = iAttributes.item(j).getNodeValue();
					}
				}
			}

			DAType basicType = Util.bTypeFromSCLString(dattr.getbType());

			boolean dchg = false;
			boolean dupd = false;
			boolean qchg = false;

			if (dattr instanceof DA) {
				dchg = ((DA) dattr).isDchg();			
				dupd = ((DA) dattr).isDupd();
				qchg = ((DA) dattr).isQchg();
			}

			// dattr can be DA or BDA -> BDA does not have triggering
			// information
			switch (basicType) {
			case BOOLEAN:

				return new BOOLEAN(new ObjectReference(ref), fc, sAddr, false, dchg, dupd);

			case CODEDENUM:

				return new CodedEnum(new ObjectReference(ref), fc, sAddr, new byte[1], dchg, dupd);

			case ENUMERATED:
				return new Enumerated(new ObjectReference(ref), fc, sAddr, dattr.getbType(), new Integer(0), dchg, dupd);

			case FLOAT32:
				return new FLOAT32(new ObjectReference(ref), fc, sAddr, new Float(0), dchg, dupd);

			case INT8:
				return new INT8(new ObjectReference(ref), fc, sAddr, new Byte((byte) 0), dchg, dupd);

			case INT8U:
				return new INT8U(new ObjectReference(ref), fc, sAddr, new Byte((byte) 0), dchg, dupd);

			case INT16:
				return new INT16(new ObjectReference(ref), fc, sAddr, new Short((short) 0), dchg, dupd);

			case INT16U:
				return new INT16U(new ObjectReference(ref), fc, sAddr, new Short((short) 0), dchg, dupd);

			case INT32:
				return new INT32(new ObjectReference(ref), fc, sAddr, new Integer(0), dchg, dupd);

			case INT32U:
				return new INT32U(new ObjectReference(ref), fc, sAddr, new Integer(0), dchg, dupd);

			case INT64:
				return new INT64(new ObjectReference(ref), fc, sAddr, new Long(0), dchg, dupd);

			case OCTET_STRING_64:
				return new OCTET_STRING_64(new ObjectReference(ref), fc, sAddr, new byte[0], dchg, dupd);

			case VISIBLE_STRING_32:
				return new VISIBLE_STRING_32(new ObjectReference(ref), fc, sAddr, "", dchg, dupd);

			case VISIBLE_STRING_64:
				return new VISIBLE_STRING_64(new ObjectReference(ref), fc, sAddr, "", dchg, dupd);

			case VISIBLE_STRING_129:
				return new VISIBLE_STRING_129(new ObjectReference(ref), fc, sAddr, "", dchg, dupd);

			case VISIBLE_STRING_255:
				return new VISIBLE_STRING_255(new ObjectReference(ref), fc, sAddr, "", dchg, dupd);

			case UNICODE_STRING_255:
				return new UNICODE_STRING_255(new ObjectReference(ref), fc, sAddr, "", dchg, dupd);

			case TIMESTAMP:
				return new Timestamp(new ObjectReference(ref), fc, sAddr, new Date(), dchg, dupd);

//			case TCMD:
//				return new Tcmd(new ObjectReference(ref), fc, sAddr, new byte[1], dchg, dupd);

			case QUALITY:
				return new Quality(new ObjectReference(ref), fc, sAddr, qchg);

//			case GENERIC_BITSTRING:
//				return new GENERIC_BITSTRING(new ObjectReference(ref), fc, sAddr, new BitSet(), dchg, dupd);

			default:
				throw new SCLParserException("Basic type " + dattr.getbType() + " not implemented!");
			}

			// return new BasicDataAttribute(new ObjectReference(ref), fc,
			// Util.bTypeFromSCLString(dattr.getbType()),
			// sAddr, null, false, false, false);

		}
		else {
			// System.out.println(" | COMPLEX TYPE: " + dattr.getbType() +
			// " OF TYPE " + dattr.getType());

			if (dattr.getbType().equals("Struct")) {
				org.openiec61850.server.scl.DAType datype = typeDefinitions.getDAType(dattr.getType());

				if (datype == null) {
					throw new SCLParserException("DAType " + dattr.getbType() + " not declared!");
				}

				List<ModelNode> subDataAttributes = new ArrayList<ModelNode>();
				for (BDA bda : datype.getBdas()) {

					Node iNodeFound = findINode(iXMLNode, dattr.getName());

					subDataAttributes.add(createDataAttribute(ref + '.' + bda.getName(), fc, bda, iNodeFound));
				}

				return new ConstructedDataAttribute(new ObjectReference(ref), fc, subDataAttributes);
			}
			// else if (dattr.getbType().equals("Enum")) {
			//
			// String sAddr = null;
			// if (iXMLNode != null) {
			//
			// NamedNodeMap iAttributes = iXMLNode.getAttributes();
			// for (int j = 0; j < iAttributes.getLength(); j++) {
			// String attributeName = iAttributes.item(j).getNodeName();
			//
			// if (attributeName.equals("sAddr")) {
			// sAddr = iAttributes.item(j).getNodeValue();
			// }
			// }
			// }
			//
			// return new BasicDataAttribute(new ObjectReference(ref), fc,
			// BasicType.INT32, sAddr, null, false, false,
			// false);
			//
			// // TODO implement Enums
			// // newDataAttribute = new ConstructedDataAttribute(parentRef,
			// // dattr.getName(), fc);
			// // newDataAttribute.setValue(new
			// // DataAttributeValue(BasicTypeEnum.INT32));
			// // newDataAttribute.getValue().setsAddr(dattr.getsAddr());
			// }
			// else if (dattr.getbType().equals("Dbpos")) {
			// // TODO implement Dbpos
			// // newDataAttribute = new ConstructedDataAttribute(parentRef,
			// // dattr.getName(), fc);
			// }
			else {
				throw new SCLParserException("Basic type " + dattr.getbType() + " not implemented!");
			}

			// return null;
		}

	} /* createDataAttribute() */

	private String parseIEDName(Node iedNode) throws SCLParserException {
		Node nameAttribute = iedNode.getAttributes().getNamedItem("name");
		if (nameAttribute != null) {
			String name = nameAttribute.getNodeValue();
			if ((name != null) && (name.length() != 0)) {
				return name;
			}
		}
		throw new SCLParserException("IED must have a name!");
	}

	private TypeDefinitions getTypeDefinitions() {
		return typeDefinitions;
	}
}
