/*
 * Copyright Fraunhofer ISE, energy & meteo Systems GmbH, and other contributors 2011
 *
 * This file is part of openIEC61850.
 * For more information visit http://www.openmuc.org 
 *
 * openIEC61850 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * openIEC61850 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with openIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openiec61850.scsm.mms;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.openiec61850.common.ServiceError;
import org.openiec61850.common.model.Array;
import org.openiec61850.common.model.BasicDataAttribute;
import org.openiec61850.common.model.ConstructedDataAttribute;
import org.openiec61850.common.model.DataObject;
import org.openiec61850.common.model.DataSet;
import org.openiec61850.common.model.FunctionalConstraint;
import org.openiec61850.common.model.ModelNode;
import org.openiec61850.common.model.datypes.basic.BOOLEAN;
import org.openiec61850.common.model.datypes.basic.CodedEnum;
import org.openiec61850.common.model.datypes.basic.Enumerated;
import org.openiec61850.common.model.datypes.basic.FLOAT32;
import org.openiec61850.common.model.datypes.basic.INT16;
import org.openiec61850.common.model.datypes.basic.INT16U;
import org.openiec61850.common.model.datypes.basic.INT32;
import org.openiec61850.common.model.datypes.basic.INT32U;
import org.openiec61850.common.model.datypes.basic.INT64;
import org.openiec61850.common.model.datypes.basic.INT8;
import org.openiec61850.common.model.datypes.basic.INT8U;
import org.openiec61850.common.model.datypes.basic.OCTET_STRING_6;
import org.openiec61850.common.model.datypes.basic.OCTET_STRING_64;
import org.openiec61850.common.model.datypes.basic.UNICODE_STRING_255;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_129;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_255;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_32;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_64;
import org.openiec61850.common.model.datypes.commonacsi.Quality;
import org.openiec61850.common.model.datypes.commonacsi.Timestamp;
import org.openiec61850.common.model.report.OptFields;
import org.openiec61850.common.model.report.ReportControlBlock;
import org.openiec61850.common.model.report.TriggerConditions;
import org.openiec61850.jmms.mms.asn1.Data;
import org.openiec61850.jmms.mms.asn1.DataSequence;
import org.openmuc.jasn1.ber.types.BerBitString;
import org.openmuc.jasn1.ber.types.BerBoolean;
import org.openmuc.jasn1.ber.types.BerInteger;
import org.openmuc.jasn1.ber.types.BerOctetString;
import org.openmuc.jasn1.ber.types.string.BerUTF8String;
import org.openmuc.jasn1.ber.types.string.BerVisibleString;

public class MmsAsn1ComplexTypeConverter {

	/**
	 * Numbers can be plain numbers like Integer, but some Enumerated data types
	 * are also implemented as numbers. In java they can be implemented using
	 * {@link Enum}s. In this case we take the ordinal value of the Enum value
	 * as the integer value.
	 * 
	 * This does not work in all cases. It is only possible, if the Enumerated
	 * values start with 0 and have no holes.
	 */
	private static Long getNumberValue(BasicDataAttribute da) throws ServiceError {
		Object rawValue = da.getValue();
		Long value;
		if (rawValue == null) {
			value = 0L;
		}
		else if (rawValue instanceof Enum<?>) {
			value = Long.valueOf(((Enum<?>) rawValue).ordinal());
		}
		else if (rawValue instanceof Number) {
			value = Long.valueOf(((Number) rawValue).longValue());
		}
		else {
			throw new ServiceError(ServiceError.TYPE_CONFLICT, "value of " + da.getReference() + " should be "
					+ da.getBasicType() + " but is " + rawValue.getClass());
		}
		return value;
	}

	public static Data convertModelNodeToMmsData(ModelNode node) throws ServiceError {
		// switch (node.getObjectType()) {
		// case BASIC_DATA_ATTRIBUTE:
		if (node instanceof BasicDataAttribute) {
			return convertBasicNodeToMmsData((BasicDataAttribute) node);
		}
		// case ARRAY_OF_FCDO:
		// case ARRAY_OF_DA:
		else if (node instanceof Array) {
			return convertArrayNodeToMmsData((Array) node);
		}
		// case FC_DATA_OBJECT:
		// case CONSTRUCTED_DATA_ATTRIBUTE:
		else if (node instanceof DataObject || node instanceof ConstructedDataAttribute
				|| node instanceof ReportControlBlock) {
			return convertStructNodeToMmsData(node);
		}

		else if (node instanceof TriggerConditions || node instanceof OptFields || node instanceof DataSet) {
			return convertNodeToMmsData(node);
		}

		else {
			throw new IllegalArgumentException("invalid ModelNode for: " + node.getNodeName() + " passed.");
		}
	}

	public static Data convertStructNodeToMmsData(ModelNode modelNode) throws ServiceError {

		Collection<ModelNode> subNodes = modelNode.getChildren();
		// Collections.sort(subNodes,
		// MMSNodeSorter.getDataContainerComparator());

		// Data data = new Data();
		LinkedList<Data> seq = new LinkedList<Data>();
		for (ModelNode item : subNodes) {
			Data child = convertModelNodeToMmsData(item);
			if (child == null) {
				continue;
			}
			seq.add(child);
		}
		if (seq.size() == 0) {
			return null;
		}
		// data.selectStructure(new DataSequence(seq));
		// return data;

		DataSequence dataSeq = new DataSequence(seq);
		return new Data(null, dataSeq, null, null, null, null, null, null, null, null, null, null, null, null, null);
	}

	public static Data convertArrayNodeToMmsData(Array array) throws ServiceError {
		// Data data = new Data();
		// DataSequence mmsArray = new DataSequence();
		// mmsArray.setValue(new ArrayList<Data>(array.size()));
		// for (ModelNode arrayItem : array) {
		// Data mmsArrayItem = convertModelNodeToMmsData(arrayItem);
		// if (mmsArrayItem != null) {
		// mmsArray.add(mmsArrayItem);
		// }
		// }
		// data.selectArray(mmsArray);
		// return data;
		ArrayList<Data> mmsArray = new ArrayList<Data>();
		for (ModelNode arrayItem : array) {
			Data mmsArrayItem = convertModelNodeToMmsData(arrayItem);
			if (mmsArrayItem != null) {
				mmsArray.add(mmsArrayItem);
			}
		}

		DataSequence dataSeq = new DataSequence(mmsArray);
		return new Data(dataSeq, null, null, null, null, null, null, null, null, null, null, null, null, null, null);

	}

	private static Data convertNodeToMmsData(ModelNode node) throws ServiceError {
		// Data data = new Data();

		if (node instanceof TriggerConditions) {
			return new Data(null, null, null, new BerBitString(((TriggerConditions) node).getValue(), 6), null, null,
					null, null, null, null, null, null, null, null, null);
		}

		else if (node instanceof OptFields) {
			// data.selectBit_string(new
			// BitString(MmsAsn1PrimitiveTypeConverter.toByteArray((((OptFields)node).getValue()))));
			// data.selectBit_string(new BitString(((OptFields)
			// node).getValue(),
			// 6));
			return new Data(null, null, null, new BerBitString(((OptFields) node).getValue(), 10), null, null, null,
					null, null, null, null, null, null, null, null);
		}

		// TODO check if data sets may be handled differently
		else {// data set
		// data.selectVisible_string(node.getReference().toString());
			try {
				return new Data(null, null, null, null, null, null, null, null, new BerVisibleString(node
						.getReference().toString()), null, null, null, null, null, null);

			} catch (UnsupportedEncodingException e) {
				throw new IllegalAccessError();
			}
		}

	}

	private static Data convertBasicNodeToMmsData(BasicDataAttribute da) throws ServiceError {

		if (da.getValue() != null) {
			// Data data = null;
			Object daValue = da.getValue();

			switch (da.getBasicType()) {
			case BOOLEAN:
				// data.selectBoolean_((Boolean) daValue);
				return new Data(null, null, new BerBoolean((Boolean) daValue), null, null, null, null, null, null,
						null, null, null, null, null, null);

			case INT8:
			case INT16:
			case INT32:
			case INT64:
				// data.selectInteger(getNumberValue(da));
				return new Data(null, null, null, null, new BerInteger(getNumberValue(da)), null, null, null, null,
						null, null, null, null, null, null);
				// case INT24U:
			case INT8U:
			case INT16U:
			case INT32U:
				// data.selectUnsigned(getNumberValue(da));
				return new Data(null, null, null, null, null, new BerInteger(getNumberValue(da)), null, null, null,
						null, null, null, null, null, null);
			case FLOAT32:
				// data.selectFloating_point(new FloatingPoint(
				// MmsAsn1PrimitiveTypeConverter
				// .toByteArray((Float) daValue)));
				// return data;
				return new Data(null, null, null, null, null, null, new BerOctetString(
						MmsAsn1PrimitiveTypeConverter.toByteArray(((Float) daValue))), null, null, null, null, null,
						null, null, null);

				// case FLOAT64:
				// data.selectFloating_point(new
				// FloatingPoint(MmsAsn1PrimitiveTypeConverter.toByteArray((Double)
				// daValue)));
				// return data;
				// case TCMD: // coded enum(?)
				// data.selectBit_string(new BitString((((byte[])daValue))));
				// return data;
				// case DBPOS: // coded enum(?)
				// data.selectBit_string(new BitString((((byte[])daValue))));
				// return data;

			case CODEDENUM:
				// data.selectBit_string(new BitString(((byte[]) daValue)));
				// return data;
				// TODO always 2 bits??

			
				byte[] value = (byte[])daValue;
	
				return new Data(null, null,
						null, new BerBitString(value, value.length*8), null,
						null, null, null, null, null, null, null, null, null,
						null);

			case ENUMERATED:
				// data.selectInteger(getNumberValue(da));
				// return data;
				return new Data(null, null, null, null, new BerInteger(getNumberValue(da)), null, null, null, null,
						null, null, null, null, null, null);
			case OCTET_STRING_6:
				if (daValue == null) {
					daValue = new byte[0];
				}
				if (da.getNodeName().equals("TimeOfEntry")) {
					// data.selectBinary_time(new TimeOfDay((byte[]) daValue));
					return new Data(null, null, null, null, null, null, null, null, null, null, new BerOctetString(
							(byte[]) daValue), null, null, null, null);
				}
				else {// other values with OCTET_STRING_6??
					// data.selectOctet_string((byte[]) daValue);
					return new Data(null, null, null, null, null, null, null, new BerOctetString((byte[]) daValue),
							null, null, null, null, null, null, null);
				}

			case OCTET_STRING_64:
				if (daValue == null) {
					daValue = new byte[0];
				}
				return new Data(null, null, null, null, null, null, null, new BerOctetString((byte[]) daValue), null,
						null, null, null, null, null, null);
			case VISIBLE_STRING_32: // fall through
			case VISIBLE_STRING_64: // fall through
			case VISIBLE_STRING_129: // fall through
			case VISIBLE_STRING_255:
				if (daValue == null) {
					daValue = "";
				}
				// data.selectVisible_string((String) daValue);
				// return data;
				try {
					return new Data(null, null, null, null, null, null, null, null, new BerVisibleString(
							(String) daValue), null, null, null, null, null, null);
				} catch (UnsupportedEncodingException e) {
					throw new IllegalAccessError();
				}
			case UNICODE_STRING_255:
				if (daValue == null) {
					daValue = "";
				}
				// data.selectMMSString(new MMSString((String) daValue));
				// return data;
				// possible to do it this way?
				return new Data(null, null, null, null, null, null, null, null, null, null, null, null, null,
						new BerUTF8String(((String) daValue).getBytes()), null);
			case TIMESTAMP:
				// if (daValue == null) {
				// daValue = new Timestamp();
				// }
				// data.selectUtc_time(MmsAsn1PrimitiveTypeConverter
				// .toMmsUtcTime((Timestamp) da));
				// return data;
				return new Data(null, null, null, null, null, null, null, null, null, null, null, null, null, null,
						new BerOctetString(MmsAsn1PrimitiveTypeConverter.toByteArray((Timestamp) da)));
			case QUALITY:
				// if (daValue == null) {
				// daValue = new Quality();
				// }
				// BitString qBits = new BitString();
				// qBits.setValue(MmsAsn1PrimitiveTypeConverter
				// .toByteArray((Quality) da), 3);
				// data.selectBit_string(qBits);

				
				byte[] qvalue = MmsAsn1PrimitiveTypeConverter.toByteArray((Quality) da);
				
				return new Data(null, null,	null, new BerBitString(qvalue, qvalue.length*8), null, null,
						null, null, null, null, null, null, null, null, null);


			}
		}
		return null;
	}

	/**
	 * Converts MMS raw data to a ModelNode. At the moment it doesn't reuse but
	 * recreates all model nodes. When fc != null, the resulting ModelNode only
	 * contains children matching the FunctionalConstraint.
	 * 
	 * Does not write the new values into the model only the copy. ModelNode dd
	 * is already a copy!
	 * 
	 */
	public static void convertMmsDataToDataContainer(ModelNode dd, FunctionalConstraint fc, Data data, String[] mmsFCs)
			throws ServiceError {

		if (data.boolean_ != null) {
			if (!(dd instanceof BOOLEAN)) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}
			((BOOLEAN) dd).setValue(data.boolean_.val);
			// return dd;
		}
		else if (data.bit_string != null) {
			if (!(dd instanceof CodedEnum) && !(dd instanceof Quality) && !(dd instanceof TriggerConditions)
					&& !(dd instanceof OptFields)) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}

			byte[] rawData = data.bit_string.bitString;
			if (dd instanceof BasicDataAttribute) {
				MmsAsn1PrimitiveTypeConverter.toBasicType((BasicDataAttribute) dd, rawData);
				// return dd;

			}
			else if (dd instanceof OptFields) {
				((OptFields) dd).setValue(rawData);
				// return dd;

			}
			else if (dd instanceof TriggerConditions) {
				((TriggerConditions) dd).setValue(rawData);
				// return dd;
			}
		}
		else if (data.integer != null) {

			if (dd instanceof INT8) {
				((INT8) dd).setValue((byte) data.integer.val);
				// return dd;

			}

			else if (dd instanceof INT16) {
				((INT16) dd).setValue((short) data.integer.val);
				// return dd;
			}

			else if (dd instanceof INT32) {
				((INT32) dd).setValue((int) data.integer.val);
				// return dd;
			}

			else if (dd instanceof INT64) {
				((INT64) dd).setValue(data.integer.val);
				// return dd;
			}

			else if (dd instanceof Enumerated) {
				((Enumerated) dd).setValue((int) data.integer.val);
				// return dd;
			}

			else {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}
		}

		else if (data.unsigned != null) {
			if (dd instanceof INT8U) {
				((INT8U) dd).setValue((byte) data.unsigned.val);
			}

			else if (dd instanceof INT16U) {
				// INT16U copy = ((INT16U) dd).copy();
				// copy.setValue(data.getInteger().shortValue());
				// daVal = copy;
				((INT16U) dd).setValue((short) data.unsigned.val);
			}

			else if (dd instanceof INT32U) {
				// INT32U copy = ((INT32U) dd).copy();
				// copy.setValue(data.getInteger().intValue());
				//
				// daVal = copy;
				((INT32U) dd).setValue((int) data.unsigned.val);
			}

			else {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}

			// else if(dd instanceof INT24){
			// INT24U copy = ((INT24U)dd).copy();
			// copy.setValue(data.getInteger());
			// daVal = copy;
			// }
			// long value = data.getUnsigned();
			// daVal = new DataAttributeValue(value, dd.getBasicType());
		}

		else if (data.floating_point != null) {
			if (!(dd instanceof FLOAT32)) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}

			byte[] rawData = data.floating_point.octetString;
			MmsAsn1PrimitiveTypeConverter.toBasicType((BasicDataAttribute) dd, rawData);
		}

		else if (data.octet_string != null) {
			// daVal = new DataAttributeValue(data.getOctet_string(),
			// BasicType.OCTET_STRING_64);
			// other lengths of octet string?
			if (dd instanceof OCTET_STRING_64) {
				// OCTET_STRING_64 copy = ((OCTET_STRING_64) dd).copy();
				// copy.setValue(data.getOctet_string());
				// daVal = copy;
				((OCTET_STRING_64) dd).setValue(data.octet_string.octetString);
			}

			else if (dd instanceof OCTET_STRING_6) {
				((OCTET_STRING_6) dd).setValue(data.octet_string.octetString);
			}

			else {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}

		}
		else if (data.visible_string != null) {
			// daVal = new DataAttributeValue(data.getVisible_string(),
			// BasicType.VISIBLE_STRING_255); // und
			// andere
			// Längen!

			if (dd instanceof VISIBLE_STRING_32) {
				// VISIBLE_STRING_32 copy = ((VISIBLE_STRING_32) dd).copy();
				// copy.setValue(data.getVisible_string());
				// daVal = copy;
				((VISIBLE_STRING_32) dd).setValue(data.visible_string.toString());
			}

			else if (dd instanceof VISIBLE_STRING_64) {
				// VISIBLE_STRING_64 copy = ((VISIBLE_STRING_64) dd).copy();
				// copy.setValue(data.getVisible_string());
				// daVal = copy;
				((VISIBLE_STRING_64) dd).setValue(data.visible_string.toString());
			}

			else if (dd instanceof VISIBLE_STRING_129) {
				// VISIBLE_STRING_129 copy = ((VISIBLE_STRING_129) dd).copy();
				// copy.setValue(data.getVisible_string());
				// daVal = copy;

				((VISIBLE_STRING_129) dd).setValue(data.visible_string.toString());
			}

			else if (dd instanceof VISIBLE_STRING_255) {
				// VISIBLE_STRING_255 copy = ((VISIBLE_STRING_255) dd).copy();
				// copy.setValue(data.getVisible_string());
				// daVal = copy;

				((VISIBLE_STRING_255) dd).setValue(data.visible_string.toString());
			}

			else {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}
		}
		else if (data.generalized_time != null) {
			// TODO type not implemented
		}
		else if (data.binary_time != null) {
			// TODO octet string 6?
			((OCTET_STRING_6) dd).setValue(data.octet_string.octetString);

		}
		else if (data.bcd != null) {
			// TODO type not implemented
		}
		else if (data.booleanArray != null) {
			// TODO type not implemented
		}
		else if (data.array != null && dd instanceof Array) {

			List<Data> items = data.array.seqOf;
			List<ModelNode> childNodes = (ArrayList<ModelNode>) ((Array) dd).getChildren();

			if (childNodes.size() != items.size()) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}

			int i = 0;
			for (Data item : items) {
				convertMmsDataToDataContainer(childNodes.get(i), fc, item, mmsFCs);
				i++;
			}

		}
		else if (data.mMSString != null) {

			// UNICODE_STRING_255 copy = ((UNICODE_STRING_255) dd).copy();
			// copy.setValue(data.getMMSString().getValue());
			// daVal = copy;
			((UNICODE_STRING_255) dd).setValue(data.mMSString.toString());

		}
		else if (data.utc_time != null) {
			// daVal = MmsAsn1PrimitiveTypeConverter.toTimestamp(dd, data
			// .getUtc_time());

			MmsAsn1PrimitiveTypeConverter.toTimestamp((Timestamp) dd, data.utc_time.octetString);
		}

		// if (daVal != null) { // Fall 1: einfaches DataAttribute
		// // ConstructedDataAttribute result = new
		// // ConstructedDataAttribute(dd.getReference().toString(), fc);
		// // result.setValue(daVal);
		// return daVal;
		// }

		else if (data.structure != null) {

			List<ModelNode> result = new ArrayList<ModelNode>();
			DataSequence dataSeq = data.structure;
			Iterator<Data> resIter = dataSeq.seqOf.iterator();

			if (dd.getChildren() == null) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
						"mmsDataToDataContainer: Structure must contain at least one element.");
			}

			List<ModelNode> sortedChildren = new LinkedList<ModelNode>();
			FunctionalConstraint fcTemp = null;
			for (String inOrderFc : mmsFCs) {
				for (ModelNode child : dd.getChildren()) {
					if (child instanceof DataSet) {
						continue;
					}

					fcTemp = FunctionalConstraint.fromString(inOrderFc);
					if (fcTemp == null) {
						continue; // for cases of RP, BR, LG etc
					}

					if (fcTemp.equals(child.getFunctionalConstraint())) {
						if ((fc == null || fc.equals(child.getFunctionalConstraint()))) {
							sortedChildren.add(child);
						}
					}
				}
			}

			for (ModelNode childDef : sortedChildren) {
				if (!resIter.hasNext()) {
					throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
							"mmsDataToDataContainer: Wrong number of results received (A).");
				}

				Data subData = resIter.next();
				// ModelNode da = convertMmsDataToDataContainer(childDef, fc,
				// subData, mmsFCs);
				convertMmsDataToDataContainer(childDef, fc, subData, mmsFCs);
				result.add(childDef);
			}

			if (resIter.hasNext()) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
						"mmsDataToDataContainer: Wrong number of results received (B).");
			}

			if (dd instanceof DataObject) {
				// return new DataObject(dd.getReference(), result, "");
				dd = new DataObject(dd.getReference(), result, "");
				// TODO ok to leave cdcName null?
			}

			else if (dd instanceof ConstructedDataAttribute) {
				// return new ConstructedDataAttribute(dd.getReference(), dd
				// .getFunctionalConstraint(), result);
				dd = new ConstructedDataAttribute(dd.getReference(), dd.getFunctionalConstraint(), result);

			}

			else {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT, "Unknown or invalid ObjectType for: "
						+ dd.getNodeName());
			}

		}
		else {
			throw new ServiceError(ServiceError.CLASS_NOT_SUPPORTED, "Response contained unknown type");
		}
		// TODO correct error code?
	}

}
