/*
 * Copyright Fraunhofer ISE, energy & meteo Systems GmbH, and other contributors 2011
 *
 * This file is part of openIEC61850.
 * For more information visit http://www.openmuc.org 
 *
 * openIEC61850 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * openIEC61850 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with openIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openiec61850.scsm.mms;

import java.nio.ByteBuffer;

import org.openiec61850.common.ServiceError;
import org.openiec61850.common.model.BasicDataAttribute;
import org.openiec61850.common.model.datypes.basic.CodedEnum;
import org.openiec61850.common.model.datypes.basic.FLOAT32;
import org.openiec61850.common.model.datypes.basic.INT16;
import org.openiec61850.common.model.datypes.basic.INT32;
import org.openiec61850.common.model.datypes.basic.OCTET_STRING_64;
import org.openiec61850.common.model.datypes.commonacsi.Quality;
import org.openiec61850.common.model.datypes.commonacsi.Quality.Source;
import org.openiec61850.common.model.datypes.commonacsi.Quality.Validity;
import org.openiec61850.common.model.datypes.commonacsi.Timestamp;

//import org.openiec61850.common.model.basictypes.Dbpos;
//import org.openiec61850.common.model.basictypes.FLOAT32;
//import org.openiec61850.common.model.basictypes.INT16;
//import org.openiec61850.common.model.basictypes.INT32;
//import org.openiec61850.common.model.basictypes.Quality;
//import org.openiec61850.common.model.basictypes.Tcmd;
//import org.openiec61850.common.model.basictypes.Timestamp;
//import org.openiec61850.common.model.basictypes.Quality.Source;
//import org.openiec61850.common.model.basictypes.Quality.Validity;

//import org.openiec61850.jmms.mms.asn1.UtcTime;

/**
 * @author tweidelt<br>
 *         Large parts taken from:
 *         http://www.daniweb.com/code/snippet216874.html
 */
public class MmsAsn1PrimitiveTypeConverter {
	public static final byte MMS_FLOAT32_EXPONENT_WIDTH = 8;
	public static final byte MMS_FLOAT64_EXPONENT_WIDTH = 11;

	public static byte toByte(byte[] b) {
		return (b == null || b.length == 0) ? 0x0 : b[0];
	}

	public static byte[] toByteArray(boolean value) {
		return new byte[] { (byte) (value ? 0x01 : 0x00) };
	}

	public static byte[] toByteArray(byte value) {
		return new byte[] { value };
	}

	// public static byte[] toByteArray(BitSet bits) {
	//
	// byte[] bytes = new byte[(bits.length()+7)/8];
	//
	// for(int i = 0; i < bits.length(); i++){
	// if(bits.get(i)){
	// bytes[i/8] = (byte) (bytes[i/8] | ( 1 << (i % 8)));
	// }
	// }
	//
	//
	// return bytes;
	//
	// }

	/**
	 * Converts a double value to a 9-byte array. First byte contains the
	 * exponent width, normally 11 bits.
	 */
	public static byte[] toByteArray(double value) {
		return ByteBuffer.allocate(1 + 8).put(MMS_FLOAT64_EXPONENT_WIDTH).putDouble(value).array();
	}

	/**
	 * Converts a float value to a 5-byte array. First byte contains the
	 * exponent width, normally 8 bits.
	 */
	public static byte[] toByteArray(float value) {
		return ByteBuffer.allocate(1 + 4).put(MMS_FLOAT32_EXPONENT_WIDTH).putFloat(value).array();
	}

	public static byte[] toByteArray(long value) {
		return new byte[] { (byte) ((value >> 56) & 0xff), (byte) ((value >> 48) & 0xff),
				(byte) ((value >> 40) & 0xff), (byte) ((value >> 32) & 0xff), (byte) ((value >> 24) & 0xff),
				(byte) ((value >> 16) & 0xff), (byte) ((value >> 8) & 0xff), (byte) ((value >> 0) & 0xff) };
	}

	/**
	 * See 61850-8-1 section
	 * "Mapping of quality common data attribute type specified in IEC 61850-7-3"
	 * in chapter "Common Data Types"
	 */
	public static byte[] toByteArray(Quality value) {
		byte[] result = new byte[2];
		switch (value.validity) {
		case GOOD:
			result[0] = (byte) (result[0] | (byte) 0x00);
			break;
		case INVALID:
			result[0] = (byte) (result[0] | (byte) 0x01);
			break;
		case QUESTIONABLE:
			result[0] = (byte) (result[0] | (byte) 0x02);
			break;
		case RESERVED:
			result[0] = (byte) (result[0] | (byte) 0x03);
			break;
		}
		if (value.overflow) {
			result[0] = (byte) (result[0] | (byte) 0x04);
		}
		if (value.outOfRange) {
			result[0] = (byte) (result[0] | (byte) 0x08);
		}
		if (value.badReference) {
			result[0] = (byte) (result[0] | (byte) 0x10);
		}
		if (value.oscillatory) {
			result[0] = (byte) (result[0] | (byte) 0x20);
		}
		if (value.failure) {
			result[0] = (byte) (result[0] | (byte) 0x40);
		}
		if (value.oldData) {
			result[0] = (byte) (result[0] | (byte) 0x80);
		}
		if (value.inconsistent) {
			result[1] = (byte) (result[1] | (byte) 0x01);
		}
		if (value.inaccurate) {
			result[1] = (byte) (result[1] | (byte) 0x02);
		}
		switch (value.source) {
		case PROCESS:
			result[1] = (byte) (result[1] | (byte) 0x00);
			break;
		case SUBSTITUTED:
			result[1] = (byte) (result[1] | (byte) 0x04);
			break;
		}
		if (value.test) {
			result[1] = (byte) (result[1] | (byte) 0x08);
		}
		if (value.operatorBlocked) {
			result[1] = (byte) (result[1] | (byte) 0x10);
		}
		return result;
	}

	public static byte[] toByteArray(short value) {
		return new byte[] { (byte) ((value >> 8) & 0xff), (byte) ((value >> 0) & 0xff) };
	}

	// public static byte[] toByteArray(Tcmd value) {
	// // FIXME Auto-generated method stub
	// return new byte[1];
	// }

	// TODO needs testing!
	/**
	 * See 61850-8-1 Section "TIMESTAMP" in chapter "Common data types"
	 */
	public static OCTET_STRING_64 toMmsUtcTime(Timestamp value) {
		// long quality = value.timeQuality.timeAccuracy & 0x001f;
		// if (value.timeQuality.leapSecondsknown) {
		// quality = quality | 0x80;
		// }
		// if (value.timeQuality.clockFailure) {
		// quality = quality | 0x40;
		// }
		// if (value.timeQuality.clockNotSynchronized) {
		// quality = quality | 0x20;
		// }
		// long timestamp = (value.secondsSinceEpoch << 32) & 0xffffffff
		// | (value.fractionOfSecond << 8) | quality;

		OCTET_STRING_64 mmsUtcTime = new OCTET_STRING_64(value.getReference(), value.getFunctionalConstraint(), "",
				toByteArray(value), false, false);

		// UtcTime mmsUtcTime = new UtcTime();
		// mmsUtcTime.setValue(toByteArray(timestamp));
		// mmsUtcTime.setValue(toByteArray(value));
		return mmsUtcTime;
	}

	public static byte[] toByteArray(Timestamp value) {
		long quality = value.timeQuality.timeAccuracy & 0x001f;
		if (value.timeQuality.leapSecondsknown) {
			quality = quality | 0x80;
		}
		if (value.timeQuality.clockFailure) {
			quality = quality | 0x40;
		}
		if (value.timeQuality.clockNotSynchronized) {
			quality = quality | 0x20;
		}
		long timestamp = (value.secondsSinceEpoch << 32) & 0xffffffff | (value.fractionOfSecond << 8) | quality;

		return toByteArray(timestamp);
	}

	// where is this used?
	// public static double toDouble(byte[] b) {
	// if (b == null || b.length != 9) {
	// return 0x0;
	// }
	// byte[] b2 = new byte[8];
	// System.arraycopy(b, 1, b2, 0, 8);
	// return Double.longBitsToDouble(toLong(b2));
	// }

	public static void toFloat(FLOAT32 dd, byte[] b) {

		FLOAT32 fl = dd;

		if (b == null || b.length != 5) {
			fl.setValue(new Float(0x0));
			// return fl;
		}
		byte[] b2 = new byte[4];
		System.arraycopy(b, 1, b2, 0, 4);

		// if (b2 == null || b2.length != 4) {
		// fl.setValue(Float.intBitsToFloat(0x0));
		// }

		fl.setValue(Float
				.intBitsToFloat(((0xff & b2[0]) << 24 | (0xff & b2[1]) << 16 | (0xff & b2[2]) << 8 | (0xff & b2[3]) << 0)));
		// return fl;
	}

	// is this used?
	public static void toInt(INT32 dd, byte[] r) {
		INT32 int32 = dd;
		if (r == null || r.length != 4) {
			int32.setValue(0x0);
			// return int32;
		}
		int32.setValue(((0xff & r[0]) << 24 | (0xff & r[1]) << 16 | (0xff & r[2]) << 8 | (0xff & r[3]) << 0));
		// return int32;
	}

	public static long toLong(byte[] b) {
		if (b == null || b.length != 8) {
			return 0x0;
		}
		return ((long) (0xff & b[0]) << 56 | (long) (0xff & b[1]) << 48 | (long) (0xff & b[2]) << 40
				| (long) (0xff & b[3]) << 32 | (long) (0xff & b[4]) << 24 | (long) (0xff & b[5]) << 16
				| (long) (0xff & b[6]) << 8 | (long) (0xff & b[7]) << 0);
	}

	// is this used?
	public static void toShort(INT16 dd, byte[] data) {

		INT16 int16 = dd;

		if (data == null || data.length != 2) {
			int16.setValue(new Short((short) 0x0));
			// return int16;
		}

		int16.setValue(new Short((short) ((0xff & data[0]) << 8 | (0xff & data[1]) << 0)));
		// return int16;
	}

	public static void toBasicType(BasicDataAttribute dd, byte[] rawData) throws ServiceError {
		switch (dd.getBasicType()) {
		// BOOLEAN:
		// INT8
		case INT16:
			toShort((INT16) dd, rawData);
			break;
		// INT24
		case INT32:
			toInt((INT32) dd, rawData);
			break;
		// INT128
		// INT8U
		// INT16U
		// INT24U
		// INT32U
		case FLOAT32:
			toFloat((FLOAT32) dd, rawData);
			break;
		// case FLOAT64:
		// return toDouble(rawData);
		// ENUMERATED
		// OCTET_STRING_64
		// VISIBLE_STRING_32
		// VISIBLE_STRING_64
		// VISIBLE_STRING_129
		// VISIBLE_STRING_255
		// UNICODE_STRING_255
		case TIMESTAMP:
			toTimestamp((Timestamp) dd, rawData);
			break;
		case QUALITY:
			toQuality((Quality) dd, rawData);
			break;
		// case DBPOS:
		// return toDbpos((Dbpos)dd, rawData);
		// case TCMD:
		case CODEDENUM:
			toCodedEnum((CodedEnum) dd, rawData);
			break;
		default:
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "invalid or unknown BasicType: "
					+ dd.getBasicType());
		}
	}

	// private static Dbpos toDbpos(Dbpos dd, byte [] rawData) throws
	// ServiceError{
	// Dbpos dbpos = dd.copy();
	// dbpos.setValue(rawData); //need to do more to the rawData?
	// return dbpos;
	// }
	//
	// private static Tcmd toTcmd(Tcmd dd, byte[] rawData) throws ServiceError{
	// Tcmd tcmd = dd.copy();
	// tcmd.setValue(rawData); //need to do more to the rawData?
	// return tcmd;
	// }

	private static void toCodedEnum(CodedEnum dd, byte[] rawData) throws ServiceError {
		// CodedEnum codedEnum = (CodedEnum)dd;
		((CodedEnum) dd).setValue(rawData); // need to do more to the rawData?
		// return codedEnum;
	}

	/**
	 * see {@link #toByteArray(Quality)}
	 */
	private static void toQuality(Quality dd, byte[] rawData) {
		Quality q = dd;
		byte validity = (byte) (rawData[0] & 0x03);
		switch (validity) {
		case 0:
			q.validity = Validity.GOOD;
			break;
		case 1:
			q.validity = Validity.INVALID;
			break;
		case 2:
			q.validity = Validity.RESERVED;
			break;
		case 3:
			q.validity = Validity.QUESTIONABLE;
			break;
		}
		q.overflow = (rawData[0] & 0x04) != 0;
		q.outOfRange = (rawData[0] & 0x08) != 0;
		q.badReference = (rawData[0] & 0x10) != 0;
		q.oscillatory = (rawData[0] & 0x20) != 0;
		q.failure = (rawData[0] & 0x40) != 0;
		q.oldData = (rawData[0] & 0x80) != 0;
		q.inconsistent = (rawData[1] & 0x01) != 0;
		q.inaccurate = (rawData[1] & 0x02) != 0;
		q.source = (rawData[1] & 0x04) != 0 ? Source.SUBSTITUTED : Source.PROCESS;
		q.test = (rawData[1] & 0x08) != 0;
		q.operatorBlocked = (rawData[1] & 0x10) != 0;
		// return q;
	}

	/**
	 * see {@link #toMmsUtcTime(Timestamp)}
	 */
	public static void toTimestamp(Timestamp dd, byte[] rawData) {
		long rawValue = toLong(rawData);

		Timestamp t = dd;
		t.secondsSinceEpoch = (rawValue >> 32) & 0xffffffff;
		t.fractionOfSecond = (int) (rawValue >> 8) & 0xff;
		t.timeQuality.leapSecondsknown = (rawValue & 0x00000080) != 0;
		t.timeQuality.clockFailure = (rawValue & 0x00000040) != 0;
		t.timeQuality.clockNotSynchronized = (rawValue & 0x00000020) != 0;
		t.timeQuality.timeAccuracy = (int) (rawValue & 0x0000001f);
		// return t;
	}

	// public static void toTimestamp(ModelNode dd, UtcTime utcTime) {
	// // if (utcTime == null) {
	// // return null;
	// // }
	// toTimestamp((Timestamp) dd, utcTime.getValue());
	// }
}
