package org.openiec61850.scsm.mms.client;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.openiec61850.common.ServiceError;
import org.openiec61850.common.model.Array;
import org.openiec61850.common.model.BasicDataAttribute;
import org.openiec61850.common.model.ConstructedDataAttribute;
import org.openiec61850.common.model.DataObject;
import org.openiec61850.common.model.FunctionalConstraint;
import org.openiec61850.common.model.LogicalNode;
import org.openiec61850.common.model.ModelNode;
import org.openiec61850.common.model.ObjectReference;
import org.openiec61850.common.model.datypes.basic.BOOLEAN;
import org.openiec61850.common.model.datypes.basic.CodedEnum;
import org.openiec61850.common.model.datypes.basic.FLOAT32;
import org.openiec61850.common.model.datypes.basic.INT16;
import org.openiec61850.common.model.datypes.basic.INT16U;
import org.openiec61850.common.model.datypes.basic.INT32;
import org.openiec61850.common.model.datypes.basic.INT32U;
import org.openiec61850.common.model.datypes.basic.INT64;
import org.openiec61850.common.model.datypes.basic.INT8;
import org.openiec61850.common.model.datypes.basic.INT8U;
import org.openiec61850.common.model.datypes.basic.OCTET_STRING_6;
import org.openiec61850.common.model.datypes.basic.OCTET_STRING_64;
import org.openiec61850.common.model.datypes.basic.UNICODE_STRING_255;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_129;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_255;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_32;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING_64;
import org.openiec61850.common.model.datypes.commonacsi.Quality;
import org.openiec61850.common.model.datypes.commonacsi.Timestamp;
import org.openiec61850.jmms.mms.asn1.ConfirmedResponsePdu;
import org.openiec61850.jmms.mms.asn1.ConfirmedServiceResponse;
import org.openiec61850.jmms.mms.asn1.GetVariableAccessAttributesResponse;
import org.openiec61850.jmms.mms.asn1.MmsPdu;
import org.openiec61850.jmms.mms.asn1.StructComponent;
import org.openiec61850.jmms.mms.asn1.TypeSpecification;
import org.openiec61850.jmms.mms.asn1.TypeSpecification.SubSeq_floating_point;
import org.openiec61850.jmms.mms.asn1.TypeSpecification.SubSeq_structure;
import org.openiec61850.jmms.mms.asn1.TypeSpecification.SubSeq_structure.SubSeqOf_components;


public class DataDefinitionResParser {

	protected static LogicalNode parseGetDataDefinitionResponse(MmsPdu mmsPdu, ObjectReference lnRef, int invokeID)
			throws ServiceError {
		MmsScsmClientAssociation.testForErrorResponse(mmsPdu);

		if (mmsPdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeGetDataDefinitionResponse: Error decoding GetDataDefinitionResponsePdu");
		}

		ConfirmedResponsePdu confirmedResponsePdu = mmsPdu.confirmedResponsePdu;
		MmsScsmClientAssociation.checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		if (confirmedServiceResponse.getVariableAccessAttributes == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeGetDataDefinitionResponse: Error decoding GetDataDefinitionResponsePdu");
		}

		GetVariableAccessAttributesResponse varAccAttrs = confirmedServiceResponse.getVariableAccessAttributes;
		TypeSpecification typeSpec = varAccAttrs.typeSpecification;
		if (typeSpec.structure == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeGetDataDefinitionResponse: Error decoding GetDataDefinitionResponsePdu");
		}

		SubSeqOf_components structure = typeSpec.structure.components;

		List<DataObject> fcDataObjects = new LinkedList<DataObject>();

		FunctionalConstraint fc;
		for (StructComponent fcComponent : structure.seqOf) {
			if (fcComponent.componentName == null) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
						"Error decoding GetDataDefinitionResponsePdu");
			}

			if (fcComponent.componentType.structure == null) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
						"Error decoding GetDataDefinitionResponsePdu");
			}

			String fcString = fcComponent.componentName.toString();

			if (fcString.equals("RP") || fcString.equals("LG") || fcString.equals("BR") || fcString.equals("GO")
					|| fcString.equals("GS") || fcString.equals("MS") || fcString.equals("US")) {
				continue;
			}

			fc = FunctionalConstraint.fromString(fcComponent.componentName.toString());

			SubSeqOf_components subStructure = fcComponent.componentType.structure.components;
			fcDataObjects.addAll(getFCDataObjectsFromSubStructure(lnRef, fc, subStructure));
		}

		return new LogicalNode(lnRef, null, fcDataObjects, null);

	}

	private static List<DataObject> getFCDataObjectsFromSubStructure(ObjectReference lnRef, FunctionalConstraint fc,
			SubSeqOf_components structure) throws ServiceError {

		List<StructComponent> structComponents = structure.seqOf;
		List<DataObject> dataObjects = new ArrayList<DataObject>(structComponents.size());

		for (StructComponent doComp : structComponents) {
			if (doComp.componentName == null) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
						"Error decoding GetDataDefinitionResponsePdu");
			}
			if (doComp.componentType.structure == null) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
						"Error decoding GetDataDefinitionResponsePdu");
			}

			ObjectReference doRef = new ObjectReference(lnRef + "." + doComp.componentName.toString());
			List<ModelNode> children = getDOSubModelNodesFromSubStructure(doRef, fc, doComp.componentType.structure.components);
			dataObjects.add(new DataObject(doRef, children, null, fc));

		}

		return dataObjects;

	}

	private static List<ModelNode> getDOSubModelNodesFromSubStructure(ObjectReference parentRef,
			FunctionalConstraint fc, SubSeqOf_components structure) throws ServiceError {

		Collection<StructComponent> structComponents = structure.seqOf;
		List<ModelNode> dataObjects = new ArrayList<ModelNode>(structComponents.size());

		for (StructComponent component : structComponents) {
			if (component.componentName == null) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
						"Error decoding GetDataDefinitionResponsePdu");
			}

			String childName = component.componentName.toString();
			ObjectReference childReference = new ObjectReference(parentRef + "." + childName);

			dataObjects.add(getModelNodesFromTypeSpecification(childReference, fc, component.componentType));

		}
		return dataObjects;
	}

	private static ModelNode getModelNodesFromTypeSpecification(ObjectReference ref, FunctionalConstraint fc,
			TypeSpecification mmsTypeSpec) throws ServiceError {

		if (mmsTypeSpec.array != null) {

//			int numArrayElements = mmsTypeSpec.getArray().getNumberOfElements().getValue().intValue();
			int numArrayElements = (int)mmsTypeSpec.array.numberOfElements.val;
			List<ModelNode> arrayChildren = new ArrayList<ModelNode>(numArrayElements);
			for (int i = 0; i < numArrayElements; i++) {
				arrayChildren.add(getModelNodesFromTypeSpecification(
						new ObjectReference(ref + "[" + Integer.toString(i) + "]"), fc, mmsTypeSpec.array.elementType));
			}

			return new Array(ref, fc, arrayChildren);

		}

		if (mmsTypeSpec.structure != null) {
			List<ModelNode> children = getDOSubModelNodesFromSubStructure(ref, fc, mmsTypeSpec.structure.components);
			return (new ConstructedDataAttribute(ref, fc, children));
		}

		// it is a single element
		BasicDataAttribute bt = convertMmsBasicTypeSpec(ref, fc, mmsTypeSpec);
		if (bt == null) {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
					"decodeGetDataDefinitionResponse: Unknown data type received");
		}
		return (bt);

	}

	private static BasicDataAttribute convertMmsBasicTypeSpec(ObjectReference ref, FunctionalConstraint fc,
			TypeSpecification mmsTypeSpec) throws ServiceError {

		if (mmsTypeSpec.boolean_ != null) {
			return new BOOLEAN(ref, fc, null, false, false, false);
		}
		if (mmsTypeSpec.bit_string != null) {
			// from fact alone that it is a BitString we cannot tell what Data
			// Attribute Type this is. Therefore we need to take a look at the
			// size and name to map the BitString to the correct Data Attribute
			// Type.

			String daNameAllUpperCase = ref.getName().toUpperCase();
			int bitStringSize = (int)mmsTypeSpec.bit_string.val;

			if (bitStringSize == -13 && (daNameAllUpperCase.equals("Q") || daNameAllUpperCase.equals("SUBQ"))) {
				return new Quality(ref, fc, null, false);
			}
//			if (daNameAllUpperCase.equals("DBPOS") || daNameAllUpperCase.equals("TCMD")) {
			//TODO other sizes for coded enum? Based on the specification in iec61850-7-3, coded enums have 2 bits for now.
			if(bitStringSize == 2){
				return new CodedEnum(ref, fc, null, null, false, false);
			}
//			if (daNameAllUpperCase.equals("TCMD")) {
//				return new CodedEnum(ref, fc, null, null, false, false);
//			}
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "MMS BitString of name: "
					+ ref.getName() + " and BitString size: " + bitStringSize + " is not supported.");
			// TODO does not handle "check" basic type
		}
		else if (mmsTypeSpec.integer != null) {
			int intBits = (int)mmsTypeSpec.integer.val;
			switch (intBits) {
			case 8:
				return new INT8(ref, fc, null, null, false, false);
			case 16:
				return new INT16(ref, fc, null, null, false, false);
			case 32:
				return new INT32(ref, fc, null, null, false, false);
			case 64:
				return new INT64(ref, fc, null, null, false, false);
			}
		}
		else if (mmsTypeSpec.unsigned != null) {
			int intBits = (int)mmsTypeSpec.unsigned.val;
			switch (intBits) {
			case 8:
				return new INT8U(ref, fc, null, null, false, false);
			case 16:
				return new INT16U(ref, fc, null, null, false, false);
				// TODO is INT24U needed?
				// case 24:
				// return new INT24U(ref, fc, null, null, false,false);
			case 32:
				return new INT32U(ref, fc, null, null, false, false);
			}
		}
		else if (mmsTypeSpec.floating_point != null) {
			SubSeq_floating_point fp = mmsTypeSpec.floating_point;
			if(fp.format_width.val == 32){
				return new FLOAT32(ref, fc, null, null, false, false);
			}
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "FLOAT of size: "
					+ fp.format_width.val + " is not supported.");
		}
		else if (mmsTypeSpec.octet_string != null) {
			int length = (int)mmsTypeSpec.octet_string.val;
			switch (length) {
			case -6:
				return new OCTET_STRING_6(ref, fc, null, null, false, false);
			case -64:
				return new OCTET_STRING_64(ref, fc, null, null, false, false);
			}
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "OCTET_STRING of size: " + length
					+ " is not supported.");

		}
		else if (mmsTypeSpec.visible_string != null) {
			int length = (int)mmsTypeSpec.visible_string.val;
			switch (length) {
			case -32:
				return new VISIBLE_STRING_32(ref, fc, null, null, false, false);
			case -64:
				return new VISIBLE_STRING_64(ref, fc, null, null, false, false);
			case -129:
				return new VISIBLE_STRING_129(ref, fc, null, null, false, false);
			case -255:
				return new VISIBLE_STRING_255(ref, fc, null, null, false, false);
			}
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "VISIBLE_STRING of size: " + length
					+ " is not supported.");
		}
		else if (mmsTypeSpec.generalized_time != null) {
			// TODO
		}
		else if (mmsTypeSpec.binary_time != null) {
			// TODO
		}
		else if (mmsTypeSpec.bcd != null) {
			// TODO
		}
		else if (mmsTypeSpec.objId != null) {
			// TODO
		}
		else if (mmsTypeSpec.mMSString != null) {
			int length = (int)mmsTypeSpec.mMSString.val;
			if (length == -255) {
				return new UNICODE_STRING_255(ref, fc, null, null, false, false);
			}
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "UNICODE_STRING of size: " + length
					+ " is not supported.");
		}
		else if (mmsTypeSpec.utc_time != null) {
			return new Timestamp(ref, fc, null, null, false, false);
		}
		return null;
	}

}
