/*
 * Copyright Fraunhofer ISE, energy & meteo Systems GmbH, and other contributors 2011
 *
 * This file is part of openIEC61850.
 * For more information visit http://www.openmuc.org 
 *
 * openIEC61850 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * openIEC61850 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with openIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openiec61850.scsm.mms.client;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.openiec61850.client.ClientACSI;
import org.openiec61850.common.ServiceError;
import org.openiec61850.common.model.LogicalDevice;
import org.openiec61850.common.model.LogicalNode;
import org.openiec61850.common.model.ModelNode;
import org.openiec61850.common.model.ObjectReference;
import org.openiec61850.common.model.Server;
import org.openiec61850.jmms.iso.acse.AcseAssociation;
import org.openiec61850.jmms.iso.acse.ClientAcseSAP;
import org.openiec61850.jmms.mms.asn1.AccessResult;
import org.openiec61850.jmms.mms.asn1.ConfirmedRequestPdu;
import org.openiec61850.jmms.mms.asn1.ConfirmedResponsePdu;
import org.openiec61850.jmms.mms.asn1.ConfirmedServiceRequest;
import org.openiec61850.jmms.mms.asn1.ConfirmedServiceResponse;
import org.openiec61850.jmms.mms.asn1.Data;
import org.openiec61850.jmms.mms.asn1.GetNameListRequest;
import org.openiec61850.jmms.mms.asn1.GetNameListRequest.SubChoice_objectScope;
import org.openiec61850.jmms.mms.asn1.GetNameListResponse;
import org.openiec61850.jmms.mms.asn1.GetVariableAccessAttributesRequest;
import org.openiec61850.jmms.mms.asn1.InitRequestDetail;
import org.openiec61850.jmms.mms.asn1.InitiateRequestPdu;
import org.openiec61850.jmms.mms.asn1.InitiateResponsePdu;
import org.openiec61850.jmms.mms.asn1.ListOfVariableSeq;
import org.openiec61850.jmms.mms.asn1.MmsPdu;
import org.openiec61850.jmms.mms.asn1.ObjectClass;
import org.openiec61850.jmms.mms.asn1.ObjectName;
import org.openiec61850.jmms.mms.asn1.ObjectName.SubSeq_domain_specific;
import org.openiec61850.jmms.mms.asn1.ReadRequest;
import org.openiec61850.jmms.mms.asn1.ReadResponse;
import org.openiec61850.jmms.mms.asn1.ServiceError.SubChoice_errorClass;
import org.openiec61850.jmms.mms.asn1.VariableAccessSpecification;
import org.openiec61850.jmms.mms.asn1.VariableAccessSpecification.SubSeqOf_listOfVariable;
import org.openiec61850.jmms.mms.asn1.VariableSpecification;
import org.openiec61850.jmms.mms.asn1.WriteRequest;
import org.openiec61850.jmms.mms.asn1.WriteRequest.SubSeqOf_listOfData;
import org.openiec61850.scsm.mms.MMSConstants;
import org.openiec61850.scsm.mms.MMSReference;
import org.openiec61850.scsm.mms.MmsAsn1ComplexTypeConverter;
import org.openmuc.jasn1.ber.BerByteArrayOutputStream;
import org.openmuc.jasn1.ber.types.BerBitString;
import org.openmuc.jasn1.ber.types.BerBoolean;
import org.openmuc.jasn1.ber.types.BerInteger;
import org.openmuc.jasn1.ber.types.BerNull;
import org.openmuc.jasn1.ber.types.string.BerVisibleString;

/**
 * @author Stefan Feuerhahn
 * @author Tobias Weidelt
 */
public class MmsScsmClientAssociation implements ClientACSI {

	private static final BerInteger version = new BerInteger(new byte[] { (byte) 0x01, (byte) 0x01 });
	private static final BerBitString proposedParameterCbbBitString = new BerBitString(new byte[] { 0x03, 0x05,
			(byte) 0xf1, 0x00 });

	private AcseAssociation acseAssociation = null;

	private InetAddress serverAddress;
	private int serverPort;

	private int invokeID = 0;

	// not being used:
	// private int negotiatedMaxServOutstandingCalling;
	// private int negotiatedMaxServOutstandingCalled;
	// private int negotiatedDataStructureNestingLevel;

	private int negotiatedMaxPduSize;

	private String[] mmsFCs = { "MX", "ST", "CO", "CF", "DC", "SP", "SG", "RP", "LG", "BR", "GO", "GS", "SV", "SE",
			"EX", "SR", "OR", "BL" };

	protected MmsScsmClientAssociation(String serverAccessPointReference, String authenticationParameter,
			ClientAcseSAP acseSAP, int proposedMaxPduSize, int proposedMaxServOutstandingCalling,
			int proposedMaxServOutstandingCalled, int proposedDataStructureNestingLevel,
			byte[] servicesSupportedCalling, int servicesSupportedCallingNumBits) throws IllegalArgumentException,
			ServiceError {

		negotiatedMaxPduSize = proposedMaxPduSize;

		associate(serverAccessPointReference, authenticationParameter, acseSAP, proposedMaxPduSize,
				proposedMaxServOutstandingCalling, proposedMaxServOutstandingCalled, proposedDataStructureNestingLevel,
				servicesSupportedCalling, servicesSupportedCallingNumBits);
	}

	private int getInvokeID() {
		invokeID = (invokeID + 1) % 2147483647;
		return invokeID;
	}

	private void readServerAPReference(String serverAccessPointReference) throws ServiceError, IllegalArgumentException {
		serverPort = 102;

		String[] serverAPRefSplits = serverAccessPointReference.split(":");
		if (serverAPRefSplits.length < 1 || serverAPRefSplits.length > 2) {
			throw new IllegalArgumentException("serverAccesPointReference argument is invalid");
		}
		if (serverAPRefSplits.length == 2) {
			try {
				serverPort = Integer.decode(serverAPRefSplits[1]);
			} catch (IllegalArgumentException e) {
				throw new IllegalArgumentException("serverAccesPointReference argument is invalid", e);
			}
			if (serverPort < 0 || serverPort > 65535) {
				throw new IllegalArgumentException("invalid port");
			}
		}

		try {
			serverAddress = InetAddress.getByName(serverAPRefSplits[0]);
		} catch (UnknownHostException e) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Hostname could not be resolved.", e);
		}
	}

	private static MmsPdu constructInitRequestPdu(int proposedMaxPduSize, int proposedMaxServOutstandingCalling,
			int proposedMaxServOutstandingCalled, int proposedDataStructureNestingLevel,
			byte[] servicesSupportedCalling, int servicesSupportedCallingNumBits) {

		InitRequestDetail initRequestDetail = new InitRequestDetail(version, proposedParameterCbbBitString,
				new BerBitString(servicesSupportedCalling, servicesSupportedCallingNumBits));

		InitiateRequestPdu initiateRequestPdu = new InitiateRequestPdu(new BerInteger(proposedMaxPduSize),
				new BerInteger(proposedMaxServOutstandingCalling), new BerInteger(proposedMaxServOutstandingCalled),
				new BerInteger(proposedDataStructureNestingLevel), initRequestDetail);

		MmsPdu initiateRequestMMSpdu = new MmsPdu(null, null, null, initiateRequestPdu, null, null);

		return initiateRequestMMSpdu;
	}

	private void handleInitiateResponse(MmsPdu responsePdu, int proposedMaxPduSize,
			int proposedMaxServOutstandingCalling, int proposedMaxServOutstandingCalled,
			int proposedDataStructureNestingLevel) throws ServiceError {

		testForErrorResponse(responsePdu);

		if (responsePdu.initiateResponsePdu == null) {
			acseAssociation.disconnect();
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding InitiateResponse Pdu");
		}

		InitiateResponsePdu initiateResponsePdu = responsePdu.initiateResponsePdu;

		if (initiateResponsePdu.localDetailCalled != null) {
			negotiatedMaxPduSize = (int) initiateResponsePdu.localDetailCalled.val;
		}

		int negotiatedMaxServOutstandingCalling = (int) initiateResponsePdu.negotiatedMaxServOutstandingCalling.val;
		int negotiatedMaxServOutstandingCalled = (int) initiateResponsePdu.negotiatedMaxServOutstandingCalled.val;

		int negotiatedDataStructureNestingLevel;
		if (initiateResponsePdu.negotiatedDataStructureNestingLevel != null) {
			negotiatedDataStructureNestingLevel = (int) initiateResponsePdu.negotiatedDataStructureNestingLevel.val;
		}
		else {
			negotiatedDataStructureNestingLevel = proposedDataStructureNestingLevel;
		}

		if (negotiatedMaxPduSize < MmsScsmClientSAP.MINIMUM_PDU_SIZE || negotiatedMaxPduSize > proposedMaxPduSize
				|| negotiatedMaxServOutstandingCalling > proposedMaxServOutstandingCalling
				|| negotiatedMaxServOutstandingCalling < 0
				|| negotiatedMaxServOutstandingCalled > proposedMaxServOutstandingCalled
				|| negotiatedMaxServOutstandingCalled < 0
				|| negotiatedDataStructureNestingLevel > proposedDataStructureNestingLevel
				|| negotiatedDataStructureNestingLevel < 0) {
			acseAssociation.disconnect();
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT, "Error negotiating paramters");
		}

		int version = (int) initiateResponsePdu.mmsInitResponseDetail.negotiatedVersionNumber.val;
		if (version != 1) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Unsupported version number was negotiated.");
		}

		byte[] servicesSupported = initiateResponsePdu.mmsInitResponseDetail.servicesSupportedCalled.bitString;
		if ((servicesSupported[0] & 0x40) != 0x40) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Obligatory services are not supported by the server.");
		}
	}

	private void associate(String serverAccessPointReference, String authenticationParameter, ClientAcseSAP acseSAP,
			int proposedMaxPduSize, int proposedMaxServOutstandingCalling, int proposedMaxServOutstandingCalled,
			int proposedDataStructureNestingLevel, byte[] servicesSupportedCalling, int servicesSupportedCallingNumBits)
			throws ServiceError, IllegalArgumentException {

		readServerAPReference(serverAccessPointReference);

		MmsPdu initiateRequestMMSpdu = constructInitRequestPdu(proposedMaxPduSize, proposedMaxServOutstandingCalling,
				proposedMaxServOutstandingCalled, proposedDataStructureNestingLevel, servicesSupportedCalling,
				servicesSupportedCallingNumBits);

		BerByteArrayOutputStream berOStream = new BerByteArrayOutputStream(500);
		try {
			initiateRequestMMSpdu.encode(berOStream, true);
		} catch (IOException e) {
			throw new ServiceError(ServiceError.FATAL, e);
		}

		try {
			acseAssociation = acseSAP.associate(serverAddress, serverPort, berOStream.getByteBuffer());

		} catch (IOException e) {
			if (acseAssociation != null) {
				acseAssociation.disconnect();
			}
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT, "Error starting association.",
					e);
		}
		ByteBuffer initResponse = acseAssociation.getAssociateResponseAPDU();

		MmsPdu initiateResponseMMSpdu = new MmsPdu();
		ByteArrayInputStream iStream = new ByteArrayInputStream(initResponse.array(), initResponse.arrayOffset()
				+ initResponse.position(), initResponse.limit() - initResponse.position());
		try {
			initiateResponseMMSpdu.decode(iStream, null);
		} catch (IOException e) {
			throw new ServiceError(ServiceError.FATAL, e);
		}

		handleInitiateResponse(initiateResponseMMSpdu, proposedMaxPduSize, proposedMaxServOutstandingCalling,
				proposedMaxServOutstandingCalled, proposedDataStructureNestingLevel);
	}

	/**
	 * The Model contains not SubDataObjects because they cannot be
	 * distinguished from Constructed Data Attributes in MMS. Also at the moment
	 * the model only contains FCDataObjects
	 */
	public Server retrieveModel() throws ServiceError {

		List<String> ldNames = retrieveLogicalDevices();
		List<List<String>> lnNames = new ArrayList<List<String>>(ldNames.size());

		for (int i = 0; i < ldNames.size(); i++) {
			lnNames.add(retrieveLogicalNodes(ldNames.get(i)));
		}
		List<LogicalDevice> lds = new ArrayList<LogicalDevice>();
		for (int i = 0; i < ldNames.size(); i++) {
			List<LogicalNode> lns = new ArrayList<LogicalNode>();
			for (int j = 0; j < lnNames.get(i).size(); j++) {
				lns.add(retrieveDataDefinitions(new ObjectReference(ldNames.get(i) + "/" + lnNames.get(i).get(j))));
			}
			lds.add(new LogicalDevice(new ObjectReference(ldNames.get(i)), lns));

		}
		return new Server(lds);
	}

	public ModelNode getDataValues(ModelNode modelNode) throws ServiceError {

		MmsPdu mmsReq = constructGetDataValuesRequest(modelNode);
		MmsPdu mmsRes = encodeWriteReadDecode(mmsReq);
		return decodeGetDataValuesResponse(mmsRes, modelNode);
	}

	private MmsPdu constructGetDataValuesRequest(ModelNode modelNode) {
		ObjectName objectName = MMSReference.toMmsObjectName(modelNode.getReference(),
				modelNode.getFunctionalConstraint());

		VariableSpecification varSpec = new VariableSpecification(objectName, null, null, null, null);

		ListOfVariableSeq seq = new ListOfVariableSeq(varSpec, null);

		List<ListOfVariableSeq> listOfVariables = new ArrayList<ListOfVariableSeq>(1);
		listOfVariables.add(seq);

		VariableAccessSpecification varAccessSpec = new VariableAccessSpecification(new SubSeqOf_listOfVariable(
				listOfVariables), null);

		ReadRequest readRequest = new ReadRequest(new BerBoolean(false), varAccessSpec);

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(null, readRequest, null, null,
				null, null, null);

		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(getInvokeID()),
				confirmedServiceRequest);
		MmsPdu mms = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);
		return mms;
	}

	private ModelNode decodeGetDataValuesResponse(MmsPdu responsePdu, ModelNode modelNode) throws ServiceError {
		testForErrorResponse(responsePdu);

		if (responsePdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding GetDataValuesReponsePdu");
		}

		ConfirmedResponsePdu confirmedResponsePdu = responsePdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		if (confirmedServiceResponse.read == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding GetDataValuesReponsePdu");
		}

		ReadResponse readResponse = confirmedServiceResponse.read;
		List<AccessResult> listOfAccessResults = readResponse.listOfAccessResult.seqOf;

		if (listOfAccessResults.size() != 1) {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "Multiple results received.");
		}

		AccessResult accRes = listOfAccessResults.iterator().next();
		// first and only (?)
		if (accRes.failure == null) {
			MmsAsn1ComplexTypeConverter.convertMmsDataToDataContainer(modelNode, modelNode.getFunctionalConstraint(),
					new Data(accRes.array, accRes.structure, accRes.boolean_, accRes.bit_string, accRes.integer,
							accRes.unsigned, accRes.floating_point, accRes.octet_string, accRes.visible_string,
							accRes.generalized_time, accRes.binary_time, accRes.bcd, accRes.booleanArray,
							accRes.mMSString, accRes.utc_time), mmsFCs);

			return modelNode;
		}
		throw new ServiceError(mmsDataAccessErrorToServiceError(accRes.failure),
				"decodeGetDataValuesResponse: DataAccessError received.");
	}

	private int mmsDataAccessErrorToServiceError(BerInteger dataAccessError) {
		switch ((int) dataAccessError.val) {
		case 1: // hardware_fault
			return ServiceError.FAILED_DUE_TO_SERVER_CONSTRAINT;
		case 2: // temporarily-unavailable
			return ServiceError.INSTANCE_LOCKED_BY_OTHER_CLIENT;
		case 3: // object-access-denied
			return ServiceError.ACCESS_VIOLATION;
		case 5: // invalid-address
			return ServiceError.PARAMETER_VALUE_INCONSISTENT;
		case 7: // type-inconsistent
			return ServiceError.TYPE_CONFLICT;
		case 10: // object-non-existent
			return ServiceError.INSTANCE_NOT_AVAILABLE;
		default:
			return ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT; // ?
		}
	}

	private MmsPdu constructGetLDDirectoryRequest(String ldRef, String continueAfter) {
		// TODO speed this up by preparing the mmspdu in the constructor
		ObjectClass objectClass = new ObjectClass(new BerInteger(MMSConstants.NAMED_VARIABLE));
		// objectClass.selectBasicObjectClass((long)
		// MMSConstants.NAMED_VARIABLE);
		// GetNameListRequest getNameListRequest = new GetNameListRequest();
		// getNameListRequest.setObjectClass(objectClass);
		// DomainSpecificChoiceType domainSpecificChoiceType = new
		// DomainSpecificChoiceType();
		// domainSpecificChoiceType.selectBasic(new BasicIdentifier(new
		// String(ldRef.getValue())));
		GetNameListRequest getNameListRequest = null;
		try {
			SubChoice_objectScope objectScopeChoiceType = new SubChoice_objectScope(null, new BerVisibleString(ldRef),
					null);
			// objectScopeChoiceType.selectDomainSpecific(domainSpecificChoiceType);
			// objectScopeChoiceType.selectDomainSpecific(new
			// Identifier(ldRef));
			// getNameListRequest.setObjectScope(objectScopeChoiceType);

			if (continueAfter != "") {
				getNameListRequest = new GetNameListRequest(objectClass, objectScopeChoiceType, new BerVisibleString(
						continueAfter));
			}
			else {
				getNameListRequest = new GetNameListRequest(objectClass, objectScopeChoiceType, null);
			}
		} catch (UnsupportedEncodingException e) {
			throw new IllegalAccessError();
		}

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(getNameListRequest, null, null,
				null, null, null, null);
		// confirmedServiceRequest.selectGetNameList(getNameListRequest);
		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(getInvokeID()),
				confirmedServiceRequest);
		// confirmedRequestPdu.setInvokeID(new
		// Unsigned32(Long.valueOf(getInvokeID())));
		// confirmedRequestPdu.setConfirmedServiceRequest(confirmedServiceRequest);
		MmsPdu confirmedRequestMMSPdu = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);
		// confirmedRequestMMSPdu.selectConfirmedRequestPdu(confirmedRequestPdu);

		// ContinueAfterChoiceType continueAfterCT = new
		// ContinueAfterChoiceType();
		// continueAfterCT.selectBasic(new BasicIdentifier(continueAfter));
		// if (continueAfter != "") {
		// // getNameListRequest.setContinueAfter(continueAfterCT);
		// getNameListRequest.setContinueAfter(new Identifier(continueAfter));
		// }

		return confirmedRequestMMSPdu;
	}

	private MmsPdu constructGetDataDefinitionRequest(ObjectReference lnRef) {

		// ObjectClass objectClass = new ObjectClass();
		// objectClass.selectBasicObjectClass((long)
		// MMSConstants.NAMED_VARIABLE);
		SubSeq_domain_specific domainSpec = null;
		try {
			domainSpec = new SubSeq_domain_specific(new BerVisibleString(lnRef.get(0)), new BerVisibleString(
					lnRef.get(1)));
		} catch (UnsupportedEncodingException e) {
			throw new IllegalAccessError();
		}

		ObjectName objectName = new ObjectName(null, domainSpec, null);
		// Domain_specificSequenceType value = new
		// Domain_specificSequenceType();
		// value.setDomainId(new Identifier(lnRef.get(0)));
		// value.setItemId(new Identifier(lnRef.get(1)));
		// objectName.selectDomain_specific(value);

		GetVariableAccessAttributesRequest getVariableAccessAttributesRequest = new GetVariableAccessAttributesRequest(
				objectName, null);
		// getVariableAccessAttributesRequest.selectName(objectName);

		// GetNameListRequest getNameListRequest = new GetNameListRequest();
		// getNameListRequest.setObjectClass(objectClass);
		// DomainSpecificChoiceType domainSpecificChoiceType = new
		// DomainSpecificChoiceType();
		// domainSpecificChoiceType.selectBasic(new BasicIdentifier(new
		// String(ldRef.getValue())));
		// ObjectScopeChoiceType objectScopeChoiceType = new
		// ObjectScopeChoiceType();
		// //
		// objectScopeChoiceType.selectDomainSpecific(domainSpecificChoiceType);
		// objectScopeChoiceType.selectDomainSpecific(new
		// Identifier(ldRef.getValue()));
		// getNameListRequest.setObjectScope(objectScopeChoiceType);

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(null, null, null,
				getVariableAccessAttributesRequest, null, null, null);
		// confirmedServiceRequest.selectGetNameList(getNameListRequest);
		// confirmedServiceRequest
		// .selectGetVariableAccessAttributes(getVariableAccessAttributesRequest);

		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(getInvokeID()),
				confirmedServiceRequest);
		// confirmedRequestPdu.setInvokeID(new Unsigned32(Long
		// .valueOf(getInvokeID())));
		// confirmedRequestPdu.setConfirmedServiceRequest(confirmedServiceRequest);
		MmsPdu confirmedRequestMMSPdu = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);
		// confirmedRequestMMSPdu.selectConfirmedRequestPdu(confirmedRequestPdu);

		// ContinueAfterChoiceType continueAfterCT = new
		// ContinueAfterChoiceType();
		// continueAfterCT.selectBasic(new BasicIdentifier(continueAfter));
		// if (continueAfter != "") {
		// // getNameListRequest.setContinueAfter(continueAfterCT);
		// getNameListRequest.setContinueAfter(new Identifier(continueAfter));
		// }

		return confirmedRequestMMSPdu;
	}

	//
	// public List<String> getLogicalDeviceDirectory(String ldRef) throws
	// ServiceError {
	// LogicalDevice ld = findOrRetrieveLogicalDevice(ldRef);
	// if (ld == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE);
	// }
	//
	// retrieveLogicalNodes(ld);
	//
	// ArrayList<String> lnReferences = new ArrayList<String>();
	// for (LogicalNode ln : ld.getLNodes()) {
	// lnReferences.add(ln.getReference().toString());
	// }
	// return lnReferences;
	// }
	//
	public List<String> retrieveLogicalNodes(String ld) throws ServiceError {
		List<String> lns = new LinkedList<String>();
		String continueAfterRef = "";
		do {
			MmsPdu getLDDirectoryRequestMMSPdu = constructGetLDDirectoryRequest(ld, continueAfterRef);
			MmsPdu getLDDirectoryResponseMMSPdu = encodeWriteReadDecode(getLDDirectoryRequestMMSPdu);
			continueAfterRef = decodeGetDirectoryResponse(getLDDirectoryResponseMMSPdu, lns);

		} while (continueAfterRef != "");
		return lns;
	}

	/**
	 * Decodes an MMS response which contains the structure of a LD and its LNs
	 * including names of DOs.
	 */
	private String decodeGetDirectoryResponse(MmsPdu getLDDirectoryResponseMmsPdu, List<String> lns)
			throws ServiceError {

		testForErrorResponse(getLDDirectoryResponseMmsPdu);

		// if (!getLDDirectoryResponseMmsPdu.isConfirmedResponsePduSelected()) {
		if (getLDDirectoryResponseMmsPdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeGetLDDirectoryResponse: Error decoding server response");
		}

		ConfirmedResponsePdu confirmedResponsePdu = getLDDirectoryResponseMmsPdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		// if (!confirmedServiceResponse.isGetNameListSelected()) {
		if (confirmedServiceResponse.getNameList == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeGetLDDirectoryResponse: Error decoding server response");
		}

		GetNameListResponse getNameListResponse = confirmedServiceResponse.getNameList;

		List<BerVisibleString> identifiers = getNameListResponse.listOfIdentifier.seqOf;

		if (identifiers.size() == 0) {
			throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
					"decodeGetLDDirectoryResponse: Instance not available");
		}

		BerVisibleString identifier = null;
		Iterator<BerVisibleString> it = identifiers.iterator();

		String idString;

		while (it.hasNext()) {
			identifier = it.next();
			idString = identifier.toString();

			if (idString.indexOf('$') == -1) {
				lns.add(idString);
			}
		}

		if (getNameListResponse.moreFollows != null && getNameListResponse.moreFollows.val == false) {
			return "";
		}
		else {
			return identifier.toString();
		}
	}

	static void testForErrorResponse(MmsPdu mmsResponsePdu) throws ServiceError {
		if (mmsResponsePdu.initiateErrorPdu == null) {
			return;
		}

		// InitiateErrorPdu errorPdu = mmsResponsePdu.initiateErrorPdu;
		org.openiec61850.jmms.mms.asn1.ServiceError asn1ServiceErr = mmsResponsePdu.initiateErrorPdu;

		// ErrorClassChoiceType errClass = asn1ServiceErr.getErrorClass();
		SubChoice_errorClass errClass = asn1ServiceErr.errorClass;
		if (errClass.others != null) { // TODO auf andere Fehlerklassen
			// prüfen
			throw new ServiceError((int) errClass.others.val); // TODO
			// geraten
		}
		throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT);
	}

	private static MmsPdu constructGetServerDirectoryRequest(int invokeID) {
		ObjectClass objectClass = new ObjectClass(new BerInteger(MMSConstants.DOMAIN));
		// objectClass.selectBasicObjectClass((long) MMSConstants.DOMAIN);
		// the previous statement was replace the following to imitate a siemens
		// iec client:
		// MMSConstants.NAMED_VARIABLE);
		// MMSConstants.NAMED_VARIABLE_LIST);

		// GetNameListRequest getNameListRequest = new GetNameListRequest();
		// getNameListRequest.setObjectClass(objectClass);

		GetNameListRequest.SubChoice_objectScope objectScope = new GetNameListRequest.SubChoice_objectScope(
				new BerNull(), null, null);
		// ObjectScopeChoiceType objectScopeChoiceType = new
		// ObjectScopeChoiceType();
		// objectScopeChoiceType.selectVmdSpecific();
		// objectScopeChoiceType.selectAaSpecific();

		GetNameListRequest getNameListRequest = new GetNameListRequest(objectClass, objectScope, null);

		// getNameListRequest.setObjectScope(objectScopeChoiceType);

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(getNameListRequest, null, null,
				null, null, null, null);

		// ConfirmedServiceRequest confirmedServiceRequest = new
		// ConfirmedServiceRequest();
		// confirmedServiceRequest.selectGetNameList(getNameListRequest);

		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(invokeID),
				confirmedServiceRequest);
		// ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu();
		// confirmedRequestPdu.setInvokeID(new
		// Unsigned32(Long.valueOf(getInvokeID())));
		// confirmedRequestPdu.setConfirmedServiceRequest(confirmedServiceRequest);

		MmsPdu confirmedRequestMMSPdu = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);
		// MmsPdu confirmedRequestMMSPdu = new MmsPdu();
		// confirmedRequestMMSPdu.selectConfirmedRequestPdu(confirmedRequestPdu);

		return confirmedRequestMMSPdu;

	}

	protected static void checkInvokeID(ConfirmedResponsePdu confirmedResponsePdu, int invokeID) throws ServiceError {
		if (confirmedResponsePdu.invokeID.val != invokeID) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding Get Server Directory Response Pdu: incorrect invokeID");
		}
	}

	private List<String> decodeGetServerDirectoryResponse(MmsPdu getServerDirectoryResponseMmsPdu) throws ServiceError {
		testForErrorResponse(getServerDirectoryResponseMmsPdu);

		if (getServerDirectoryResponseMmsPdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding Get Server Directory Response Pdu");
		}

		ConfirmedResponsePdu confirmedResponsePdu = getServerDirectoryResponseMmsPdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		if (confirmedServiceResponse.getNameList == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding Get Server Directory Response Pdu");
		}

		List<BerVisibleString> identifiers = confirmedServiceResponse.getNameList.listOfIdentifier.seqOf;
		ArrayList<String> objectRefs = new ArrayList<String>(); // ObjectReference[identifiers.size()];

		for (BerVisibleString identifier : identifiers) {
			objectRefs.add(identifier.toString());
		}

		return objectRefs;
	}

	private List<String> retrieveLogicalDevices() throws ServiceError {
		MmsPdu requestPdu = constructGetServerDirectoryRequest(getInvokeID());
		MmsPdu responsePdu = encodeWriteReadDecode2(requestPdu);
		return decodeGetServerDirectoryResponse(responsePdu);
	}

	/**
	 * "Encode, write, read, decode"
	 */
	private MmsPdu encodeWriteReadDecode(MmsPdu requestPdu) throws ServiceError {
		BerByteArrayOutputStream baos = new BerByteArrayOutputStream(500);
		try {
			// encoder.encode(requestPdu, baos);
			requestPdu.encode(baos, true);
		} catch (Exception e) {
			acseAssociation.disconnect();
			throw new RuntimeException("Error encoding MmsPdu.", e);
		}
		// byte[] encodedRequestPdu = baos.toByteArray();

		// send pdu
		try {
			acseAssociation.send(baos.getByteBuffer());
		} catch (IOException e) {
			acseAssociation.disconnect();
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT, "Error sending packet.", e);
		}

		ByteBuffer encodedResponsePdu;
		try {
			encodedResponsePdu = acseAssociation.receive();
		} catch (IOException e) {
			acseAssociation.disconnect();
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT, "Error receiving packet.", e);
		}

		MmsPdu decodedResponsePdu = new MmsPdu();
		try {
			// decodedResponsePdu = decoder.decode(new ByteArrayInputStream(
			// encodedResponsePdu), MmsPdu.class);
			decodedResponsePdu
					.decode(new ByteArrayInputStream(encodedResponsePdu.array(), encodedResponsePdu.arrayOffset()
							+ encodedResponsePdu.position(), encodedResponsePdu.limit() - encodedResponsePdu.position()),
							null);

		} catch (Exception e) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT, "Error decoding MmsPdu", e);
		}
		return decodedResponsePdu;
	}

	/**
	 * "Encode, write, read, decode"
	 */
	private MmsPdu encodeWriteReadDecode2(MmsPdu requestPdu) throws ServiceError {

		BerByteArrayOutputStream berOStream = new BerByteArrayOutputStream(500);
		try {
			requestPdu.encode(berOStream, true);
		} catch (IOException e) {
			throw new ServiceError(ServiceError.FATAL, e);
		}
		// byte[] requestPduEncoded = berOStream.getArray();

		// ByteArrayOutputStream baos = new ByteArrayOutputStream();
		// try {
		// encoder.encode(requestPdu, baos);
		// } catch (Exception e) {
		// acseAssociation.disconnect();
		// throw new RuntimeException("Error encoding MmsPdu.", e);
		// }
		// byte[] encodedRequestPdu = baos.toByteArray();

		// send pdu
		try {
			acseAssociation.send(berOStream.getByteBuffer());
		} catch (IOException e) {
			acseAssociation.disconnect();
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT, "Error sending packet.", e);
		}

		ByteBuffer encodedResponsePdu;
		try {
			encodedResponsePdu = acseAssociation.receive();
		} catch (IOException e) {
			acseAssociation.disconnect();
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT, "Error receiving packet.", e);
		}

		MmsPdu decodedResponsePdu = new MmsPdu();
		try {

			decodedResponsePdu
					.decode(new ByteArrayInputStream(encodedResponsePdu.array(), encodedResponsePdu.arrayOffset()
							+ encodedResponsePdu.position(), encodedResponsePdu.limit() - encodedResponsePdu.position()),
							null);
		} catch (Exception e) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT, "Error decoding MmsPdu", e);
		}
		return decodedResponsePdu;
	}

	//
	// public void close() {
	// if (pConnection != null) {
	// pConnection.disconnect();
	// }
	// }
	//
	// public List<String> getLogicalNodeDirectory(String lnRef, ACSIClass
	// acsiClass) throws ServiceError {
	// if (lnRef == null || !new ObjectReference(lnRef).isLNRef()) {
	// return null;
	// }
	//
	// List<String> dataNames = new LinkedList<String>();
	// IModelNode logicalNode = findOrRetrieveNode(lnRef);
	// if (logicalNode != null && logicalNode instanceof LogicalNode) {
	//
	// if (acsiClass == ACSIClass.DATA_SET) {
	// // retrieve DS list from server to get any configured dataSets
	// // as well
	// retrieveDataSetList((LogicalNode) logicalNode);
	// List<DataSet> lnDataSets = ((LogicalNode) logicalNode).getDataSets();
	// for (DataSet ds : lnDataSets) {
	// dataNames.add(ds.getNodeName());
	// }// non persistent DS are not included
	// }
	//
	// else {// data object
	//
	// if (!logicalNode.hasChildren()) {
	//
	// retrieveLogicalNodeDirectory((LogicalNode) logicalNode);
	// logicalNode = findOrRetrieveNode(lnRef);
	//
	// }
	//
	// for (IModelNode child : logicalNode) {
	// switch (child.getObjectType()) {
	// case DATA_OBJECT:
	// case ARRAY_OF_DO:
	// if (acsiClass == ACSIClass.DATA_OBJECT) {
	// dataNames.add(child.getNodeName());
	// }
	// break;
	// }
	// }
	//
	// }
	// }
	//
	// if (dataNames.size() == 0) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "No Objects found in LogicalNode. Logical Node should contain at least one DO.");
	// }
	// return dataNames;
	//
	// }
	//
	// protected void retrieveDataSetList(LogicalNode lnNode) throws
	// ServiceError {
	// // call getNameList using ld device name
	// LogicalDevice ldNode =
	// findOrRetrieveLogicalDevice(lnNode.getReference().getLDName());
	//
	// if (ldNode != null) {
	// // in case it exceeds the maximum mmsPdu size
	// String continueAfterRef = "";
	// do {
	// MmsPdu requestPdu =
	// constructGetLogicalNodeDataSetRequest(ldNode.getReference().toString(),
	// "");
	// // getParentRef() + "/"
	// // + lnNode.getNodeName(), "");// want LD/LN
	// MmsPdu responsePdu = encodeWriteReadDecode(requestPdu);
	// continueAfterRef = decodeGetDataSetListResponse(responsePdu, ldNode);
	//
	// } while (continueAfterRef != "");
	// }
	// else {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "retrieveDataSetList: Logical device not found");
	// }
	//
	// }
	//
	// private MmsPdu constructGetLogicalNodeDataSetRequest(String ldRef, String
	// continueAfter) throws ServiceError {
	// ObjectClass objectClass = new ObjectClass();
	// objectClass.selectBasicObjectClass((long)
	// MMSConstants.NAMED_VARIABLE_LIST);
	// GetNameListRequest getNameListRequest = new GetNameListRequest();
	// getNameListRequest.setObjectClass(objectClass);
	//
	// ObjectScopeChoiceType objectScopeChoiceType = new
	// ObjectScopeChoiceType();
	//
	// objectScopeChoiceType.selectDomainSpecific(new Identifier(ldRef));
	// getNameListRequest.setObjectScope(objectScopeChoiceType);
	//
	// ConfirmedServiceRequest confirmedServiceRequest = new
	// ConfirmedServiceRequest();
	// confirmedServiceRequest.selectGetNameList(getNameListRequest);
	// ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu();
	// confirmedRequestPdu.setInvokeID(new
	// Unsigned32(Long.valueOf(getInvokeID())));
	// confirmedRequestPdu.setConfirmedServiceRequest(confirmedServiceRequest);
	// MmsPdu confirmedRequestMMSPdu = new MmsPdu();
	// confirmedRequestMMSPdu.selectConfirmedRequestPdu(confirmedRequestPdu);
	//
	// if (continueAfter != "") {
	// getNameListRequest.setContinueAfter(new Identifier(continueAfter));
	// }
	//
	// return confirmedRequestMMSPdu;
	//
	// }
	//
	// private String decodeGetDataSetListResponse(MmsPdu responsePdu,
	// LogicalDevice ldNode) throws ServiceError {
	// testForErrorResponse(responsePdu);
	//
	// if (!responsePdu.isConfirmedResponsePduSelected()) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "decodeGetDataSetListResponse: Error decoding server response");
	// }
	//
	// ConfirmedResponsePdu confirmedResponsePdu =
	// responsePdu.getConfirmedResponsePdu();
	// checkInvokeID(confirmedResponsePdu);
	//
	// ConfirmedServiceResponse confirmedServiceResponse =
	// confirmedResponsePdu.getConfirmedServiceResponse();
	// if (!confirmedServiceResponse.isGetNameListSelected()) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "decodeGetDataSetListResponse: Error decoding server response");
	// }
	//
	// GetNameListResponse getNameListResponse =
	// confirmedServiceResponse.getGetNameList();
	//
	// Collection<Identifier> identifiers =
	// getNameListResponse.getListOfIdentifier();
	//
	// if (identifiers.size() == 0) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "decodeGetDataSetListResponse: Instance not available");
	// }
	//
	// // begin getting list of data sets
	// Identifier identifier = null;
	// Iterator<Identifier> it = identifiers.iterator();
	//
	// String idString;
	// List<DataSet> lnDataSets;
	//
	// int firstDollar;
	// LogicalNode lnNode;
	// String dataSetName;
	//
	// while (it.hasNext()) {
	// boolean add = true;
	// identifier = it.next();
	// idString = identifier.getValue(); // format of:
	// // LnName$FC$dataSetName
	//
	// // does the data set already exist?
	// // if yes, do not replace in case it refers to DOs/DAs
	// firstDollar = idString.indexOf('$');
	//
	// if (firstDollar < 0) {
	// throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "");
	// }
	//
	// // find logical node
	// lnNode = ldNode.getLN(idString.substring(0, firstDollar));
	//
	// if (lnNode == null) {
	// throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
	// "decodeGetDataSetListResponse: can't find logical node");
	// }
	//
	// lnDataSets = lnNode.getDataSets();
	// dataSetName = idString.substring(idString.lastIndexOf('$') + 1,
	// idString.length());
	//
	// for (DataSet dsCheck : lnDataSets) {
	// if (dataSetName.equals(dsCheck.getNodeName())) {
	// add = false;
	// }
	// }
	// if (add) {
	// DataSet ds = new DataSet(ldNode.getNodeName() + "/" +
	// idString.replace('$', '.'));
	// lnNode.addDataSet(ds);
	// }
	// }
	//
	// if (getNameListResponse.getMoreFollows() != null &&
	// !getNameListResponse.getMoreFollows()) {
	// return "";
	// }
	// else {
	// return identifier.getValue();
	// }
	// }
	//
	// protected void retrieveLogicalNodeDirectory(LogicalNode ln) throws
	// ServiceError {
	// if (ln != null) {
	// // retrieves model structures incl. the names of data objects
	// LogicalDevice parent =
	// findOrRetrieveLogicalDevice(ln.getReference().getParentRef());
	// retrieveLogicalNodes(parent);// this gets the LNs, DOs and FCs
	// }
	// // retrieveDataDefinitions(ln);
	// }
	//
	// public List<String> getDataDirectory(String doRef) throws ServiceError {
	// IModelNode doNode = getDataDefinition(doRef);
	// List<String> names = new LinkedList<String>();
	// for (IModelNode child : doNode) {
	// names.add(child.getNodeName());
	// }
	// return names;
	// }
	//
	// public IModelNode getDataDefinition(String ref) throws ServiceError {
	//
	// IModelNode node = findOrRetrieveNode(ref);
	// if (node != null && (node.hasChildren() || node.getObjectType() ==
	// ObjectType.BASIC_DATA_ATTRIBUTE)) {
	// return node;
	// }
	//
	// // Node is not complete
	// ObjectReference oRef = new ObjectReference(ref);
	// LogicalDevice ld = model.getLD(oRef);
	// if (ld == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE);
	// }
	// LogicalNode ln = ld.getLN(oRef);
	// if (ln == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE);
	// }
	// retrieveDataDefinitions(ln);
	// return model.findNode(ref);
	// }
	//
	protected LogicalNode retrieveDataDefinitions(ObjectReference lnRef) throws ServiceError {
		MmsPdu getDataDefinitionRequestMMSPdu = constructGetDataDefinitionRequest(lnRef);
		MmsPdu getDataDefinitionResponseMMSPdu = encodeWriteReadDecode(getDataDefinitionRequestMMSPdu);
		return DataDefinitionResParser.parseGetDataDefinitionResponse(getDataDefinitionResponseMMSPdu, lnRef, invokeID);
	}

	public ModelNode setDataValues(ModelNode container) throws ServiceError {
		MmsPdu mmsReq = constructSetDataValuesRequest(container);
		MmsPdu mmsRes = encodeWriteReadDecode(mmsReq);
		handleSetDataValuesResponse(mmsRes);

		// TODO return updated node or nothing?
		return null;
	}

	private MmsPdu constructSetDataValuesRequest(ModelNode container) throws ServiceError {
		// Create and configure Pdus in reverse order
		ObjectName objectName = MMSReference.toMmsObjectName(container.getReference(),
				container.getFunctionalConstraint());

		// VariableAccessSpecification variableAccessSpec = new
		// VariableAccessSpecification();
		// variableAccessSpec.selectVariableListName(objectName);

		VariableSpecification varSpec = new VariableSpecification(objectName, null, null, null, null);

		List<ListOfVariableSeq> listOfVariable = new LinkedList<ListOfVariableSeq>();
		ListOfVariableSeq varSeq = new ListOfVariableSeq(varSpec, null);
		// varSeq.setVariableSpecification(varSpec);
		listOfVariable.add(varSeq);

		// variableAccessSpec.selectListOfVariable(listOfVariable);
		VariableAccessSpecification variableAccessSpec = new VariableAccessSpecification(new SubSeqOf_listOfVariable(
				listOfVariable), null);

		List<Data> listOfData = new LinkedList<Data>();
		Data mmsData = MmsAsn1ComplexTypeConverter.convertModelNodeToMmsData(container);
		listOfData.add(mmsData);

		WriteRequest writeRequest = new WriteRequest(variableAccessSpec, new SubSeqOf_listOfData(listOfData));
		// writeRequest.setVariableAccessSpecification(variableAccessSpec);
		// writeRequest.setListOfData(listOfData);

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(null, null, writeRequest, null,
				null, null, null);
		// confirmedServiceRequest.selectWrite(writeRequest);

		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(getInvokeID()),
				confirmedServiceRequest);
		// confirmedRequestPdu.setInvokeID(new Unsigned32(Long
		// .valueOf(getInvokeID())));
		// confirmedRequestPdu.setConfirmedServiceRequest(confirmedServiceRequest);

		MmsPdu mms = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);
		// mms.selectConfirmedRequestPdu(confirmedRequestPdu);

		return mms;
	}

	// /*
	// * private static Data convertModelNodeToMmsData(IModelNode node) throws
	// * ServiceError { switch(node.getObjectType()) { case SERVER: //not
	// allowed,
	// * but possible case LOGICAL_DEVICE: //not allowed, but possible case
	// * LOGICAL_NODE: //not allowed, but possible case DATA_OBJECT: case
	// * CONSTRUCTED_DATA_ATTRIBUTE: case DATA_SET: Collection<Data> listOfData
	// =
	// * new LinkedList<Data>(); for(IModelNode child : node) {
	// * listOfData.add(convertModelNodeToMmsData(child)); } Data mmsStruct =
	// new
	// * Data(); mmsStruct.selectStructure(new DataSequence(listOfData)); return
	// * mmsStruct;
	// *
	// * case ARRAY_OF_DO: case ARRAY_OF_DA: Collection<Data> arrayOfData = new
	// * LinkedList<Data>(); for(IModelNode child : node) {
	// * arrayOfData.add(convertModelNodeToMmsData(child)); } Data mmsArray =
	// new
	// * Data(); mmsArray.selectArray(new DataSequence(arrayOfData)); return
	// * mmsArray;
	// *
	// * case BASIC_TYPE: return convertBasicNodeToMmsData(node);
	// *
	// * default: throw new
	// * ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
	// * "SetDataValues: unsupported ObjectType"); } }
	// *
	// * private static Data convertBasicNodeToMmsData(IModelNode node) {
	// * switch(node.getBasicType()) { //case } return null; }
	// */
	//
	private void handleSetDataValuesResponse(MmsPdu mmsRes) throws ServiceError {
		testForErrorResponse(mmsRes);
	}

	// Note: fc and dsMemberRef should correspond one to one
	// public void createDataSet(String dsRef, List<String> dsMemberRef,
	// List<FunctionalConstraint> fc) throws ServiceError {
	// // store the data set on client side
	// int at = dsRef.indexOf('@');
	// IModelNode mem;
	// FunctionalConstraint funcCon = null;
	// Iterator<String> itMem = dsMemberRef.iterator();
	// Iterator<FunctionalConstraint> itFC = fc.iterator();
	// String member;
	// LogicalNode lnNode = null;
	//
	// if (dsMemberRef.size() != fc.size()) {
	// throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
	// "createDataSet: the functional constraints do not match the list of members");
	// }
	//
	// DataSet dsTest = retrieveDataSet(dsRef);
	// if (dsTest != null) {
	// throw new ServiceError(ServiceError.INSTANCE_IN_USE,
	// "createDataSet: already have data set with same name");
	// }
	//
	// DataSet dataSet = new DataSet(dsRef);
	//
	// // get the members
	// while (itMem.hasNext()) {
	// funcCon = itFC.next();
	// member = itMem.next();
	// mem = findOrRetrieveNode(member);
	// if (mem == null) {
	// mem = getDataDefinition(member);
	// }
	// if (mem.getObjectType() == ObjectType.DATA_OBJECT) {
	// DataObject obj = new DataObject(mem.getReference().toString());
	// getFuncConstrainedDataObject(mem, funcCon, obj);
	// mem = obj;
	// }
	// dataSet.addChild((DataContainer) mem);
	// }
	// // at this point the data set is deletable
	// dataSet.setDeletable(true);
	//
	// if (at < 0) {
	// int dot = dsRef.indexOf('.');
	// lnNode = (LogicalNode) model.findNode(dsRef.substring(0, dot));
	// if (lnNode == null) {// getLogicalDeviceDirectory not called
	// throw new ServiceError(
	// ServiceError.PARAMETER_VALUE_INCONSISTENT,
	// "Logical node not in model: call getLogicalDeviceDirectory");
	// }
	// lnNode.addDataSet(dataSet);
	// } else {
	// nonPersistentDS.add(dataSet);
	// }
	//
	// // send mms msg to server to create the data set
	// MmsPdu mmsReq = constructCreateDataSetRequest(dsRef, dataSet);
	// MmsPdu mmsRes = encodeWriteReadDecode(mmsReq);
	// handleCreateDataSetResponse(mmsRes);
	// }

	/*
	 * dsRef = either LD/LN.DataSetName (persistent) or @DataSetname
	 * (non-persistent) Names in dsMemberRef should be in the form:
	 * LD/LNName.DoName or LD/LNName.DoName.DaName
	 */
	// private MmsPdu constructCreateDataSetRequest(String dsRef, DataSet
	// dataSet)
	// throws ServiceError {
	// // construct object name
	// ObjectName dataSetObjName = MMSReference.toMmsObjectNameDS(dsRef);
	//
	// // set the list of variables
	// Collection<SequenceType> listOfVariable = new LinkedList<SequenceType>();
	// FunctionalConstraint fc = null;
	//
	// for (DataContainer dc : dataSet.getChildren()) {
	// // find fc for this object
	// if (dc.getObjectType() == ObjectType.DATA_OBJECT) {
	// fc = dc.getChildren().get(0).getFunctionalConstraint();
	// // since DO does not have a FC, get it from child
	// } else {
	// fc = dc.getFunctionalConstraint();
	// }
	//
	// ObjectName dsMember = MMSReference.toMmsObjectName(dc
	// .getReference().toString(), fc);
	// VariableSpecification memVarSpec = new VariableSpecification();
	// memVarSpec.selectName(dsMember);
	//
	// SequenceType memSeqTy = new SequenceType();
	// memSeqTy.setVariableSpecification(memVarSpec);
	// listOfVariable.add(memSeqTy);
	// }
	//
	// // construct the DefineNamedVariableListRequest
	// DefineNamedVariableListRequest createDSRequest = new
	// DefineNamedVariableListRequest();
	// createDSRequest.setVariableListName(dataSetObjName);
	// createDSRequest.setListOfVariable(listOfVariable);
	//
	// ConfirmedServiceRequest confirmedServiceRequest = new
	// ConfirmedServiceRequest();
	// confirmedServiceRequest.selectDefineNamedVariableList(createDSRequest);
	//
	// ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu();
	// confirmedRequestPdu.setInvokeID(new Unsigned32(Long
	// .valueOf(getInvokeID())));
	// confirmedRequestPdu.setConfirmedServiceRequest(confirmedServiceRequest);
	//
	// MmsPdu mms = new MmsPdu();
	// mms.selectConfirmedRequestPdu(confirmedRequestPdu);
	//
	// return mms;
	//
	// }

	// private void handleCreateDataSetResponse(MmsPdu responsePdu) throws
	// ServiceError {
	// // check for errors
	// testForErrorResponse(responsePdu);
	// }
	//
	// // dsRef should be of form LD/LN.DataSetName or @DataSetName
	// public List<String> getDataSetDirectory(String dsRef) throws ServiceError
	// {
	// // check if data set exists
	// DataSet dataSet = retrieveDataSet(dsRef);
	// if (dataSet == null) {
	// // is there a case where the client side knows a dataset without
	// // requesting getLogicalNodeDirectory first?
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "Data set does not exist, call getLogicalNodeDirectory first");
	// }
	//
	// if (!dataSet.getDataObjects().isEmpty()) {
	// List<String> members = new LinkedList<String>();
	//
	// for (DataContainer mem : dataSet.getDataObjects()) {
	// members.add(mem.getReference().toString());
	// }
	//
	// return members;
	//
	// }
	//
	// else {
	// MmsPdu mmsReq = constructGetDataSetDirectory(dsRef);
	// MmsPdu mmsRes = encodeWriteReadDecode(mmsReq);
	// return decodeGetDataSetDirectoryResponse(mmsRes, dsRef);
	// }
	//
	// }
	//
	// private MmsPdu constructGetDataSetDirectory(String dsRef) throws
	// ServiceError {
	//
	// ObjectName dataSetObj = MMSReference.toMmsObjectNameDS(dsRef);
	// GetNamedVariableListAttributesRequest getDataSetDirectory = new
	// GetNamedVariableListAttributesRequest();
	// getDataSetDirectory.setValue(dataSetObj);
	//
	// ConfirmedServiceRequest confirmedServiceRequest = new
	// ConfirmedServiceRequest();
	// confirmedServiceRequest.selectGetNamedVariableListAttributes(getDataSetDirectory);
	//
	// ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu();
	// confirmedRequestPdu.setInvokeID(new
	// Unsigned32(Long.valueOf(getInvokeID())));
	// confirmedRequestPdu.setConfirmedServiceRequest(confirmedServiceRequest);
	//
	// MmsPdu mms = new MmsPdu();
	// mms.selectConfirmedRequestPdu(confirmedRequestPdu);
	//
	// return mms;
	//
	// }
	//
	// private List<String> decodeGetDataSetDirectoryResponse(MmsPdu
	// responsePdu, String dsRef) throws ServiceError {
	// // check response msg
	// testForErrorResponse(responsePdu);
	//
	// if (!responsePdu.isConfirmedResponsePduSelected()) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "decodeGetDataSetDirectoryResponse: Error decoding server response");
	// }
	//
	// ConfirmedResponsePdu confirmedResponsePdu =
	// responsePdu.getConfirmedResponsePdu();
	// checkInvokeID(confirmedResponsePdu);
	//
	// ConfirmedServiceResponse confirmedServiceResponse =
	// confirmedResponsePdu.getConfirmedServiceResponse();
	// if (!confirmedServiceResponse.isGetNamedVariableListAttributesSelected())
	// {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "decodeGetDataSetDirectoryResponse: Error decoding server response");
	// }
	//
	// GetNamedVariableListAttributesResponse getNamedVariableListAttResponse =
	// confirmedServiceResponse
	// .getGetNamedVariableListAttributes();
	//
	// boolean deletable = getNamedVariableListAttResponse.getMmsDeletable();
	//
	// Collection<GetNamedVariableListAttributesResponse.SequenceType> variables
	// = getNamedVariableListAttResponse
	// .getListOfVariable();
	//
	// if (variables.size() == 0) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "decodeGetDataSetDirectoryResponse: Instance not available");
	// }
	//
	// // find data set
	// DataSet dataSet = retrieveDataSet(dsRef);
	//
	// // add members to data set
	// VariableSpecification memVS;
	// DataContainer member;
	// List<String> memList = new LinkedList<String>();
	// String memberRef;
	// FunctionalConstraint fc;
	//
	// for (GetNamedVariableListAttributesResponse.SequenceType var : variables)
	// {
	// memVS = var.getVariableSpecification();
	// memberRef = MMSReference.toObjectReference(memVS.getName());
	// fc =
	// MMSReference.getFunctionalConstraint(memVS.getName().getDomain_specific().getItemId().getValue());
	//
	// member = (DataContainer) findOrRetrieveNode(memberRef);
	// if (member == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "decodeGetDataSetDirectoryResponse: data set memeber does not exist, call getDataDefinition first");
	// }
	//
	// if (member.getObjectType() == ObjectType.DATA_OBJECT) {
	// DataObject obj = new DataObject(member.getReference().toString());
	// getFuncConstrainedDataObject(member, fc, obj);
	// member = obj;
	// }
	//
	// dataSet.addChild(member);
	// memList.add(memberRef);
	// }
	//
	// dataSet.setDeletable(deletable);
	//
	// return memList;
	// }
	//
	// // int[] has format: int[0] = numberMatched, int[1] = numberDeleted
	// // make the response more intuitive??
	// public int[] deleteDataSet(String dataSetRef) throws ServiceError {
	// // first find dataSet in model
	// int at = dataSetRef.indexOf('@');
	// DataSet dataSet = null;
	//
	// if (at < 0) {
	// int dot = dataSetRef.indexOf('.');
	// LogicalNode lnNode = (LogicalNode)
	// findOrRetrieveNode(dataSetRef.substring(0, dot));
	// if (lnNode == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "cannot find logical node");
	// }
	// dataSet = lnNode.getDataSet(dataSetRef.substring(dot + 1));
	// if (dataSet == null || !dataSet.isDeletable()) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "Data set does not exist or cannot be deleted");
	// }
	// lnNode.deleteDataSet(dataSet.getNodeName());
	// }
	//
	// else {
	// for (DataSet ds : nonPersistentDS) {
	// if (dataSetRef.equals(ds.getReference().toString())) {
	// if (!ds.isDeletable()) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "Data set does not exist or cannot be deleted");
	// }
	// dataSet = ds;
	// nonPersistentDS.remove(ds);
	// break;
	// }
	// }
	//
	// if (dataSet == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "Data set does not exist or cannot be deleted");
	// }
	// }
	//
	// // delete the data set on the server side
	// MmsPdu request = constructDeleteDataSetRequest(dataSetRef);
	// MmsPdu response = encodeWriteReadDecode(request);
	// return decodeDeleteDataSet(response);
	// }
	//
	// private MmsPdu constructDeleteDataSetRequest(String dataSetRef) throws
	// ServiceError {
	// ObjectName objName = MMSReference.toMmsObjectNameDS(dataSetRef);
	// Collection<ObjectName> listOfVariableListName = new
	// LinkedList<ObjectName>();
	// // only deleting one element
	// listOfVariableListName.add(objName);
	//
	// DeleteNamedVariableListRequest requestDeleteDS = new
	// DeleteNamedVariableListRequest();
	// requestDeleteDS.setListOfVariableListName(listOfVariableListName);
	//
	// ConfirmedServiceRequest requestService = new ConfirmedServiceRequest();
	// requestService.selectDeleteNamedVariableList(requestDeleteDS);
	//
	// ConfirmedRequestPdu request = new ConfirmedRequestPdu();
	// request.setInvokeID(new Unsigned32(Long.valueOf(getInvokeID())));
	// request.setConfirmedServiceRequest(requestService);
	//
	// MmsPdu mmsRequest = new MmsPdu();
	// mmsRequest.selectConfirmedRequestPdu(request);
	//
	// return mmsRequest;
	// }
	//
	// private int[] decodeDeleteDataSet(MmsPdu responsePdu) throws ServiceError
	// {
	// testForErrorResponse(responsePdu);
	//
	// if (!responsePdu.isConfirmedResponsePduSelected()) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "decodeDeleteDataSetResponse: Error decoding server response");
	// }
	//
	// ConfirmedResponsePdu confirmedResponsePdu =
	// responsePdu.getConfirmedResponsePdu();
	// checkInvokeID(confirmedResponsePdu);
	//
	// ConfirmedServiceResponse confirmedServiceResponse =
	// confirmedResponsePdu.getConfirmedServiceResponse();
	// if (!confirmedServiceResponse.isDeleteNamedVariableListSelected()) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "decodeDeleteDataSetResponse: Error decoding server response");
	// }
	//
	// DeleteNamedVariableListResponse deleteNamedVariableListResponse =
	// confirmedServiceResponse
	// .getDeleteNamedVariableList();
	//
	// int[] numMatchedDeleted = new int[2];
	// numMatchedDeleted[0] =
	// deleteNamedVariableListResponse.getNumberMatched().getValue().intValue();
	// numMatchedDeleted[1] =
	// deleteNamedVariableListResponse.getNumberDeleted().getValue().intValue();
	//
	// return numMatchedDeleted;
	//
	// }
	//
	// public List<IModelNode> getDataSetValues(String dataSetRef) throws
	// ServiceError {
	// // always check server for updated values
	// // get the data set since only the values are read
	// DataSet dataSet = retrieveDataSet(dataSetRef);
	//
	// if (dataSet == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "getDataSetValues: data set does not exist");
	// }
	//
	// if (!dataSet.hasChildren()) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "getDataSetValues: data set has no children, call getDataSetDirectory first!!");
	// }
	//
	// MmsPdu request = constructGetDataSetValuesRequest(dataSetRef);
	// MmsPdu responsePdu = encodeWriteReadDecode(request);
	//
	// return decodeGetDataSetValuesResponse(responsePdu, dataSet);
	// }
	//
	// private MmsPdu constructGetDataSetValuesRequest(String dataSetRef) throws
	// ServiceError {
	// // TODO merge with constructGetDataValuesRequest??
	// ObjectName dsObjName = MMSReference.toMmsObjectNameDS(dataSetRef);
	//
	// VariableAccessSpecification varAccSpec = new
	// VariableAccessSpecification();
	// varAccSpec.selectVariableListName(dsObjName);
	//
	// ReadRequest getDataSetValuesRequest = new ReadRequest();
	//
	// getDataSetValuesRequest.setSpecificationWithResult(true);
	// // left as default of 'false' for getDataValues
	//
	// getDataSetValuesRequest.setVariableAccessSpecification(varAccSpec);
	//
	// ConfirmedServiceRequest requestService = new ConfirmedServiceRequest();
	// requestService.selectRead(getDataSetValuesRequest);
	//
	// ConfirmedRequestPdu request = new ConfirmedRequestPdu();
	// request.setInvokeID(new Unsigned32(Long.valueOf(getInvokeID())));
	// request.setConfirmedServiceRequest(requestService);
	//
	// MmsPdu mmsRequest = new MmsPdu();
	// mmsRequest.selectConfirmedRequestPdu(request);
	//
	// return mmsRequest;
	//
	// }
	//
	// private List<IModelNode> decodeGetDataSetValuesResponse(MmsPdu
	// responsePdu, DataSet ds) throws ServiceError {
	// // check response for error
	// testForErrorResponse(responsePdu);
	//
	// if (!responsePdu.isConfirmedResponsePduSelected()) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "Error decoding GetDataValuesReponsePdu");
	// }
	//
	// ConfirmedResponsePdu confirmedResponsePdu =
	// responsePdu.getConfirmedResponsePdu();
	// checkInvokeID(confirmedResponsePdu);
	//
	// ConfirmedServiceResponse confirmedServiceResponse =
	// confirmedResponsePdu.getConfirmedServiceResponse();
	// if (!confirmedServiceResponse.isReadSelected()) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "Error decoding GetDataValuesReponsePdu");
	// }
	//
	// ReadResponse readResponse = confirmedServiceResponse.getRead();
	// Collection<AccessResult> listOfAccessResults =
	// readResponse.getListOfAccessResult();
	//
	// if (listOfAccessResults.size() != ds.getDataObjects().size()) {
	// // should be one AccessResult for each data set memeber
	// throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
	// "Too many results received.");
	// }
	//
	// Iterator<AccessResult> it = listOfAccessResults.iterator();
	// AccessResult accRes = null;
	//
	// Iterator<DataContainer> dsIt = ds.getDataObjects().iterator();
	// DataContainer member = null;
	//
	// List<IModelNode> results = new LinkedList<IModelNode>();
	// int failures = 0;
	// FunctionalConstraint fc = null;
	//
	// while (it.hasNext()) {
	// accRes = it.next();
	// member = dsIt.next();
	// if (accRes.isSuccessSelected()) {
	// if (member.getObjectType() == ObjectType.DATA_OBJECT) {
	// for (DataContainer child : member.recursive()) {
	// if (child instanceof ConstructedDataAttribute &&
	// child.getFunctionalConstraint() != null) {
	// fc = child.getFunctionalConstraint();
	// break;
	// }
	// }
	//
	// // copy leaf DA into a temporary data object
	// // DataObject tempDO = new
	// // DataObject(member.getReference().toString());
	// // int count = 0;
	// // for (DataContainer mem : member.recursive()) {
	// // if (mem.isLeaf()) {
	// // tempDO.addDataAttribute(new
	// // DataAttribute(member.getReference().toString() +
	// // ".member"
	// // + count++, fc));
	// // }
	// // }
	//
	// results.add(MmsAsn1ComplexTypeConverter.convertMmsDataToDataContainer(member,
	// fc,
	// accRes.getSuccess()));
	//
	// // get results out of temporary DO and put into DO
	// // Iterator<DataContainer> iter =
	// // tempDO.getDataAttributes().iterator();
	// // for (DataContainer child : member.recursive()) {
	// // if (child.isLeaf()) {
	// // child.setValue(iter.next().getValue());
	// // System.out.println("Child: " +
	// // child.getReference().toString() + " Value: "
	// // + child.getValue().toString());
	// // }
	// // }
	//
	// }
	// else {
	// fc = member.getFunctionalConstraint();
	// results.add(MmsAsn1ComplexTypeConverter.convertMmsDataToDataContainer(member,
	// fc,
	// accRes.getSuccess()));
	// }
	//
	// }
	// else if (accRes.isFailureSelected()) {
	// failures++;
	// results.add(null);
	// // leave to client to handle?
	// }
	// }
	//
	// // indicate failure this way?
	// if (failures != 0) {
	// logger.error("decodeGetDataValuesResponse: " + failures +
	// " DataAccessError received.");
	// }
	//
	// return results;
	//
	// }
	//
	// private boolean getFuncConstrainedDataObject(IModelNode mem,
	// FunctionalConstraint fc, IModelNode obj)
	// throws ServiceError {
	//
	// // DataObject funcConstrainedDO = new
	// // DataObject(mem.getReference().toString());
	// // same name as original data object
	// boolean x = false;
	// boolean y = false;
	// for (DataContainer child : ((DataContainer) mem).getChildren()) {
	// if (child.getFunctionalConstraint() == fc) {
	// ((DataContainer) obj).addChild(child);
	// x = true;
	//
	// }
	// else if (child.getFunctionalConstraint() == null) {
	// ConstructedDataAttribute da = new
	// ConstructedDataAttribute(child.getReference().toString(), fc);
	// y = getFuncConstrainedDataObject(child, fc, da);
	// if (y == true) {
	// ((DataContainer) obj).addChild(da);
	//
	// }
	// }
	// }
	// return x || y;
	//
	// }
	//
	// // too tedious to fill out every DataAttributeValue? Consider revising
	// input
	// public List<String> setDataSetValues(String dsRef,
	// List<DataAttributeValue> setValues) throws ServiceError {
	// // setValues should contain all the DataAttributeValue in the order
	// // they occur for getDataSetValues
	//
	// DataSet dataSet = retrieveDataSet(dsRef);
	// if (dataSet == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "setDataSetValues: cannot find data set");
	// }
	// Iterator<DataAttributeValue> itValues = setValues.iterator();
	//
	// // set the values of the members themselves
	// setDataSetMembers(dataSet.getDataObjects(), itValues);
	//
	// // prepare, send msg and decode response
	// MmsPdu mmsRequest = constructSetDataSetValues(dataSet);
	// MmsPdu mmsResponse = encodeWriteReadDecode(mmsRequest);
	// return decodeSetDataSetValues(mmsResponse);
	//
	// }
	//
	// private void setDataSetMembers(List<DataContainer> member,
	// Iterator<DataAttributeValue> newValue)
	// throws ServiceError {
	//
	// DataAttributeValue test = null;
	// // TODO check if the number of values equals the number of leaft data
	// // attributes
	// // before changing values?
	// for (DataContainer mem : member) {
	//
	// for (DataContainer child : mem.recursive()) {
	// if (child.getObjectType() == ObjectType.BASIC_DATA_ATTRIBUTE) {
	// test = newValue.next();
	// child.setValue(test);
	// }
	// }
	// }
	// }
	//
	// private MmsPdu constructSetDataSetValues(DataSet dataSet) throws
	// ServiceError {
	// // insert data set name
	// ObjectName dsName =
	// MMSReference.toMmsObjectNameDS(dataSet.getReference().toString());
	// VariableAccessSpecification varAccSpec = new
	// VariableAccessSpecification();
	// varAccSpec.selectVariableListName(dsName);
	//
	// // construct list of data
	// Collection<Data> listOfData = new LinkedList<Data>();
	//
	// for (DataContainer member : dataSet.getDataObjects()) {
	// listOfData.add(MmsAsn1ComplexTypeConverter.convertModelNodeToMmsData(member,
	// member.getFunctionalConstraint()));
	// }
	//
	// WriteRequest setDataSetValuesRequest = new WriteRequest();
	// setDataSetValuesRequest.setVariableAccessSpecification(varAccSpec);
	// setDataSetValuesRequest.setListOfData(listOfData);
	//
	// ConfirmedServiceRequest requestService = new ConfirmedServiceRequest();
	// requestService.selectWrite(setDataSetValuesRequest);
	//
	// ConfirmedRequestPdu request = new ConfirmedRequestPdu();
	// request.setInvokeID(new Unsigned32(Long.valueOf(getInvokeID())));
	// request.setConfirmedServiceRequest(requestService);
	//
	// MmsPdu mmsRequest = new MmsPdu();
	// mmsRequest.selectConfirmedRequestPdu(request);
	//
	// return mmsRequest;
	//
	// }
	//
	// private List<String> decodeSetDataSetValues(MmsPdu responsePdu) throws
	// ServiceError {
	// testForErrorResponse(responsePdu);
	//
	// if (!responsePdu.isConfirmedResponsePduSelected()) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "Error decoding SetDataSetValuesReponsePdu");
	// }
	//
	// ConfirmedResponsePdu confirmedResponsePdu =
	// responsePdu.getConfirmedResponsePdu();
	// checkInvokeID(confirmedResponsePdu);
	//
	// ConfirmedServiceResponse confirmedServiceResponse =
	// confirmedResponsePdu.getConfirmedServiceResponse();
	// if (!confirmedServiceResponse.isWriteSelected()) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "Error decoding SetDataSetValuesReponsePdu");
	// }
	//
	// WriteResponse writeResponse = confirmedServiceResponse.getWrite();
	// Collection<WriteResponseChoiceType> writeResChoiceType =
	// writeResponse.getValue();
	// List<String> response = new ArrayList<String>(writeResChoiceType.size());
	//
	// for (WriteResponseChoiceType result : writeResChoiceType) {
	//
	// if (result.isFailureSelected()) {
	// response.add("Failure: reason = " +
	// result.getFailure().getValue().toString());
	// }
	// else {
	// response.add("Success");
	// }
	// }
	// return response;
	//
	// }
	//
	// private DataSet retrieveDataSet(String dsRef) throws ServiceError {
	// int at = dsRef.indexOf('@');
	// DataSet dataSet = null;
	//
	// if (at < 0) {
	// int dot = dsRef.indexOf('.');
	// LogicalNode lnNode = (LogicalNode) findOrRetrieveNode(dsRef.substring(0,
	// dot));
	// if (lnNode == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "cannot find logical node");
	// }
	// dataSet = lnNode.getDataSet(dsRef.substring(dot + 1));
	// }
	//
	// else {
	// for (DataSet ds : nonPersistentDS) {
	// if (dsRef.equals(ds.getReference().toString())) {
	// dataSet = ds;
	// }
	// }
	// }
	// return dataSet;
	// }

	public void close() {
		// TODO Auto-generated method stub

	}

	// public ModelNode setDataValues(ModelNode modelNode) throws ServiceError {
	// // TODO Auto-generated method stub
	// return null;
	// }

}
