/*
 * Copyright Fraunhofer ISE, energy & meteo Systems GmbH, and other contributors 2011
 *
 * This file is part of openIEC61850.
 * For more information visit http://www.openmuc.org 
 *
 * openIEC61850 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * openIEC61850 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with openIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openiec61850.scsm.mms.server;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import org.openiec61850.common.ACSIClass;
import org.openiec61850.common.ServiceError;
import org.openiec61850.common.model.Array;
import org.openiec61850.common.model.BasicDataAttribute;
import org.openiec61850.common.model.ConstructedDataAttribute;
import org.openiec61850.common.model.DataObject;
import org.openiec61850.common.model.DataSet;
import org.openiec61850.common.model.FunctionalConstraint;
import org.openiec61850.common.model.LogicalDevice;
import org.openiec61850.common.model.LogicalNode;
import org.openiec61850.common.model.ModelNode;
import org.openiec61850.common.model.ObjectReference;
import org.openiec61850.common.model.Server;
import org.openiec61850.common.model.report.BufferedReportControlBlock;
import org.openiec61850.common.model.report.OptFields;
import org.openiec61850.common.model.report.Report;
import org.openiec61850.common.model.report.ReportControlBlock;
import org.openiec61850.common.model.report.ReportEntryData;
import org.openiec61850.common.model.report.TriggerConditions;
import org.openiec61850.common.model.report.UnbufferedReportContrlBlock;
import org.openiec61850.jmms.iso.acse.AcseAssociation;
import org.openiec61850.jmms.mms.asn1.AccessResult;
import org.openiec61850.jmms.mms.asn1.ConfirmedRequestPdu;
import org.openiec61850.jmms.mms.asn1.ConfirmedResponsePdu;
import org.openiec61850.jmms.mms.asn1.ConfirmedServiceRequest;
import org.openiec61850.jmms.mms.asn1.ConfirmedServiceResponse;
import org.openiec61850.jmms.mms.asn1.Data;
import org.openiec61850.jmms.mms.asn1.DefineNamedVariableListRequest;
import org.openiec61850.jmms.mms.asn1.DeleteNamedVariableListRequest;
import org.openiec61850.jmms.mms.asn1.DeleteNamedVariableListResponse;
import org.openiec61850.jmms.mms.asn1.GetNameListRequest;
import org.openiec61850.jmms.mms.asn1.GetNameListResponse;
import org.openiec61850.jmms.mms.asn1.GetNameListResponse.SubSeqOf_listOfIdentifier;
import org.openiec61850.jmms.mms.asn1.GetNamedVariableListAttributesResponse;
import org.openiec61850.jmms.mms.asn1.GetNamedVariableListAttributesResponse.SubSeqOf_listOfVariable;
import org.openiec61850.jmms.mms.asn1.GetVariableAccessAttributesRequest;
import org.openiec61850.jmms.mms.asn1.GetVariableAccessAttributesResponse;
import org.openiec61850.jmms.mms.asn1.InformationReport;
import org.openiec61850.jmms.mms.asn1.InitResponseDetail;
import org.openiec61850.jmms.mms.asn1.InitiateRequestPdu;
import org.openiec61850.jmms.mms.asn1.InitiateResponsePdu;
import org.openiec61850.jmms.mms.asn1.ListOfVariableSeq;
import org.openiec61850.jmms.mms.asn1.MmsPdu;
import org.openiec61850.jmms.mms.asn1.ObjectName;
import org.openiec61850.jmms.mms.asn1.ObjectName.SubSeq_domain_specific;
import org.openiec61850.jmms.mms.asn1.ReadRequest;
import org.openiec61850.jmms.mms.asn1.ReadResponse;
import org.openiec61850.jmms.mms.asn1.ReadResponse.SubSeqOf_listOfAccessResult;
import org.openiec61850.jmms.mms.asn1.ServiceError.SubChoice_errorClass;
import org.openiec61850.jmms.mms.asn1.StructComponent;
import org.openiec61850.jmms.mms.asn1.TypeSpecification;
import org.openiec61850.jmms.mms.asn1.TypeSpecification.SubSeq_array;
import org.openiec61850.jmms.mms.asn1.TypeSpecification.SubSeq_floating_point;
import org.openiec61850.jmms.mms.asn1.TypeSpecification.SubSeq_structure;
import org.openiec61850.jmms.mms.asn1.TypeSpecification.SubSeq_structure.SubSeqOf_components;
import org.openiec61850.jmms.mms.asn1.UnconfirmedPDU;
import org.openiec61850.jmms.mms.asn1.UnconfirmedService;
import org.openiec61850.jmms.mms.asn1.VariableAccessSpecification;
import org.openiec61850.jmms.mms.asn1.VariableSpecification;
import org.openiec61850.jmms.mms.asn1.WriteRequest;
import org.openiec61850.jmms.mms.asn1.WriteResponse;
import org.openiec61850.scsm.mms.MMSConstants;
import org.openiec61850.scsm.mms.MMSReference;
import org.openiec61850.scsm.mms.MmsAsn1ComplexTypeConverter;
import org.openiec61850.scsm.mms.MmsAsn1PrimitiveTypeConverter;
import org.openiec61850.server.Association;
import org.openiec61850.server.ServerACSI;
import org.openiec61850.server.ServerACSI.ServerDirectoryType;
import org.openiec61850.server.scsm.SCSMConnectionHandler;
import org.openmuc.jasn1.ber.BerByteArrayOutputStream;
import org.openmuc.jasn1.ber.types.BerBitString;
import org.openmuc.jasn1.ber.types.BerBoolean;
import org.openmuc.jasn1.ber.types.BerInteger;
import org.openmuc.jasn1.ber.types.BerNull;
import org.openmuc.jasn1.ber.types.BerOctetString;
import org.openmuc.jasn1.ber.types.string.BerVisibleString;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Implements the SCSMConnectionHandler. One MMSConnectionHandler is created for
 * every new Association Request.
 * 
 * @author Stefan Feuerhahn
 * @author Tobias Weidelt
 * 
 */
public class MMSConnectionHandler implements SCSMConnectionHandler {

	private static final int DEFAULT_PROPOSED_MAX_SERV_OUTSTANDING_CALLING = 5;
	private static final int DEFAULT_PROPOSED_MAX_SERV_OUTSTANDING_CALLED = 5;
	private static final int DEFAULT_PROPOSED_DATA_STRUCTURE_NESTING_LEVEL = 10;
	private static final int DEFAULT_MAX_Pdu_SIZE = 65000;
	private static final int MINIMUM_Pdu_SIZE = 1000;

	private static Logger logger = LoggerFactory.getLogger(MMSConnectionHandler.class);

	private AcseAssociation pConnection = null;
	private boolean associationStarted = false;

	private int proposedMaxPduSize = DEFAULT_MAX_Pdu_SIZE;
	// TODO the following 3 parameters are ignored at the moment
	private final int proposedMaxServOutstandingCalling = DEFAULT_PROPOSED_MAX_SERV_OUTSTANDING_CALLING;
	private final int proposedMaxServOutstandingCalled = DEFAULT_PROPOSED_MAX_SERV_OUTSTANDING_CALLED;
	private final int proposedDataStructureNestingLevel = DEFAULT_PROPOSED_DATA_STRUCTURE_NESTING_LEVEL;

	private int negotiatedMaxServOutstandingCalling;
	private int negotiatedMaxServOutstandingCalled;
	private int negotiatedDataStructureNestingLevel;
	private long negotiatedMaxPduSize;

	private final ServerACSI serverACSI;

	// private static IEncoder<MmsPdu> iencoder;
	// private static IDecoder idecoder;

	private boolean notifyAssociationOfClosedSocket;

	private boolean insertRef;
	String continueAfter;

	// The following FCs are not part of this enum because they are not really
	// FCs and only defined in part 8-1:
	// RP (report), LG (log), BR (buffered report), GO, GS, MS, US

	private static String[] mmsFCs = { "MX", "ST", "CO", "CF", "DC", "SP", "SG", "RP", "LG", "BR", "GO", "GS", "SV",
			"SE", "EX", "SR", "OR", "BL" };

	public MMSConnectionHandler(ServerACSI serverACSI) {
		this.serverACSI = serverACSI;

		// try {
		// iencoder = CoderFactory.getInstance().newEncoder("BER");
		// idecoder = CoderFactory.getInstance().newDecoder("BER");
		// } catch (Exception e) {
		// throw new RuntimeException("Error creating encoder and decoder.");
		// }
	}

	/**
	 * Set the maximum MMS Pdu size in bytes. The default size is
	 * DEFAULT_MAX_Pdu_SIZE.
	 * 
	 * @param maxPduSize
	 *            should be at least MINIMUM_Pdu_SIZE
	 */
	public void setMaxPduSize(int maxPduSize) {
		if (associationStarted) {
			throw new RuntimeException("MaxPduSize cannot be changed once the association has been started");
		}
		if (maxPduSize >= MINIMUM_Pdu_SIZE) {
			this.proposedMaxPduSize = maxPduSize;
		}
		else {
			throw new IllegalArgumentException("Maximum size is too small");
		}
	}

	private MmsPdu constructAssociationResponsePdu() {

		// TODO what are the CBB Parameter?
		// byte[] negotiatedParameterCbbBitString = new byte[] { (byte) (0xf1 &
		// 0xff), 0x00 };

		byte[] negotiatedParameterCbbBitString = new byte[] { (byte) (0xfb), 0x00 };

		// byte[] servicesSupportedCalledBitString = new byte[] { 0x40, 0x00,
		// 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		// 0x00, 0x00 };
		// set getNameList

		byte[] servicesSupportedCalledBitString = new byte[] { (byte) 0xee, (byte) 0x1c, 0x00, 0x00, 0x04, 0x08, 0x00,
				0x00, 0x79, (byte) 0xef, 0x18 };

		InitResponseDetail initRespDetail = new InitResponseDetail(new BerInteger(1), new BerBitString(
				negotiatedParameterCbbBitString, negotiatedParameterCbbBitString.length * 8 - 5), new BerBitString(
				servicesSupportedCalledBitString, servicesSupportedCalledBitString.length * 8 - 3));

		InitiateResponsePdu initRespPdu = new InitiateResponsePdu(new BerInteger(negotiatedMaxPduSize), new BerInteger(
				negotiatedMaxServOutstandingCalling), new BerInteger(negotiatedMaxServOutstandingCalled),
				new BerInteger(negotiatedDataStructureNestingLevel), initRespDetail);

		// InitResponseDetail initResponseDetail = new InitResponseDetail();
		// initResponseDetail.setNegotiatedVersionNumber(new Integer16(2));
		// initResponseDetail.setNegotiatedParameterCBB(new
		// ParameterSupportOptions(new BitString(
		// negotiatedParameterCbbBitString, 5)));
		// initResponseDetail.setServicesSupportedCalled(new
		// ServiceSupportOptions(new BitString(
		// servicesSupportedCalledBitString, 3)));

		// InitiateResponsePdu initiateResponsePdu = new InitiateResponsePdu();
		// initiateResponsePdu.setLocalDetailCalled(new
		// Integer32(negotiatedMaxPduSize));
		// initiateResponsePdu.setNegotiatedMaxServOutstandingCalling(new
		// Integer16(negotiatedMaxServOutstandingCalling));
		// initiateResponsePdu.setNegotiatedMaxServOutstandingCalled(new
		// Integer16(negotiatedMaxServOutstandingCalled));
		// initiateResponsePdu.setNegotiatedDataStructureNestingLevel(new
		// Integer8(negotiatedDataStructureNestingLevel));
		// initiateResponsePdu.setMmsInitResponseDetail(initResponseDetail);

		MmsPdu initiateResponseMMSpdu = new MmsPdu(null, null, null, null, initRespPdu, null);

		// MmsPdu initiateResponseMMSpdu = new MmsPdu();
		// initiateResponseMMSpdu.selectInitiateResponsePdu(initiateResponsePdu);

		return initiateResponseMMSpdu;
	}

	private boolean processAssociationRequest(InitiateRequestPdu associationRequestMMSpdu) {
		// if (!associationRequestMMSpdu.isInitiateRequestPduSelected()) {
		// return false;
		// }

		if (associationRequestMMSpdu.localDetailCalling != null) {
			negotiatedMaxPduSize = (int) associationRequestMMSpdu.localDetailCalling.val;
		}
		if (negotiatedMaxPduSize > proposedMaxPduSize) {
			negotiatedMaxPduSize = proposedMaxPduSize;
		}
		else if (negotiatedMaxPduSize < MINIMUM_Pdu_SIZE) {
			negotiatedMaxPduSize = MINIMUM_Pdu_SIZE;
		}

		negotiatedMaxServOutstandingCalling = (int) associationRequestMMSpdu.proposedMaxServOutstandingCalling.val;

		if (negotiatedMaxServOutstandingCalling > proposedMaxServOutstandingCalling) {
			negotiatedMaxServOutstandingCalling = proposedMaxServOutstandingCalling;
		}

		negotiatedMaxServOutstandingCalled = (int) associationRequestMMSpdu.proposedMaxServOutstandingCalled.val;

		if (negotiatedMaxServOutstandingCalled > proposedMaxServOutstandingCalled) {
			negotiatedMaxServOutstandingCalled = proposedMaxServOutstandingCalled;
		}

		if (associationRequestMMSpdu.proposedDataStructureNestingLevel != null) {
			negotiatedDataStructureNestingLevel = (int) associationRequestMMSpdu.proposedDataStructureNestingLevel.val;

		}
		if (negotiatedDataStructureNestingLevel > proposedDataStructureNestingLevel) {
			negotiatedDataStructureNestingLevel = proposedDataStructureNestingLevel;
		}

		if (negotiatedMaxServOutstandingCalling < 0 || negotiatedMaxServOutstandingCalled < 0
				|| negotiatedDataStructureNestingLevel < 0) {
			return false;
		}

		return true;

	}

	public void connectionIndication(AcseAssociation pConnection, ByteBuffer associationRequest) throws ServiceError,
			IOException {

		logger.debug("someone connected");

		this.pConnection = pConnection;

		Association association = associate(pConnection, associationRequest);
		if (association == null) {
			return;
		}
		associationStarted = true;

		// add the pConnection with the client ID here?

		try {
			handleConnection(pConnection, association);
		} finally {
			associationStarted = false;
		}
	}

	private void handleConnection(AcseAssociation pConnection, Association association) throws IOException {
		while (true) {
			try {
				MmsPdu mmsRequestPdu = readAnMmsPdu(pConnection, association);
				if (mmsRequestPdu == null) {
					return;
				}

				// if (!mmsRequestPdu.isConfirmedRequestPduSelected()) {
				if (mmsRequestPdu.confirmedRequestPdu == null) {
					throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
							"Got an invalid MMS packet: is not an ConfirmedRequestPdu");
				}

				ConfirmedRequestPdu confirmedRequestPdu = mmsRequestPdu.confirmedRequestPdu;
				ConfirmedServiceRequest confirmedServiceRequest = confirmedRequestPdu.confirmedServiceRequest;

				ConfirmedServiceResponse confirmedServiceResponse = null;

				if (confirmedServiceRequest.getNameList != null) {

					// GetNameListRequest getNameListRequest =
					// confirmedServiceRequest.getGetNameList();
					// GetNameListResponse response;
					GetNameListRequest getNameListRequest = confirmedServiceRequest.getNameList;
					GetNameListResponse response = null;

					// in other MMS implementations BasicObjectClass is
					// sometimes named extendedObjectClass
					// if
					// (!getNameListRequest.getObjectClass().isBasicObjectClassSelected())
					// {
					if (getNameListRequest.objectClass.basicObjectClass == null) {
						throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
								"Got an invalid MMS packet: ObjectClass was not selected in GetNameList request");
					}

					// long basicObjectClass =
					// getNameListRequest.getObjectClass().getBasicObjectClass();
					long basicObjectClass = getNameListRequest.objectClass.basicObjectClass.val;
					if (basicObjectClass == MMSConstants.DOMAIN) {
						logger.debug("Got a GetServerDirectory (MMS GetNameList[DOMAIN]) request");
						response = handleGetServerDirectoryRequest(getNameListRequest, association);
					}
					else if (basicObjectClass == MMSConstants.NAMED_VARIABLE) {
						logger.debug("Got a Get{LD|LN}Directory (MMS GetNameList[NAMED_VARIABLE]) request");
						response = handleGetDirectoryRequest(getNameListRequest, association);
					}
					else if (basicObjectClass == MMSConstants.NAMED_VARIABLE_LIST) {
						logger.debug("Got a GetLogicalNodeDirectory[DataSet] (MMS GetNameList[NAMED_VARIABLE_LIST]) request");
						response = handleGetDataSetListRequest(getNameListRequest, association);
					}
					else if (basicObjectClass == MMSConstants.JOURNAL) {
						logger.debug("Got a GetLogicalNodeDirectory[Log] (MMS GetNameList[JOURNAL]) request");
						response = handleGetNameListJournalRequest(getNameListRequest);
						// TODO
					}

					confirmedServiceResponse = new ConfirmedServiceResponse(response, null, null, null, null, null,
							null);

					// // confirmedServiceResponse.selectGetNameList(response);
					//
				}
				// else if
				// (confirmedServiceRequest.isGetVariableAccessAttributesSelected())
				// {
				else if (confirmedServiceRequest.getVariableAccessAttributes != null) {
					logger.debug("Got a GetVariableAccessAttributes request (GetDataDirectory/GetDataDefinition)");
					GetVariableAccessAttributesResponse response = handleGetVariableAccessAttributesRequest(

					confirmedServiceRequest.getVariableAccessAttributes, association);
					confirmedServiceResponse = new ConfirmedServiceResponse(null, null, null, response, null, null,
							null);

					//
					// confirmedServiceResponse.selectGetVariableAccessAttributes(response);
				}
				// // else if (confirmedServiceRequest.isReadSelected()) {
				else if (confirmedServiceRequest.read != null) {
					// both getDataValues and getDataSetValues map to this
					ReadResponse response = handleGetDataValuesRequest(confirmedServiceRequest.read, association);

					confirmedServiceResponse = new ConfirmedServiceResponse(null, response, null, null, null, null,
							null);
					// // confirmedServiceResponse.selectRead(response);
				}
				// // else if (confirmedServiceRequest.isWriteSelected()) {
				else if (confirmedServiceRequest.write != null) {
					logger.debug("Got a Write request");
					// // confirmedServiceResponse = null;
					WriteResponse response = handleWriteRequest(confirmedServiceRequest.write, association);

					confirmedServiceResponse = new ConfirmedServiceResponse(null, null, response, null, null, null,
							null);

					// confirmedServiceResponse.selectWrite(response);

				}

				// // for Data Sets
				// else if (confirmedServiceRequest
				// .isDefineNamedVariableListSelected()) {
				else if (confirmedServiceRequest.defineNamedVariableList != null) {
					logger.debug("Got a CreateDataSet request");

					BerNull response = handleDefineNamedVariableListRequest(
							confirmedServiceRequest.defineNamedVariableList, association);

					// confirmedServiceResponse = new
					// ConfirmedServiceResponse();
					// confirmedServiceResponse
					// .selectDefineNamedVariableList(response);

					confirmedServiceResponse = new ConfirmedServiceResponse(null, null, null, null, response, null,
							null);
				}

				else if (confirmedServiceRequest.getNamedVariableListAttributes != null) {
					logger.debug("Got a GetDataSetDirectory request");
					GetNamedVariableListAttributesResponse response = handleGetNamedVariableListAttritubes(
							confirmedServiceRequest.getNamedVariableListAttributes, association);

					confirmedServiceResponse = new ConfirmedServiceResponse(null, null, null, null, null, response,
							null);

				}

				else if (confirmedServiceRequest.deleteNamedVariableList != null) {
					logger.debug("Got a DeleteDataSet request");
					DeleteNamedVariableListResponse response = handleDeleteNamedVariableListRequest(
							confirmedServiceRequest.deleteNamedVariableList, association);
					// confirmedServiceResponse = new
					// ConfirmedServiceResponse();
					// confirmedServiceResponse
					// .selectDeleteNamedVariableList(response);
					confirmedServiceResponse = new ConfirmedServiceResponse(null, null, null, null, null, null,
							response);
				}

				else {
					throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
							"invalid MMS packet: unknown request type.");
				}

				ConfirmedResponsePdu confirmedResponsePdu = new ConfirmedResponsePdu(confirmedRequestPdu.invokeID,
						confirmedServiceResponse);
				// confirmedResponsePdu.setInvokeID(confirmedRequestPdu.getInvokeID());
				//
				// confirmedResponsePdu.setConfirmedServiceResponse(confirmedServiceResponse);

				// MmsPdu mmsResponsePdu = new MmsPdu();
				// mmsResponsePdu.selectConfirmedResponsePdu(confirmedResponsePdu);

				MmsPdu mmsResponsePdu = new MmsPdu(null, confirmedResponsePdu, null, null, null, null);
				sendAnMmsPdu(pConnection, mmsResponsePdu);
			}

			catch (ServiceError e) {
				logger.warn(e.getMessage());
				sendAnMmsPdu(pConnection, createServiceErrorResponse(e));
			}
		}
	}

	private Association associate(AcseAssociation pConnection, ByteBuffer associationRequest) throws ServiceError {

		MmsPdu mmsPdu = new MmsPdu();
		// ByteArrayInputStream iStream = new ByteArrayInputStream(
		// associationRequest);
		ByteArrayInputStream iStream = new ByteArrayInputStream(associationRequest.array(),
				associationRequest.arrayOffset() + associationRequest.position(), associationRequest.limit()
						- associationRequest.position());
		try {
			mmsPdu.decode(iStream, null);
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
			throw new ServiceError(0);
		}

		// // Decode the Association Request:
		// MmsPdu associationRequestMMSpdu = null;
		// try {
		// associationRequestMMSpdu = idecoder.decode(new
		// ByteArrayInputStream(associationRequest), MmsPdu.class);
		// } catch (Exception e) {
		// pConnection.associationMalformed();
		// }

		Association association = null;

		// TODO association needs to be processed
		if (!processAssociationRequest(mmsPdu.initiateRequestPdu)) {
			throw new ServiceError(ServiceError.FATAL);
			// pConnection.associationMalformed();
		}
		else {
			association = serverACSI.associate(this, new String(""));

			MmsPdu initiateResponseMMSpdu = constructAssociationResponsePdu();

			// byte[] initiateResponsePdu =
			// encodeMMSpdu(initiateResponseMMSpdu);

			BerByteArrayOutputStream berOStream = new BerByteArrayOutputStream(500);
			try {
				initiateResponseMMSpdu.encode(berOStream, true);
			} catch (IOException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
				throw new ServiceError(0);
			}
			// byte[] initiateResponsePdu = berOStream.getArray();

			try {
				// pConnection.accept(berOStream.buffer, berOStream.index + 1,
				// berOStream.buffer.length
				// - (berOStream.index + 1));
				pConnection.accept(berOStream.getByteBuffer());
			} catch (IOException e) {
				pConnection.close();
				association = null;
			}
		}

		return association;
	}

	private void sendAnMmsPdu(AcseAssociation pConnection, MmsPdu mmsResponsePdu) throws IOException {
		try {
			// byte[] mmsResponsePdubytes = encodeMMSpdu(mmsResponsePdu);
			BerByteArrayOutputStream baos = new BerByteArrayOutputStream(500, true);
			mmsResponsePdu.encode(baos, false);
			// pConnection.send(baos.getArray());
			pConnection.send(baos.getByteBuffer());
		} catch (IOException e) {
			pConnection.close();
			throw e;
		}
	}

	private MmsPdu readAnMmsPdu(AcseAssociation pConnection, Association association) throws ServiceError {

		// MmsPdu mmsRequestPdu = null;
		MmsPdu mmsRequestPdu = null;
		try {
			ByteBuffer rcvdPdu = pConnection.receive();
			// if (rcvdPdu.length == 2 && rcvdPdu[0] == -117 && rcvdPdu[1] == 0)
			// {
			// logger.debug("Conclude-RequestPdu received. Closing Connection...");
			// }
			// else {
			// mmsRequestPdu = idecoder.decode(new
			// ByteArrayInputStream(rcvdPdu), MmsPdu.class);
			// mmsRequestPdu.decode(new ByteArrayInputStream(rcvdPdu), new
			// BerIdentifier(BerIdentifier.CONTEXT_CLASS,
			// BerIdentifier.CONSTRUCTED, 0));
			mmsRequestPdu = new MmsPdu();
			mmsRequestPdu.decode(new ByteArrayInputStream(rcvdPdu.array(), rcvdPdu.arrayOffset() + rcvdPdu.position(),
					rcvdPdu.limit() - rcvdPdu.position()), null);
			// }
		} catch (IOException e) {
			if (notifyAssociationOfClosedSocket) {
				serverACSI.abort(association, "Socket was closed.");
			}
			logger.debug("IOException reading, client probably closed socket. closing ConnectionHandler. {}");
			pConnection.close();
		} catch (Exception e) {
			throw new RuntimeException(e);
			// throw new
			// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
			// "1Got an invalid MMS packet", e);
		}
		return mmsRequestPdu;
	}

	private WriteResponse handleWriteRequest(WriteRequest mmsWriteRequest, Association association) throws ServiceError {

		// VariableAccessSpecification variableAccessSpecification =
		// mmsWriteRequest
		// .getVariableAccessSpecification();

		VariableAccessSpecification variableAccessSpecification = mmsWriteRequest.variableAccessSpecification;

		ObjectReference reference = null;
		FunctionalConstraint fc = null;

		List<Data> listOfData = mmsWriteRequest.listOfData.seqOf;
		// List<WriteResponseChoiceType> mmsReponseValues = new
		// ArrayList<WriteResponseChoiceType>(
		// listOfData.size());

		List<WriteResponse.SubChoice> mmsResponseValues = new ArrayList<WriteResponse.SubChoice>(listOfData.size());

		// if (variableAccessSpecification.isListOfVariableSelected()) {
		if (variableAccessSpecification.listOfVariable != null) {
			List<ListOfVariableSeq> varSeq = variableAccessSpecification.listOfVariable.seqOf;

			// case of more than one? -> this works for only one
			// for (ListOfVariableSeq vs :
			// varSeq) {
			VariableSpecification varSpec = varSeq.get(0).variableSpecification;
			reference = (MMSReference.toObjectReference(varSpec.name.domain_specific.domainId.toString(),
					varSpec.name.domain_specific.itemId.toString()));

			fc = MMSReference.getFunctionalConstraint(varSpec.name.domain_specific.itemId.toString());
			// }

			for (Data mmsData : listOfData) {
				// WriteResponseChoiceType result = new
				// WriteResponseChoiceType();
				WriteResponse.SubChoice result = null;
				try {

					if (fc == null) {

						ReportControlBlock rcb = (ReportControlBlock) serverACSI.getRCB(association,
								new ObjectReference(reference.getParentRef()));
						ModelNode container = rcb.getChild(reference.getName()).copy();

						// if (container instanceof DataSet) {
						// // switched to a different existing data set
						// // ModelNode lnNode = serverACSI
						// // .getDataDefinition(
						// // association,
						// // new
						// //
						// ObjectReference(mmsData.getVisible_string().substring(0,mmsData.getVisible_string().lastIndexOf("."))),
						// // null);
						//
						// ModelNode lnNode = serverACSI.getDataDefinition(
						// association,
						// new
						// ObjectReference(mmsData.visible_string.toString().substring(0,
						// mmsData.visible_string.toString().lastIndexOf("."))),
						// null);
						//
						// container = ((LogicalNode) lnNode).getDataSet(
						// mmsData.visible_string.toString().substring(
						// mmsData.visible_string.toString().lastIndexOf(".") +
						// 1)).copy();
						// }
						// else {

						MmsAsn1ComplexTypeConverter.convertMmsDataToDataContainer(container, null, mmsData, mmsFCs);
						// }

						if (rcb instanceof UnbufferedReportContrlBlock) {

							serverACSI.setURCBValues(association, (UnbufferedReportContrlBlock) rcb, container);
						}
						else {
							// TODO for buffered report blocks
						}
					}
					else {
						ModelNode container = serverACSI.getDataDefinition(association, reference, fc).copy();
						MmsAsn1ComplexTypeConverter.convertMmsDataToDataContainer(container, fc, mmsData, mmsFCs);
						serverACSI.setDataValues(association, container);
					}
					result = new WriteResponse.SubChoice(null, new BerNull());
				} catch (ServiceError e) {
					logger.warn(e.getMessage());
					// result.selectFailure(createMmsDataAccessError(e));
					// correct error code?
					result = new WriteResponse.SubChoice(new BerInteger(0L), null);
				}
				mmsResponseValues.add(result);
			}
		}
		else if (variableAccessSpecification.variableListName != null) {
			// data set
			reference = MMSReference.toObjectReference(
					variableAccessSpecification.variableListName.domain_specific.domainId.toString(),
					variableAccessSpecification.variableListName.domain_specific.itemId.toString());

			ModelNode dataSet = serverACSI.getDataSetDirectory(association, reference.toString());
			fc = MMSReference
					.getFunctionalConstraint(variableAccessSpecification.variableListName.domain_specific.itemId
							.toString());

			List<ModelNode> members = ((DataSet) dataSet).getChildren();
			int index = 0;
			ModelNode dd = null;

			List<ModelNode> writeValues = new ArrayList<ModelNode>();

			for (Data mmsData : listOfData) {
				dd = members.get(index).copy();
				index++;
				MmsAsn1ComplexTypeConverter.convertMmsDataToDataContainer(dd, fc, mmsData, mmsFCs);
				writeValues.add(dd);
			}

			List<ServiceError> results = serverACSI.setDataSetValues(association, writeValues);

			for (ServiceError e : results) {
				if (e.equals(ServiceError.NO_ERROR)) {
					WriteResponse.SubChoice result = new WriteResponse.SubChoice(null, new BerNull());
					mmsResponseValues.add(result);
				}

				else {

					WriteResponse.SubChoice result = new WriteResponse.SubChoice(new BerInteger(0L), null);
					mmsResponseValues.add(result);
				}
			}

			// the node is passed onto the Data Factory -> needs to
			// handle both data objects and data attributes!
			// serverACSI.setDataValues(association, container);
			// result.selectSuccess();

			// } catch (ServiceError e) {
			// logger.warn(e.getMessage());
			// result.selectFailure(createMmsDataAccessError(e));
			// }
			// mmsReponseValues.add(result);
			// }
		}
		else {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
					"handleWriteRequest: invalid MMS request");
		}

		return new WriteResponse(mmsResponseValues);
	}

	private GetNameListResponse handleGetNameListJournalRequest(GetNameListRequest getNameListRequest) {
		// TODO liefert momentan leere Liste zurück:

		Vector<BerVisibleString> identifiers = new Vector<BerVisibleString>();
		SubSeqOf_listOfIdentifier listOf = new SubSeqOf_listOfIdentifier(identifiers);

		return new GetNameListResponse(listOf, new BerBoolean(false));
	}

	private GetNameListResponse handleGetDataSetListRequest(GetNameListRequest getNameListRequest,
			Association association) throws ServiceError {

		List<String> dsList = new LinkedList<String>();
		ModelNode ldDD = null;

		if (getNameListRequest.objectScope.domainSpecific != null) {

			String mmsDomainId = getNameListRequest.objectScope.domainSpecific.toString();

			if (mmsDomainId == "") {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
						"handleGetDataSetListRequest: logical node invalid");
			}

			ldDD = serverACSI.getDataDefinition(association, new ObjectReference(mmsDomainId), null);

			for (ModelNode ln : ldDD) {
				List<String> dsNames = serverACSI.getLogicalNodeDirectory(association, ln.getReference().toString(),
						ACSIClass.DATA_SET);
				for (String s : dsNames) {
					dsList.add(ln.getNodeName() + "$" + s);
				}
			}
		}

		insertRef = true;
		if (getNameListRequest.continueAfter != null) {
			continueAfter = getNameListRequest.continueAfter.toString();
			insertRef = false;
		}

		Vector<BerVisibleString> identifiers = new Vector<BerVisibleString>();
		BerVisibleString identifier;

		int identifierSize = 0;
		boolean moreFollows = false;

		if (dsList != null) {
			for (String dsRef : dsList) {
				if (insertRef == true) {
					if (identifierSize > negotiatedMaxPduSize - 200) {
						moreFollows = true;
						logger.info("---maxMMSPduSize--reached--");
						break;
					}
					try {
						identifier = new BerVisibleString(dsRef);
					} catch (UnsupportedEncodingException e) {
						logger.warn(e.getMessage());
						throw new IllegalAccessError("received an UnsupportedEncodingException, terminating");
					}
					identifiers.add(identifier);
					identifierSize += dsRef.length() + 2;
				}
				else {
					if (dsRef.equals(continueAfter)) {
						insertRef = true;
					}
				}
			}
		}

		SubSeqOf_listOfIdentifier listOf = new SubSeqOf_listOfIdentifier(identifiers);

		return new GetNameListResponse(listOf, new BerBoolean(moreFollows));
	}

	private ReadResponse handleGetDataValuesRequest(ReadRequest mmsReadRequest, Association association)
			throws ServiceError {

		VariableAccessSpecification variableAccessSpecification = mmsReadRequest.variableAccessSpecification;

		List<AccessResult> listOfAccessResult = new LinkedList<AccessResult>();

		if (mmsReadRequest.specificationWithResult == null || mmsReadRequest.specificationWithResult.val == false) {
			logger.debug("Got a GetDataValues (MMS) request.");

			if (variableAccessSpecification.listOfVariable == null) {
				throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
						"handleGetDataValuesRequest: Got an invalid MMS packet");
			}

			List<ListOfVariableSeq> listOfVariable = variableAccessSpecification.listOfVariable.seqOf;

			if (listOfVariable.size() != 1) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
						"handleGetDataValuesRequest: more than one variableAccessSpecification is not allowed");
			}

			ListOfVariableSeq variableSeq = listOfVariable.iterator().next();

			VariableSpecification varSpec = variableSeq.variableSpecification;

			if (varSpec.name == null) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
						"handleGetDataValuesRequest: Name is not selected in valiableSpecification");
			}

			ObjectName objName = varSpec.name;
			if (objName.domain_specific == null) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
						"handleGetDataValuesRequest: domain_specific is not selected in ObjectName");
			}
			SubSeq_domain_specific domSpec = objName.domain_specific;

			String mmsDomainID = domSpec.domainId.toString();
			String mmsItemId = domSpec.itemId.toString();

			ObjectReference objectReference = MMSReference.toObjectReference(mmsDomainID, mmsItemId);
			FunctionalConstraint fc = MMSReference.getFunctionalConstraint(mmsItemId);

			if (fc == null) {
				String funcCon = mmsItemId.substring(mmsItemId.indexOf("$") + 1, mmsItemId.lastIndexOf("$"));

				if (!(funcCon.equals("RP")) && !(funcCon.equals("BR")) && !(funcCon.equals("LG"))) {
					throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
							"handleGetDataValuesRequest: DomainID: " + mmsDomainID + " and ItemID: " + mmsItemId
									+ " doesn't give a valid reference of functionally constraint data.");
				}
			}

			if (objectReference == null || objectReference.size() < 3) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
						"handleGetDataValuesRequest: DomainID: " + mmsDomainID + " and ItemID: " + mmsItemId
								+ " doesn't give a valid reference of functionally constraint data.");
			}

			logger.debug(" ->ObjectReference: {} FC: {}", objectReference, fc);

			ModelNode container;

			if (fc == null) { // it may be either RP or BR

				container = serverACSI.getRCB(association, objectReference);

				if (container == null) {
					throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE, "object not found: " + objectReference);
				}

				if (!(container instanceof ReportControlBlock)) {
					throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "invalid object reference "
							+ objectReference);
				}

				// TODO other types with fc = null??
			}
			else {
				container = serverACSI.getDataValues(association, objectReference.toString(), fc);

			}
			AccessResult accessRes = null;
			Data result = null;
			if (container != null) {
				logger.debug(" will convertModeNode");
				// data = MmsAsn1ComplexTypeConverter
				// .convertModelNodeToMmsData(container);

				result = MmsAsn1ComplexTypeConverter.convertModelNodeToMmsData(container);
			}

			// AccessResult accessResult;
			if (result == null) {
				// DataAccessError dataAccessError = new DataAccessError();
				// dataAccessError.setValue(0L); // TODO korrekten Fehlercode
				// übergeben
				// accessResult.selectFailure(dataAccessError);
				accessRes = new AccessResult(new BerInteger(0L), null, null, null, null, null, null, null, null, null,
						null, null, null, null, null, null);
				if (container == null) {
					logger.debug("DataAccessError");
				}
				else {
					logger.debug("DataAccessError accessing " + container.getReference());
				}
			}
			else {
				accessRes = new AccessResult(null, result.array, result.structure, result.boolean_, result.bit_string,
						result.integer, result.unsigned, result.floating_point, result.octet_string,
						result.visible_string, result.generalized_time, result.binary_time, result.bcd,
						result.booleanArray, result.mMSString, result.utc_time);
			}

			listOfAccessResult.add(accessRes); // exactly one element
		}

		else {
			// then this is a getDataSetValues request

			String dataSetName = MMSReference.toObjectReferenceDS(variableAccessSpecification.variableListName);

			if (dataSetName == "") {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
						"handleGetDataValuesRequest: data set name incorrect");
			}

			DataSet dataSetCopy = ((DataSet) serverACSI.getDataSetDirectory(association, dataSetName)).copy();

			// updates the values
			List<ModelNode> dsMembersRead = serverACSI.getDataSetValues(association, dataSetCopy.getChildren());

			// these should be updated now
			// List<ModelNode> dsMembersRead = dataSetCopy.getChildren();
			// this is a list of the data set members with updated values

			// construct list of access result
			// one access result for each data container in dsMemDataRead

			for (ModelNode dc : dsMembersRead) {
				// data = MmsAsn1ComplexTypeConverter
				// .convertModelNodeToMmsData(dc);

				Data result = MmsAsn1ComplexTypeConverter.convertModelNodeToMmsData(dc);

				AccessResult accessResult = null;
				if (result == null) {
					accessResult = new AccessResult(new BerInteger(0L), null, null, null, null, null, null, null, null,
							null, null, null, null, null, null, null);
				}
				else {
					accessResult = new AccessResult(null, result.array, result.structure, result.boolean_,
							result.bit_string, result.integer, result.unsigned, result.floating_point,
							result.octet_string, result.visible_string, result.generalized_time, result.binary_time,
							result.bcd, result.booleanArray, result.mMSString, result.utc_time);
				}
				listOfAccessResult.add(accessResult);
			}

		}

		SubSeqOf_listOfAccessResult listOf = new SubSeqOf_listOfAccessResult(listOfAccessResult);
		return new ReadResponse(null, listOf);
		// readResponse.setListOfAccessResult(listOfAccessResult);
	}

	/**
	 * //TODO Provisorium, hier ist noch Arbeit erforderlich...
	 * 
	 * @param e
	 * 
	 */
	private MmsPdu createServiceErrorResponse(ServiceError e) {
		// ErrorClassChoiceType errClass = new ErrorClassChoiceType();
		SubChoice_errorClass errClass = null;

		switch (e.getErrorCode()) {
		/*
		 * remotes/origin/master case ServiceError.CLASS_NOT_SUPPORTED:
		 * errClass.selectAccess(1L); //object-access-unsupported einfach mal so
		 * gewählt
		 */
		case ServiceError.NO_ERROR:

			break;
		case ServiceError.INSTANCE_NOT_AVAILABLE:
			errClass = new SubChoice_errorClass(null, null, null, null, null, null, null, new BerInteger(
					e.getErrorCode()), null, null, null, null, null);
			break;
		case ServiceError.INSTANCE_IN_USE:
			errClass = new SubChoice_errorClass(null, null, new BerInteger(e.getErrorCode()), null, null, null, null,
					null, null, null, null, null, null);
			break;
		case ServiceError.ACCESS_VIOLATION:
			errClass = new SubChoice_errorClass(null, null, null, null, null, null, null, new BerInteger(
					e.getErrorCode()), null, null, null, null, null);
			break;
		case ServiceError.ACCESS_NOT_ALLOWED_IN_CURRENT_STATE:
			/*
			 * destination-file? Put into "file" or "others"?
			 */
			errClass = new SubChoice_errorClass(null, null, null, null, null, null, null, null, null, null, null, null,
					new BerInteger(e.getErrorCode()));
			break;
		case ServiceError.INSTANCE_LOCKED_BY_OTHER_CLIENT:
			errClass = new SubChoice_errorClass(null, null, null, null, null, null, null, null, null, null, null,
					new BerInteger(2), null);
			break;
		case ServiceError.TYPE_CONFLICT:
			errClass = new SubChoice_errorClass(null, null, null, null, null, null, null, null, null, null, null,
					new BerInteger(4), null);
			break;
		// case ServiceError.APPLICATION_UNREACHABLE:
		// case ServiceError.MEMORY_UNAVAILABLE:
		// case ServiceError.PROCESSOR_RESOURCE_UNAVAILABLE:
		// case ServiceError.CONNECTION_LOST:
		// case ServiceError.FAILED_DUE_TO_SERVER_CONSTRAINT:
		// case ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT:
		// case ServiceError.CONTROL_MUST_BE_SELECTED:
		// case ServiceError.CLASS_NOT_SUPPORTED:
		// case ServiceError.PARAMETER_VALUE_INCONSISTENT:
		// case ServiceError.PARAMETER_VALUE_INAPPROPRIATE:

		default:
			// errClass.selectOthers(new Long(e.getErrorCode())); // TODO
			errClass = new SubChoice_errorClass(null, null, null, null, null, null, null, null, null, null, null, null,
					new BerInteger(e.getErrorCode()));
		}
		org.openiec61850.jmms.mms.asn1.ServiceError asn1ServiceError = null;
		try {
			asn1ServiceError = new org.openiec61850.jmms.mms.asn1.ServiceError(errClass, null, new BerVisibleString(
					e.getMessage()));
			// asn1ServiceError.setErrorClass(errClass);
		} catch (UnsupportedEncodingException f) {
			logger.warn(e.getMessage());
			asn1ServiceError = new org.openiec61850.jmms.mms.asn1.ServiceError(errClass, null, null);
		}

		// InitiateErrorPdu initErrorPdu = new
		// InitiateErrorPdu(asn1ServiceError);
		// initErrorPdu.setValue(asn1ServiceError);

		MmsPdu mmsPdu = new MmsPdu(null, null, null, null, null, asn1ServiceError);
		// mmsPdu.selectInitiateErrorPdu(initErrorPdu);
		return mmsPdu;
	}

	// private MmsPdu createServiceErrorResponse(ServiceError e) {
	// ErrorClassChoiceType errClass = new ErrorClassChoiceType();
	//
	// switch (e.getErrorCode()) {
	// /*
	// * remotes/origin/master case ServiceError.CLASS_NOT_SUPPORTED:
	// * errClass.selectAccess(1L); //object-access-unsupported einfach mal so
	// * gewählt
	// */
	// default:
	// errClass.selectOthers(new Long(e.getErrorCode())); // TODO
	// }
	//
	// org.openiec61850.jmms.mms.asn1.ServiceError asn1ServiceError = new
	// org.openiec61850.jmms.mms.asn1.ServiceError();
	// asn1ServiceError.setErrorClass(errClass);
	//
	// InitiateErrorPdu initErrorPdu = new InitiateErrorPdu();
	// initErrorPdu.setValue(asn1ServiceError);
	//
	// MmsPdu mmsPdu = new MmsPdu();
	// mmsPdu.selectInitiateErrorPdu(initErrorPdu);
	// return mmsPdu;
	// }

	/**
	 * //TODO Provisorium, hier ist noch Arbeit erforderlich...
	 * 
	 * @param e
	 */
	// protected DataAccessError createMmsDataAccessError(ServiceError e) {
	// long result;
	// switch (e.getErrorCode()) {
	// default:
	// result = 0;
	// }
	// return new DataAccessError(result);
	// }

	// private byte[] encodeMMSpdu(MmsPdu mmsPdu) {
	// ByteArrayOutputStream baos = new ByteArrayOutputStream();
	// try {
	// iencoder.encode(mmsPdu, baos);
	// } catch (Exception e) {
	// throw new RuntimeException("Error encoding MMS PDU", e);
	// }
	// return baos.toByteArray();
	// }

	private GetNameListResponse handleGetDirectoryRequest(GetNameListRequest getNameListRequest, Association association)
			throws ServiceError {

		// the ObjectScope can be vmdSpecific,domainSpecific, or aaSpecific
		// vmdSpecific and aaSpecific are not part of 61850-8-1 but are used by
		// some IEC 61850 clients anyways.
		// this stack will return an empty list on vmdSpecific and aaSpecific
		// requests.

		if (getNameListRequest.objectScope.aaSpecific != null || getNameListRequest.objectScope.vmdSpecific != null) {
			SubSeqOf_listOfIdentifier listOfIden = new SubSeqOf_listOfIdentifier(new Vector<BerVisibleString>());
			GetNameListResponse getNameListResponse = new GetNameListResponse(listOfIden, new BerBoolean(false));
			return getNameListResponse;
		}

		String mmsDomainId = getNameListRequest.objectScope.domainSpecific.toString();

		ModelNode logicalDevice = serverACSI.getDataDefinition(association, new ObjectReference(mmsDomainId), null);
		if (logicalDevice == null || !(logicalDevice instanceof LogicalDevice)) {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
					"Got an invalid MMS request: given Domain name in GetNameList request is not a Logical Device name");
		}

		logger.debug(" ->ObjectReference: {}", mmsDomainId);

		insertRef = true;
		// if (getNameListRequest.isContinueAfterPresent()) {
		// continueAfter = getNameListRequest.getContinueAfter().getValue();
		// logger.debug(" ->continue after: {}", getNameListRequest
		// .getContinueAfter().getValue());
		// insertRef = false;
		// }

		if (getNameListRequest.continueAfter != null) {
			continueAfter = getNameListRequest.continueAfter.toString();
			logger.debug(" ->continue after: {}", continueAfter.toString());
			insertRef = false;

		}

		List<String> mmsReferences = new LinkedList<String>();

		Collection<String> buffered = new LinkedList<String>();
		Collection<String> unbuffered = new LinkedList<String>();

		for (ModelNode logicalNode : logicalDevice) {
			mmsReferences.add(logicalNode.getNodeName());

			boolean reportsAdded = false;

			for (String mmsFC : mmsFCs) {
				FunctionalConstraint fc = FunctionalConstraint.fromString(mmsFC);
				if (fc != null) {

					List<DataObject> dataObjects = ((LogicalNode) logicalNode).getChildren(fc);
					if (dataObjects != null) {
						mmsReferences.add(logicalNode.getNodeName() + "$" + mmsFC);
						for (DataObject dataObject : dataObjects) {
							insertMMSRef(dataObject, mmsReferences, logicalNode.getNodeName() + "$" + mmsFC);

						}
					}
				}
				// TODO need to handle logs?
				else { // it could be RP or BR or LG??

					if (!reportsAdded) {
						List<ReportControlBlock> rcbs = ((LogicalNode) logicalNode).getReportControlBlocks();
						if (!rcbs.isEmpty()) {

							for (ReportControlBlock rcb : rcbs) {
								if (rcb instanceof UnbufferedReportContrlBlock) {

									String main = logicalNode.getNodeName() + "$" + "RP" + "$" + rcb.getNodeName();

									unbuffered.add(main);
									for (ModelNode node : rcb.getChildren()) {

										// if (node instanceof DataSet) {
										// unbuffered.add(main + "$DatSet");
										//
										// }
										// else {
										unbuffered.add(main + "$" + node.getNodeName());

										// }
									}

								}

								else if (rcb instanceof BufferedReportControlBlock) {

									String main = logicalNode.getNodeName() + "$" + "BR" + "$" + rcb.getNodeName();

									buffered.add(main);

									for (ModelNode node : rcb.getChildren()) {

										// if (node instanceof DataSet) {
										// buffered.add(main + "$DatSet");
										//
										// }
										// else {
										buffered.add(main + "$" + node.getNodeName());

										// }
									}

								}
							}

							if (!unbuffered.isEmpty()) {
								mmsReferences.add(logicalNode.getNodeName() + "$" + mmsFC);

								mmsReferences.addAll(unbuffered);
								unbuffered.clear();
							}

							if (!buffered.isEmpty()) {
								mmsReferences.add(logicalNode.getNodeName() + "$" + mmsFC);

								mmsReferences.addAll(buffered);
								buffered.clear();
							}

							reportsAdded = true;

						}
					}
				}
			}
		}

		Vector<BerVisibleString> identifiers = new Vector<BerVisibleString>();

		int identifierSize = 0;
		boolean moreFollows = false;
		for (String mmsReference : mmsReferences) {
			if (insertRef == true) {
				if (identifierSize > negotiatedMaxPduSize - 200) {
					moreFollows = true;
					logger.debug(" ->maxMMSPduSize of " + negotiatedMaxPduSize + " Bytes reached");
					break;
				}

				BerVisibleString identifier = null;

				try {
					identifier = new BerVisibleString(mmsReference);
				} catch (UnsupportedEncodingException e) {
					logger.warn(e.getMessage());
				}
				// identifier.setValue(mmsReference);
				identifiers.add(identifier);
				identifierSize += mmsReference.length() + 2;
			}
			else {
				if (mmsReference.equals(continueAfter)) {
					insertRef = true;
				}
			}
		}

		SubSeqOf_listOfIdentifier listOfIden = new SubSeqOf_listOfIdentifier(identifiers);

		return new GetNameListResponse(listOfIden, new BerBoolean(moreFollows));
	}

	/**
	 * GetVariableAccessAttributes (GetDataDefinition/GetDataDirectory) can be
	 * called with different kinds of references. Examples: 1. DGEN1 2. DGEN1$CF
	 * 3. DGEN1$CF$GnBlk
	 * 
	 */
	private GetVariableAccessAttributesResponse handleGetVariableAccessAttributesRequest(
			GetVariableAccessAttributesRequest getVariableAccessAttributesRequest, Association association)
			throws ServiceError {
		if (getVariableAccessAttributesRequest.name == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Got an invalid MMS packet: name is not selected in GetVariableAccessAttributesRequest");
		}

		if (getVariableAccessAttributesRequest.name.domain_specific == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Got an invalid MMS packet: Domain specific is not selected in GetVariableAccessAttributesRequest");
		}

		String domainId = getVariableAccessAttributesRequest.name.domain_specific.domainId.toString();
		String itemId = getVariableAccessAttributesRequest.name.domain_specific.itemId.toString();

		ObjectReference objectReference = MMSReference.toObjectReference(domainId, itemId);
		FunctionalConstraint reqFc = MMSReference.getFunctionalConstraint(itemId);

		logger.debug(" ->ObjectReference: {} FC: {}", objectReference, reqFc);

		if (objectReference.size() < 2) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Got an invalid MMS packet: domainId and/or itemId are not valid");
		}

		ModelNode modelNode = serverACSI.getDataDefinition(association, objectReference, reqFc);

		if (modelNode == null || !(modelNode instanceof LogicalNode)) {
			throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
					"GetVariableAccessAttributes (GetDataDefinition): no object with domainId " + domainId
							+ " and ItemID " + itemId + " was found. objectReference: " + objectReference + " fc "
							+ reqFc + " modelNode " + modelNode);
		}

		List<StructComponent> structComponents = new LinkedList<StructComponent>();
		List<StructComponent> urcbStructComponents = new LinkedList<StructComponent>();
		List<StructComponent> brcbStructComponents = new LinkedList<StructComponent>();
		boolean rptCompAdded = false;

		for (String mmsFC : mmsFCs) {
			FunctionalConstraint fc = FunctionalConstraint.fromString(mmsFC);
			if (fc != null) {

				List<DataObject> fcDataObjects = ((LogicalNode) modelNode).getChildren(fc);
				if (fcDataObjects != null) {
					List<StructComponent> doStructComponents = new LinkedList<StructComponent>();

					for (ModelNode child : fcDataObjects) {
						doStructComponents.add(getMmsStructComponent(child));
					}

					SubSeqOf_components comp = new SubSeqOf_components(doStructComponents);
					SubSeq_structure struct = new SubSeq_structure(null, comp);

					TypeSpecification fcTypeSpec = new TypeSpecification(null, null, struct, null, null, null, null,
							null, null, null, null, null, null, null, null, null);

					StructComponent structCom = null;
					try {
						structCom = new StructComponent(new BerVisibleString(mmsFC), fcTypeSpec);
					} catch (UnsupportedEncodingException e) {
						logger.warn(e.getMessage());
						throw new IllegalAccessError();
					}
					structComponents.add(structCom);

				}

			}
			else {
				if (!rptCompAdded) {
					List<ReportControlBlock> rcbs = ((LogicalNode) modelNode).getReportControlBlocks();
					if (!rcbs.isEmpty()) {

						for (ReportControlBlock rcb : rcbs) {
							if (rcb instanceof UnbufferedReportContrlBlock) {
								urcbStructComponents.add(getMmsStructComponent(rcb));
							}

							else if (rcb instanceof BufferedReportControlBlock) {

								brcbStructComponents.add(getMmsStructComponent(rcb));
							}
						}

						if (!urcbStructComponents.isEmpty()) {

							SubSeqOf_components comp = new SubSeqOf_components(urcbStructComponents);
							SubSeq_structure struct = new SubSeq_structure(null, comp);

							TypeSpecification fcTypeSpec = new TypeSpecification(null, null, struct, null, null, null,
									null, null, null, null, null, null, null, null, null, null);

							StructComponent structCom = null;
							try {
								structCom = new StructComponent(new BerVisibleString(mmsFC), fcTypeSpec);
							} catch (UnsupportedEncodingException e) {
								logger.warn(e.getMessage());
								throw new IllegalAccessError();
							}
							structComponents.add(structCom);
						}
					}

					if (!brcbStructComponents.isEmpty()) {

						SubSeqOf_components comp = new SubSeqOf_components(brcbStructComponents);
						SubSeq_structure struct = new SubSeq_structure(null, comp);

						TypeSpecification fcTypeSpec = new TypeSpecification(null, null, struct, null, null, null,
								null, null, null, null, null, null, null, null, null, null);
						StructComponent structCom = null;
						try {
							structCom = new StructComponent(new BerVisibleString(mmsFC), fcTypeSpec);
						} catch (UnsupportedEncodingException e) {
							logger.warn(e.getMessage());
							throw new IllegalAccessError();
						}
						structComponents.add(structCom);
					}
					rptCompAdded = true;
				}
				// }

				// TODO handle other mms fcs

			}
		}

		SubSeqOf_components comp = new SubSeqOf_components(structComponents);
		SubSeq_structure struct = new SubSeq_structure(null, comp);

		TypeSpecification typSpec = new TypeSpecification(null, null, struct, null, null, null, null, null, null, null,
				null, null, null, null, null, null);

		return new GetVariableAccessAttributesResponse(new BerBoolean(false), null, typSpec);
		// getVariableAccessAttributesResp.setMmsDeletable(false);
		// getVariableAccessAttributesResp.setTypeSpecification(typeSpecification);

	}

	private StructComponent getMmsStructComponent(ModelNode node) throws ServiceError {

		// structComponent.setComponentName(new Identifier(node.getNodeName()));

		TypeSpecification typeSpecification = getMmsTypeSpecification(node);
		if (typeSpecification == null) {
			return null;
		}
		// structComponent.setComponentType(typeSpecification);
		StructComponent structCom = null;
		try {
			structCom = new StructComponent(new BerVisibleString(node.getNodeName()), typeSpecification);
		} catch (UnsupportedEncodingException e) {
			logger.warn(e.getMessage());
			throw new IllegalAccessError("received an UnsupportedEncodingException");
		}

		return structCom;
	}

	private TypeSpecification getMmsTypeSpecification(ModelNode node) throws ServiceError {

		// TypeSpecification typeSpecification = new TypeSpecification();

		TypeSpecification typeSpec = null;

		// TypeSpecification(ObjectName typeName, SubSeq_array array,
		// SubSeq_structure structure, BerNull boolean_,
		// BerInteger bit_string, BerInteger integer, BerInteger unsigned,
		// SubSeq_floating_point floating_point,
		// BerInteger octet_string, BerInteger visible_string, BerNull
		// generalized_time, BerBoolean binary_time, BerInteger bcd,
		// BerNull objId, BerInteger mMSString, BerNull utc_time) {

		if (node instanceof BasicDataAttribute) {

			// from ISO 9506-1 Section 14.2.2.4.1

			switch (((BasicDataAttribute) node).getBasicType()) {
			case BOOLEAN:
				// typeSpecification.selectBoolean_();
				typeSpec = new TypeSpecification(null, null, null, new BerNull(), null, null, null, null, null, null,
						null, null, null, null, null, null);
				break;

			// case DBPOS:
			// case TCMD:
			case CODEDENUM:
				// typeSpecification.selectBit_string(new Integer32(2));
				typeSpec = new TypeSpecification(null, null, null, null, new BerInteger(2), null, null, null, null,
						null, null, null, null, null, null, null);
				break;
			// case ENUMERATED:
			// // TODO adjust size dynamically?
			// typeSpecification.selectInteger(new Unsigned8(32));
			// break;
			// case FLOAT32:
			// Floating_pointSequenceType floating_pointSequenceType = new
			// Floating_pointSequenceType();
			// floating_pointSequenceType.setFormat_width(new Unsigned8(32));
			// floating_pointSequenceType.setExponent_width(new Unsigned8(8));
			// typeSpecification
			// .selectFloating_point(floating_pointSequenceType);
			case INT8U:
				typeSpec = new TypeSpecification(null, null, null, null, null, null, new BerInteger(8), null, null,
						null, null, null, null, null, null, null);
				break;
			case INT8:
				// typeSpecification.selectInteger(new Unsigned8(8));
				typeSpec = new TypeSpecification(null, null, null, null, null, new BerInteger(8), null, null, null,
						null, null, null, null, null, null, null);
				break;
			case INT16U:
				typeSpec = new TypeSpecification(null, null, null, null, null, null, new BerInteger(16), null, null,
						null, null, null, null, null, null, null);
				break;
			case INT16:
				// typeSpecification.selectInteger(new Unsigned8(16));
				typeSpec = new TypeSpecification(null, null, null, null, null, new BerInteger(16), null, null, null,
						null, null, null, null, null, null, null);
				break;
			case INT32U:
				typeSpec = new TypeSpecification(null, null, null, null, null, null, new BerInteger(32), null, null,
						null, null, null, null, null, null, null);
				break;
			case INT32:
				// typeSpecification.selectInteger(new Unsigned8(32));
				typeSpec = new TypeSpecification(null, null, null, null, null, new BerInteger(32), null, null, null,
						null, null, null, null, null, null, null);
				break;
			case INT64:
				// typeSpecification.selectInteger(new Unsigned8(64));
				typeSpec = new TypeSpecification(null, null, null, null, null, new BerInteger(64), null, null, null,
						null, null, null, null, null, null, null);
				break;
			// case INT8U:
			// typeSpecification.selectUnsigned(new Unsigned8(8));
			// break;
			// case INT16U:
			// typeSpecification.selectUnsigned(new Unsigned8(16));
			// break;
			// case INT32U:
			// typeSpecification.selectUnsigned(new Unsigned8(32));
			// break;
			case FLOAT32:
				// Floating_pointSequenceType floating_pointSequenceType = new
				// Floating_pointSequenceType();
				// floating_pointSequenceType.setFormat_width(new
				// Unsigned8(32));
				// floating_pointSequenceType.setExponent_width(new
				// Unsigned8(8));
				// typeSpecification
				// .selectFloating_point(floating_pointSequenceType);
				typeSpec = new TypeSpecification(null, null, null, null, null, null, null, new SubSeq_floating_point(
						new BerInteger(32), new BerInteger(8)), null, null, null, null, null, null, null, null);

				break;
			case ENUMERATED:
				// TODO adjust size dynamically?
				// typeSpecification.selectInteger(new Unsigned8(32));
				typeSpec = new TypeSpecification(null, null, null, null, null, new BerInteger(32), null, null, null,
						null, null, null, null, null, null, null);
				break;
			case OCTET_STRING_6:
				if (node.getNodeName().equals("TimeOfEntry")) {
					// typeSpecification.selectBinary_time(true);
					typeSpec = new TypeSpecification(null, null, null, null, null, null, null, null, null, null, null,
							new BerBoolean(true), null, null, null, null);

				}
				else {

					// typeSpecification.selectOctet_string(new Integer32(-6));
					typeSpec = new TypeSpecification(null, null, null, null, null, new BerInteger(-6), null, null,
							null, null, null, null, null, null, null, null);
				}
				break;
			case OCTET_STRING_64:
				// typeSpecification.selectOctet_string(new Integer32(-64));

				typeSpec = new TypeSpecification(null, null, null, null, null, null, null, null, new BerInteger(-64),
						null, null, null, null, null, null, null);

				break;
			case QUALITY:
				// typeSpecification.selectBit_string(new Integer32(-13));
				typeSpec = new TypeSpecification(null, null, null, null, new BerInteger(-13), null, null, null, null,
						null, null, null, null, null, null, null);
				break;
			case TIMESTAMP:
				// typeSpecification.selectUtc_time();
				typeSpec = new TypeSpecification(null, null, null, null, null, null, null, null, null, null, null,
						null, null, null, null, new BerNull());
				break;
			// case UNICODE_STRING_255:
			case VISIBLE_STRING_255:
				typeSpec = new TypeSpecification(null, null, null, null, null, null, null, null, null, new BerInteger(
						-255), null, null, null, null, null, null);
				break;
			case UNICODE_STRING_255:
				// typeSpecification.selectMMSString(new Integer32(-255));

				typeSpec = new TypeSpecification(null, null, null, null, null, null, null, null, null, null, null,
						null, null, null, new BerInteger(-255), null);
				break;
			case VISIBLE_STRING_32:
				// typeSpecification.selectVisible_string(new Integer32(-32));
				typeSpec = new TypeSpecification(null, null, null, null, null, null, null, null, null, new BerInteger(
						-32), null, null, null, null, null, null);
				break;
			case VISIBLE_STRING_64:
				// typeSpecification.selectVisible_string(new Integer32(-64));
				typeSpec = new TypeSpecification(null, null, null, null, null, null, null, null, null, new BerInteger(
						-64), null, null, null, null, null, null);
				break;
			case VISIBLE_STRING_129:
				// typeSpecification.selectVisible_string(new Integer32(-129));
				typeSpec = new TypeSpecification(null, null, null, null, null, null, null, null, null, new BerInteger(
						-129), null, null, null, null, null, null);
				break;

			default:
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
						"getMmsTypeSpecification: unknown BasicType");
			}
		}

		else if (node instanceof Array) {
			Array array = (Array) node;
			// ArraySequenceType mmsArray = new ArraySequenceType();
			// mmsArray.setNumberOfElements(new Unsigned32((long)
			// array.size()));
			TypeSpecification arrayItemTypeSpec = getMmsTypeSpecification(array.getChild("0"));

			if (arrayItemTypeSpec == null) {
				return null;
			}
			// mmsArray.setElementType(arrayItemTypeSpec);
			SubSeq_array subArray = new SubSeq_array(null, new BerInteger(array.size()), arrayItemTypeSpec);
			// typeSpecification.selectArray(mmsArray);
			typeSpec = new TypeSpecification(null, subArray, null, null, null, null, null, null, null, null, null,
					null, null, null, null, null);
		}

		else if (node instanceof Server || node instanceof LogicalDevice || node instanceof LogicalNode
				|| node instanceof DataObject || node instanceof ConstructedDataAttribute
				|| node instanceof ReportControlBlock) {

			List<StructComponent> structComponents = new LinkedList<StructComponent>();
			StructComponent childStructComponent;
			for (ModelNode child : node) {

				// if (node instanceof ReportControlBlock && child instanceof
				// DataSet) {
				// // TypeSpecification typSpec4 = new TypeSpecification();
				// // typSpec4.selectVisible_string(new Integer32(-64));
				// // childStructComponent = new StructComponent();
				// // childStructComponent.setComponentName(new Identifier(
				// // "DatSet"));
				// //
				// // childStructComponent.setComponentType(typSpec4);
				// TypeSpecification typ = new TypeSpecification(
				// null, null, null, null, null, null, null, null, null, new
				// BerInteger(-64), null, null,
				// null, null, null, null);
				//
				// try {
				// childStructComponent = new StructComponent(new
				// BerVisibleString(
				// "DatSet"), typ);
				// } catch (UnsupportedEncodingException e) {
				// logger.warn(e.getMessage());
				// throw new IllegalAccessError();
				// }
				// }
				//
				// else {
				childStructComponent = getMmsStructComponent(child);
				// }

				if (childStructComponent != null) {
					structComponents.add(childStructComponent);
				}
			}
			if (structComponents.isEmpty()) {
				return null;
			}

			SubSeqOf_components componentsSequenceType = new SubSeqOf_components(structComponents);
			SubSeq_structure structure = new SubSeq_structure(null, componentsSequenceType);

			// typeSpecification.selectStructure(structureSequenceType);
			typeSpec = new TypeSpecification(null, null, structure, null, null, null, null, null, null, null, null,
					null, null, null, null, null);
		}

		else if (node instanceof TriggerConditions) {
			// typeSpecification.selectBit_string(new Integer32(-8));

			typeSpec = new TypeSpecification(null, null, null, null, new BerInteger(-8), null, null, null, null, null,
					null, null, null, null, null, null);
		}

		else if (node instanceof OptFields) {
			// typeSpecification.selectBit_string(new Integer32(-10));
			typeSpec = new TypeSpecification(null, null, null, null, new BerInteger(-10), null, null, null, null, null,
					null, null, null, null, null, null);
		}

		else {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
					"getMmsTypeSpecification: unknown ObjectType: ");
		}

		return typeSpec;
	}

	// private DataDefinition buildDataDefTree(String ldName) throws
	// ServiceError {
	// DataDefinition ldDD = null;
	// boolean ldFound = false;
	// for (DataDefinition myldDD : ldDDs) {
	// if (myldDD.getNodeName().equals(ldName)) {
	// ldFound = true;
	// ldDD = myldDD;
	// }
	// }
	//
	// if (ldFound == false) {
	// ldDD = new DataDefinition(ldName, ObjectType.LOGICAL_DEVICE);
	// List<String> lnRefs = serverACSI.getLogicalDeviceDirectory(association,
	// ldName);
	// for (String lnRef : lnRefs) {
	// DataDefinition lnDD = new DataDefinition(new
	// ObjectReference(lnRef).getLNName(),
	// ObjectType.LOGICAL_NODE);
	// List<String> doRefs = serverACSI.getLogicalNodeDirectory(association,
	// lnRef, ACSIClass.DATA_OBJECT);
	// for (String doRef : doRefs) {
	// DataDefinition doDD = new DataDefinition(new
	// ObjectReference(doRef).getDOName(),
	// ObjectType.DATA_OBJECT);
	// List<DataDefinition> subdoDDs = serverACSI.getDataDefinition(association,
	// doRef);
	// for (DataDefinition subdoDD : subdoDDs) {
	// doDD.addChild(subdoDD);
	// }
	// lnDD.addChild(doDD);
	// }
	// ldDD.addChild(lnDD);
	// }
	// ldDDs.add(ldDD);
	// }
	// return ldDD;
	// }

	private static void insertMMSRef(ModelNode node, List<String> mmsRefs, String parentRef) {

		String ref = parentRef + '$' + node.getNodeName();
		mmsRefs.add(ref);
		if (!(node instanceof Array)) {
			for (ModelNode childNode : node) {
				insertMMSRef(childNode, mmsRefs, ref);
			}
		}
	}

	private GetNameListResponse handleGetServerDirectoryRequest(GetNameListRequest getNameListRequest,
			Association association) throws ServiceError {
		List<String> ldNames = null;

		ldNames = serverACSI.getServerDirectory(association, ServerDirectoryType.LOGICAL_DEVICES);

		Vector<BerVisibleString> identifiers = new Vector<BerVisibleString>();
		BerVisibleString identifier = null;

		for (String ldName : ldNames) {
			try {
				identifier = new BerVisibleString(ldName);
			} catch (UnsupportedEncodingException e) {
				logger.warn(e.getMessage());
			}
			identifiers.add(identifier);
		}

		GetNameListResponse getNameListResponse = new GetNameListResponse(new SubSeqOf_listOfIdentifier(identifiers),
				new BerBoolean(false));

		return getNameListResponse;
	}

	// Data Set functions
	private BerNull handleDefineNamedVariableListRequest(
			DefineNamedVariableListRequest mmsDefineNamedVariableListRequest, Association association)
			throws ServiceError {
		// get Data Set name
		String dataSetName = MMSReference.toObjectReferenceDS(mmsDefineNamedVariableListRequest.variableListName);
		if (dataSetName.isEmpty()) {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
					"handleDefineNamedVariableListRequest: invalid MMS request (No DataSet Name Specified)");
		}

		// get list of DO/DA references within Data Set
		List<DefineNamedVariableListRequest.SubSeqOf_listOfVariable.SubSeq> nameList = mmsDefineNamedVariableListRequest.listOfVariable.seqOf;
		List<String> dsMemberList = new ArrayList<String>();
		List<FunctionalConstraint> memFC = new ArrayList<FunctionalConstraint>();
		ObjectName dsMember;
		for (DefineNamedVariableListRequest.SubSeqOf_listOfVariable.SubSeq a : nameList) {
			dsMember = a.variableSpecification.name;

			dsMemberList.add((MMSReference.toObjectReference(dsMember.domain_specific.domainId.toString(),
					dsMember.domain_specific.itemId.toString())).toString());
			memFC.add(MMSReference.getFunctionalConstraint(dsMember.domain_specific.itemId.toString()));
		}

		serverACSI.createDataSet(association, dataSetName, dsMemberList, memFC);

		return new BerNull(); // this is empty

	}

	private GetNamedVariableListAttributesResponse handleGetNamedVariableListAttritubes(
			ObjectName mmsGetNamedVarListAttReq, Association association) throws ServiceError {

		// get name of the data set
		String dataSetName = MMSReference.toObjectReferenceDS(mmsGetNamedVarListAttReq);

		DataSet dataSet = (DataSet) serverACSI.getDataSetDirectory(association, dataSetName);

		List<ModelNode> dataSetDirectory = dataSet.getChildren();
		List<GetNamedVariableListAttributesResponse.SubSeqOf_listOfVariable.SubSeq> listOfVariable = new ArrayList<GetNamedVariableListAttributesResponse.SubSeqOf_listOfVariable.SubSeq>();
		ObjectName objRef;

		for (ModelNode mem : dataSetDirectory) {
			objRef = MMSReference.toMmsObjectName(mem.getReference(), mem.getFunctionalConstraint());

			VariableSpecification varSpec = new VariableSpecification(objRef, null, null, null, null);

			listOfVariable
					.add(new GetNamedVariableListAttributesResponse.SubSeqOf_listOfVariable.SubSeq(varSpec, null));
		}
		return new GetNamedVariableListAttributesResponse(new BerBoolean(dataSet.isDeletable()),
				new SubSeqOf_listOfVariable(listOfVariable));
	}

	private DeleteNamedVariableListResponse handleDeleteNamedVariableListRequest(
			DeleteNamedVariableListRequest mmsDelNamVarListReq, Association association) throws ServiceError {
		// get the name of the data set
		List<ObjectName> listOfVariableListName = mmsDelNamVarListReq.listOfVariableListName.seqOf;

		int[] total = new int[2];
		int[] numberMatchedDeleted; // order: numberMatched first then
		// numberDeleted

		for (ObjectName dsName : listOfVariableListName) {

			numberMatchedDeleted = serverACSI.deleteDataSet(association, MMSReference.toObjectReferenceDS(dsName));
			total[0] += numberMatchedDeleted[0];
			total[1] += numberMatchedDeleted[1];
		}
		return new DeleteNamedVariableListResponse(new BerInteger(total[0]), new BerInteger(total[1]));

	}

	public void sendReport(Report report) throws IOException {
		try {
			// InformationReport infoReport = new InformationReport();

			// ObjectName rptName = new ObjectName();
			// rptName.selectVmd_specific(new Identifier("RPT"));
			//
			// VariableAccessSpecification varAccessSpec = new
			// VariableAccessSpecification();
			// varAccessSpec.selectVariableListName(rptName);
			//
			// infoReport.setVariableAccessSpecification(varAccessSpec);

			VariableAccessSpecification varAccSpec = new VariableAccessSpecification(null, new ObjectName(
					new BerVisibleString("RPT"), null, null));

			OptFields optFlds = report.getOptFlds();

			List<AccessResult> listOfAccessResult = new LinkedList<AccessResult>();

			// rptID

			AccessResult accessRes1 = new AccessResult(null, null, null, null, null, null, null, null, null,
					new BerVisibleString(report.getRptId()), null, null, null, null, null, null);
			listOfAccessResult.add(accessRes1);

			// Reported OptFlds
			// Data rptOptFlds = new Data();
			// rptOptFlds.selectBit_string(new BitString(report.getOptFlds()
			// .getValue()));
			// AccessResult accessRes2 = new AccessResult();
			// accessRes2.selectSuccess(rptOptFlds);
			AccessResult accessRes2 = new AccessResult(null, null, null, null, new BerBitString(report.getOptFlds()
					.getValue(), 10), null, null, null, null, null, null, null, null, null, null, null);

			listOfAccessResult.add(accessRes2);

			if (optFlds.isSeqNum()) {
				// Data seqNum = new Data();
				// seqNum.selectUnsigned(new Long(report.getSqNum()));
				// AccessResult accessRes3 = new AccessResult();
				// accessRes3.selectSuccess(seqNum);
				AccessResult accessRes3 = new AccessResult(null, null, null, null, null, null, new BerInteger(
						report.getSqNum()), null, null, null, null, null, null, null, null, null);
				listOfAccessResult.add(accessRes3);
			}

			if (optFlds.isTimeStamp()) {
				// Data timeStamp = new Data();
				// timeStamp.selectBinary_time(new TimeOfDay(
				// MmsAsn1PrimitiveTypeConverter.toByteArray(report
				// .getTimeOfEntry())));
				// AccessResult accessRes4 = new AccessResult();
				// accessRes4.selectSuccess(timeStamp);
				AccessResult accessRes4 = new AccessResult(null, null, null, null, null, null, null, null, null, null,
						null, new BerOctetString(MmsAsn1PrimitiveTypeConverter.toByteArray(report.getTimeOfEntry())),
						null, null, null, null);
				listOfAccessResult.add(accessRes4);
			}

			if (optFlds.isDataSet()) {
				// Data dataSet = new Data();
				// dataSet.selectVisible_string(report.getDataSet().replace(".",
				// "$"));
				// AccessResult accessRes5 = new AccessResult();
				// accessRes5.selectSuccess(dataSet);
				AccessResult accessRes5 = new AccessResult(null, null, null, null, null, null, null, null, null,
						new BerVisibleString(report.getDataSet().replace(".", "$")), null, null, null, null, null, null);
				listOfAccessResult.add(accessRes5);
			}

			if (optFlds.isBufOvfl()) {
				// Data bufOvfl = new Data();
				// bufOvfl.selectBoolean_(report.getBufOvfl());
				// AccessResult accessRes6 = new AccessResult();
				// accessRes6.selectSuccess(bufOvfl);
				AccessResult accessRes6 = new AccessResult(null, null, null, new BerBoolean(report.getBufOvfl()), null,
						null, null, null, null, null, null, null, null, null, null, null);
				listOfAccessResult.add(accessRes6);
			}

			if (optFlds.isEntryId()) {
				// Data entryId = new Data();
				// entryId.selectOctet_string(MmsAsn1PrimitiveTypeConverter
				// .toByteArray(report.getEntryId()));
				// AccessResult accessRes7 = new AccessResult();
				// accessRes7.selectSuccess(entryId);
				AccessResult accessRes7 = new AccessResult(null, null, null, null, null, null, null, null,
						new BerOctetString(MmsAsn1PrimitiveTypeConverter.toByteArray(report.getEntryId())), null, null,
						null, null, null, null, null);
				listOfAccessResult.add(accessRes7);
			}

			if (optFlds.isConfigRef()) {
				// Data confRev = new Data();
				// confRev.selectUnsigned(report.getConvRev());
				// AccessResult accessRes8 = new AccessResult();
				// accessRes8.selectSuccess(confRev);
				AccessResult accessRes8 = new AccessResult(null, null, null, null, null, null, new BerInteger(
						report.getConvRev()), null, null, null, null, null, null, null, null, null);
				listOfAccessResult.add(accessRes8);
			}

			// TODO check if segmentation number is used

			// if (optFlds.isSeqNum()) {
			// Data subSeqNum = new Data();
			// subSeqNum.selectUnsigned(new Long(report.getSubSqNum()));
			// AccessResult accessRes9 = new AccessResult();
			// accessRes9.selectSuccess(subSeqNum);
			// listOfAccessResult.add(accessRes9);
			//
			// Data moreFollows = new Data();
			// moreFollows.selectBoolean_(report.getMoreSegmentsFollow());
			// AccessResult accessRes10 = new AccessResult();
			// accessRes10.selectSuccess(moreFollows);
			// listOfAccessResult.add(accessRes10);
			// }

			// TODO padding required?
			// Data inclusionString = new Data();
			// inclusionString.selectBit_string(new BitString(report
			// .getInclusionBitString()));
			// AccessResult accessRes11 = new AccessResult();
			// accessRes11.selectSuccess(inclusionString);
			AccessResult accessRes11 = new AccessResult(null, null, null, null, new BerBitString(
					report.getInclusionBitString(), 8), null, null, null, null, null, null, null, null, null, null,
					null);
			listOfAccessResult.add(accessRes11);

			List<ReportEntryData> entryData = report.getEntryData();
			List<AccessResult> dataRefs = new ArrayList<AccessResult>();
			List<AccessResult> values = new ArrayList<AccessResult>();
			List<AccessResult> reasonCodes = new ArrayList<AccessResult>();

			AccessResult dataRefAC;
			// Data dataRef;
			AccessResult reasonCodeAC;
			AccessResult value;
			// Data reasonCode;
			byte[] reasonCodeValue = new byte[1];

			for (ReportEntryData data : entryData) {
				if (optFlds.isDataRef()) {
					// dataRef = new Data();
					// dataRef.selectVisible_string(data.getValue().getReference()
					// .get(0)
					// + "/"
					// + MMSReference.getMMSItemId(data.getValue()
					// .getReference(), data.getValue()
					// .getFunctionalConstraint()));
					// dataRefAC = new AccessResult();
					// dataRefAC.selectSuccess(dataRef);
					dataRefAC = new AccessResult(null, null, null, null, null, null, null, null, null,
							new BerVisibleString(data.getValue().getReference().get(0)
									+ "/"
									+ MMSReference.getMMSItemId(data.getValue().getReference(), data.getValue()
											.getFunctionalConstraint())), null, null, null, null, null, null);
					dataRefs.add(dataRefAC);
				}

				Data result = MmsAsn1ComplexTypeConverter.convertModelNodeToMmsData(data.getValue());

				if (result == null) {
					value = new AccessResult(new BerInteger(0L), null, null, null, null, null, null, null, null, null,
							null, null, null, null, null, null);
				}
				else {
					value = new AccessResult(null, result.array, result.structure, result.boolean_, result.bit_string,
							result.integer, result.unsigned, result.floating_point, result.octet_string,
							result.visible_string, result.generalized_time, result.binary_time, result.bcd,
							result.booleanArray, result.mMSString, result.utc_time);
				}

				values.add(value);

				if (optFlds.isReasonCode()) {
					// reasonCode = new Data();
					switch (data.getReasonCode()) {
					case DCHG:
						reasonCodeValue[0] = (byte) (reasonCodeValue[0] | (1 << 6));
						break;
					case QCHG:
						reasonCodeValue[0] = (byte) (reasonCodeValue[0] | (1 << 5));
						break;
					case DUPD:
						reasonCodeValue[0] = (byte) (reasonCodeValue[0] | (1 << 4));
						break;
					case INTEGRITY:
						reasonCodeValue[0] = (byte) (reasonCodeValue[0] | (1 << 3));
						break;
					case GI:
						reasonCodeValue[0] = (byte) (reasonCodeValue[0] | (1 << 2));
						break;
					case APPTRIGGER:
						reasonCodeValue[0] = (byte) (byte) (reasonCodeValue[0] | (1 << 1));
					}
					// reasonCode.selectBit_string(new
					// BitString(reasonCodeValue,
					// 1));
					// reasonCodeAC = new AccessResult();
					// reasonCodeAC.selectSuccess(reasonCode);
					reasonCodeAC = new AccessResult(null, null, null, null, new BerBitString(reasonCodeValue, 8), null,
							null, null, null, null, null, null, null, null, null, null);
					reasonCodes.add(reasonCodeAC);
				}
			}

			listOfAccessResult.addAll(dataRefs);
			listOfAccessResult.addAll(values);
			listOfAccessResult.addAll(reasonCodes);

			// infoReport.setListOfAccessResult(listOfAccessResult);
			// UnconfirmedService unconfirmedSer = new UnconfirmedService();
			// unconfirmedSer.selectInformationReport(infoReport);
			//
			// UnconfirmedPDU unconfirmedPDU = new UnconfirmedPDU();
			// unconfirmedPDU.setUnconfirmedService(unconfirmedSer);
			//
			// MmsPdu mmsResponsePdu = new MmsPdu();
			// mmsResponsePdu.selectUnconfirmedPDU(unconfirmedPDU);
			// sendAnMmsPdu(pConnection, mmsResponsePdu);
			InformationReport infoReport = new InformationReport(varAccSpec,
					new InformationReport.SubSeqOf_listOfAccessResult(listOfAccessResult));
			UnconfirmedService unconfirmedSer = new UnconfirmedService(infoReport);
			MmsPdu mmsResponsePdu = new MmsPdu(null, null, new UnconfirmedPDU(unconfirmedSer), null, null, null);
			sendAnMmsPdu(pConnection, mmsResponsePdu);
		} catch (ServiceError e) {
			logger.warn(e.getMessage());
			sendAnMmsPdu(pConnection, createServiceErrorResponse(e));
		}
	}

	public void close() {
		if (pConnection != null) {
			pConnection.disconnect();
		}
	}

}
